/*
 * irc-gw.c, an Internet Relay Chat proxy designed to work with TIS fwtk
 *
 * (C) Copyright 1998 by ArkanoiD <ark@mpak.convey.ru>
 * based on "BNC" program by James Seter (noonie@toledolink.com)
 * DCC proxy code from "tircproxy" by Bjarni R. Einarsson (bre@mmedia.is)
 * Version 1.0beta patchlevel 2
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <netdb.h>
#include <signal.h>
#include <fcntl.h>
#include "firewall.h"

typedef struct {
	int 	inbytes;
	int 	outbytes;
} iostats;

#define	OP_DENY	001
#define	OP_LOG	002
#define	OP_AUTH	004
#define	OP_XTND	010

#define MINPORT 1024
#define DGETS_TIMEOUT 10

typedef struct {
	char 	*name;
	int	flg;
} irc_action;

static 	irc_action	dcc_actions[] = {
{	"send",		0 },
{	"get",		0 },
{	"chat",		0 },
{	"xmit",		0 },
{	"unknown",	0 },
{	0,		0 },
};

typedef unsigned long   ipaddr_t;

int mu,dp,cu,po;
char ps[20];

static  char *proxy_dcc(int destaddr, int destport, ipaddr_t proxy_ip, char *cmd, int isclient, int dcc_log);

static	int			logmore = 0; 		/* obsolete */
static	int			transparent = 0;
static	char			fakeduser[12] ="";
static	char			fakedhost[512] ="";
static	char			plug_server[512] ="";
static  char 			riaddr[512];
static  char 			rladdr[512];
static  char			**validests = (char **)0;
static  Cfg                     *confp;
static	time_t			ontime;
static	time_t			offtime;
static	int			timeout = PROXY_TIMEOUT;

ipaddr_t 	cside_ip = 0;   /* our IP as we appear to client */
ipaddr_t 	sside_ip = 0;   /* our IP as we appear to server */

int 		iostat_fd[2];	/* fd pipe to collect dcc statistics */

extern void 	waitwaitwait();

int do_connect(char *hostname, u_short port)
{
	int s;
	struct hostent *he;
	struct sockaddr_in sin;

	memset(&sin, 0, sizeof(struct sockaddr_in));
	sin.sin_family = AF_INET;
	sin.sin_port = htons(port);
	sin.sin_addr.s_addr = inet_addr(hostname);
	if(sin.sin_addr.s_addr == INADDR_NONE)
	{
		he = gethostbyname(hostname);
		if(!he)
			return -1;
		memcpy(&sin.sin_addr, he->h_addr, he->h_length);
	}
	s = socket(AF_INET, SOCK_STREAM, 0);
	if(s < 0)
		return -2;
	if(connect(s, (struct sockaddr *)&sin, sizeof(sin)) < 0)
	{
		close(s);
		return -2;
	}
	return s;
}

oktotalkto()
{                       
        Cfg     *cf;
        int     x;
   
        cf = cfg_get("hosts",confp);
        while(cf != (Cfg *)0) {
                if(cf->argc < 1)
                        goto skip;

                for(x = 0; x < cf->argc && cf->argv[x][0] != '-'; x++) {
                        if(hostmatch(cf->argv[x],riaddr)) {
                                if(cf->flags & PERM_DENY) {
                                        syslog(LLEV,"deny host=%.512s/%.20s use of gateway",rladdr,riaddr);  
                                        return(0);
                                }
                                syslog(LLEV,"permit host=%.512s/%.20s use of gateway",rladdr,riaddr);
                                return(acceptrule(cf) == 0);
                        }
                }
skip:
                cf = cfg_get("hosts",(Cfg*)0);
        } 
        syslog(LLEV,"deny host=%.512s/%.20s use of gateway",rladdr,riaddr);
        return(0);
}

static  void                    
accept_setdest(v,c)
char    *v;
int     c;      
{                       
        int     dests = 0;

        if(validests == (char **)0)
                validests = (char **)malloc(sizeof(char *) * 2);
        else {
                for(dests = 0; validests[dests] != (char *)0; dests++)
                        ;
                validests = (char **)realloc(validests,(dests + 2) * sizeof(char *));
        }
        if(validests == (char **)0)
                return;
        validests[dests] = v;   
        validests[dests + 1] = (char *)0;
}

static	void
accept_setdcclog(v,c)
char	*v;
int	c;
{
	irc_action	*act;

	for (act = dcc_actions; act->name != (char *)0; act++)
		if(!strcasecmp(v,act->name)) {
			act->flg |= OP_LOG;
			return;
		}
	if (act->name == (char *)0)
		syslog(LLEV,"fwtkcfgerr: config line %d: bad operand %.100s",c,v);
}

static	void
accept_setdccdeny(v,c)
char	*v;
int	c;
{
	irc_action	*act;

	for (act = dcc_actions; act->name != (char *)0; act++)
		if(!strcasecmp(v,act->name)) {
			act->flg |= OP_DENY;
			return;
		}
	if (act->name == (char *)0)
		syslog(LLEV,"fwtkcfgerr: config line %d: bad operand %.100s",c,v);
}
	
acceptrule(c)
Cfg     *c;
{
        int     x;
        void    (*op)();
 
        for(x = 1; x < c->argc; x++) {
                if(c->argv[x][0] != '-')
                        continue;

		if (!strcmp(c->argv[x], "-pass")) {
			x++;
			if (x >= c->argc) {
				syslog(LLEV,"fwtkcfgerr: -pass needs an argument, line %d",c->ln);
				exit(1);
			}
			strncpy(ps, c->argv[x], strlen(c->argv[x]));
			ps[strlen(c->argv[x])] = '\0';
			po=1;
			continue;
		}

                if (!strcmp(c->argv[x], "-bogus-host")) {
                        x++;
                        if (x >= c->argc) {
                                syslog(LLEV,"fwtkcfgerr: -bogus-host needs an argument, line %d",c->ln);
                                exit(1);
                        }
                        strncpy(fakedhost, c->argv[x], strlen(c->argv[x]));
                        fakedhost[strlen(c->argv[x])] = '\0';
                        continue;
                }

                if (!strcmp(c->argv[x], "-bogus-user")) {
                        x++;
                        if (x >= c->argc) {
                                syslog(LLEV,"fwtkcfgerr: -bogus-user needs an argument, line %d",c->ln);
                                exit(1);
                        }
                        strncpy(fakeduser, c->argv[x], strlen(c->argv[x]));
                        fakeduser[strlen(c->argv[x])] = '\0';
                        continue;
                }

		if (!strcmp(c->argv[x], "-plug-to")) {
                        x++;
                        if (x >= c->argc) {
                                syslog(LLEV,"fwtkcfgerr: -plug-to needs an argument, line %d",c->ln);
                                exit(1);
                        }
                        strncpy(plug_server, c->argv[x], strlen(c->argv[x]));
                        plug_server[strlen(c->argv[x])] = '\0';
                        continue;
                }

		if (!strcmp(c->argv[x], "-client-side")) {
                        x++;
                        if (x >= c->argc) {
                                syslog(LLEV,"fwtkcfgerr: -client-side needs an argument, line %d",c->ln);
                                exit(1);
                        }
                        cside_ip=ntohl(inet_addr(c->argv[x]));
                        continue;
                }

		if (!strcmp(c->argv[x], "-server-side")) {
                        x++;
                        if (x >= c->argc) {
                                syslog(LLEV,"fwtkcfgerr: -server-side needs an argument, line %d",c->ln);
                                exit(1);
                        }
                        sside_ip=ntohl(inet_addr(c->argv[x]));
                        continue;
                }
		if(!strcmp(c->argv[x],"-log")) {
			logmore = 1;
			continue;
		}

		if(!strcmp(c->argv[x],"-transparent")) {
			transparent = 1;
			continue;
		}

		/* options that take parameters and lists */


		op = 0;
		if(!strcmp(c->argv[x],"-dest"))
			op = accept_setdest;
		if(!strcmp(c->argv[x],"-dcc-log"))
			op = accept_setdcclog;
		if(!strcmp(c->argv[x],"-dcc-deny"))
			op = accept_setdccdeny;
		if(op == 0) {
			syslog(LLEV,"fwtkcfgerr: bad option line %d: %.100s",c->ln,c->argv[x]);
			exit(1);
		}
		if(++x >= c->argc) {
			syslog(LLEV,"fwtkcfgerr: malformed line %d: missing option",c->ln);
			exit(1);
		}
                if(!strcmp(c->argv[x],"{")) {
                        while(1) {
                                if(++x >= c->argc) {
                                        syslog(LLEV,"fwtkcfgerr: malformed line %d: missing option",c->ln);
                                        exit(1);
                                }
                                if(!strcmp(c->argv[x],"}"))
                                        break;
                                (*op)(c->argv[x],c->ln);
                        }
                } else
                        (*op)(c->argv[x],c->ln);

        }
	return(0);
}

notice(fd,nick,s)
int	fd;
char	*s;
char	*nick;

{
	char	buffer[1024];
	sprintf(buffer,"NOTICE %.10s :%.1000s\r\n",nick,s);
	if (write(fd, buffer, strlen(buffer)) != strlen(buffer))
		return(1);
}
	
showfile(fd,nick,fn)
int	fd;
char	*nick;
char	*fn;
{
	FILE    *f;
        char    buf[BUFSIZ];
        char    yuf[BUFSIZ];
        char    *c;
        int     x;              

	if((f = fopen(fn,"r")) == (FILE *)0)
                return(1);
        while(fgets(buf,sizeof(buf),f) != (char *)0) {
                if((c = index(buf,'\n')) != (char *)0)
                        *c = '\0';
                x = fgetc(f);
                if(x == EOF)
			sprintf(yuf,"NOTICE %.10s :%.256s\r\n",nick,buf);
		else {
			sprintf(yuf,"NOTICE %.10s :%.256s\r\n",nick,buf);
			ungetc(x,f);
		}
		if(write(fd,yuf,strlen(yuf))!= strlen(yuf)) {
			fclose(f);
			return(1);
		}
	}
	fclose(f);
	return(0);
}

baddest(fd,nick,dest)
int     fd;
char    *dest;
char	*nick;
{
        Cfg     *cf;
        char    buf[BUFSIZ];

        syslog(LLEV,"deny host=%.512s/%.20s connect to %.256s",rladdr,riaddr,dest);
if((cf = cfg_get("denydest-msg",confp)) != (Cfg *)0) {
                if(cf->argc != 1) {
                        syslog(LLEV,"fwtkcfgerr: denydest-msg must have one parameter, line %d",cf->ln);
                        return(1);
                }
                if(showfile(fd,nick,cf->argv[0])) {
                        syslog(LLEV,"fwtksyserr: cannot display denydest-msg %.512s: %m",cf->argv[0]);
                        return(1);
                }
                return(0);
        }
        sprintf(buf,"You are not permitted to connect to %.512s",dest);
        return(notice(fd,nick,buf));
}


/*
 * Bind to the specified ip and port.
 */
static int bind_to_port(bind_port,backlog)
short bind_port;
int backlog;
{
	struct sockaddr_in	addr;
	int			sock;

	/*
	 * Allocate a socket.
	 */
	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
		syslog(LLEV,"fwtksyserr: Failed to create socket");
		exit(1);
	}

#ifdef DEBUG
	/*
	 * Set the SO_REUSEADDR option for debugging.
	 */
	{
	 	int on = 1;

		setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on));
	}
#endif

	/*
	 * Set the address to listen to.
	 */
	addr.sin_family = AF_INET;
	addr.sin_port = htons(bind_port);
	addr.sin_addr.s_addr = INADDR_ANY;

	/*
	 * Bind our socket to the above address.
	 */
	if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0)
	{
		syslog(LLEV,"fwtksyserr: bind() failed");
		exit(1);
	}

	/*
	 * Establish a large listen backlog.
	 */
	if (listen(sock, backlog) < 0)
	{
		syslog(LLEV,"fwtksyserr: Listen failed");
		exit(1);
	}

	return (sock);
}


static int filtered_write(int to_sock, char *buff, int blen, 
					ipaddr_t proxy_ip,
					char *nick,
			   		int isclient)
{
  	char				addr[512];
	char				out_string[1024]="";
	char				ctcp_string[1024];
	char				ctcp_out[1024];
	char				*dcc_cmd;
	char				*dcc_fname;
	char				*text_string;
   	unsigned int	  	 	dcc_addr, dcc_port, dcc_size ;
	irc_action			*op;
	int				dcc_log;

	buff[blen] = '\0' ;
	
	while (1)
	{

		text_string=strsep(&buff,"\01");

		if (!(text_string==NULL))
			strncat(out_string,text_string,sizeof(out_string)-strlen(out_string));
		if (buff==NULL) {
/*			syslog(LLEV,"text: %s",text_string); */
			goto parsed;
		}

		strncpy(ctcp_string,strsep(&buff,"\01"),512);
		if (!(ctcp_string==NULL)) {
/*			syslog(LLEV,"ctcp: %s",ctcp_string); */

			if (!strncmp(ctcp_string,"DCC ",4)) {
				char *xx;
				strtok(ctcp_string," ");
				dcc_cmd=strtok(NULL," ");
				dcc_fname=strtok(NULL," ");

				xx=strtok(NULL," ");
				if (xx)
					dcc_addr=strtoul(xx,NULL,10);
				else 	goto baddcc;
				xx=strtok(NULL," ");
				if (xx)
					dcc_port=strtoul(xx,NULL,10);
				else 	goto baddcc;
				xx=strtok(NULL," ");
				if (xx)
					dcc_size=strtoul(xx,NULL,10);
				else 	syslog(LLEV,"warning: dcc requested with size parameter missing");
	
				for (op = dcc_actions; op->name != (char *)0; op++) {
					if (!strcmp(op->name,"send") && !strncasecmp(dcc_cmd,"send",4) && !isclient) {
						if (op->flg & OP_DENY) {
							syslog(LLEV,"deny host=%.512s/%.20s operation DCC SEND",rladdr,riaddr);
							notice(0,nick,"*** DCC SEND is not permitted");
							goto skip;
						} else 
							dcc_log = (op->flg & OP_LOG);
						
					} else
					if (!strcmp(op->name,"get") && !strncasecmp(dcc_cmd,"send",4) && isclient) {
						if (op->flg & OP_DENY) {
							syslog(LLEV,"deny host=%.512s/%.20s operation DCC GET",rladdr,riaddr);
							notice(0,nick,"*** DCC GET is not permitted");
							goto skip;
						} else
							dcc_log = (op->flg & OP_LOG);
					} else
					if (!strcmp(op->name,"chat") && !strncasecmp(dcc_cmd,"chat",4)) {
						if (op->flg & OP_DENY) {
							syslog(LLEV,"dcc cmd %s",dcc_cmd);
							syslog(LLEV,"deny host=%.512s/%.20s operation DCC CHAT",rladdr,riaddr);
							notice(0,nick,"*** DCC CHAT is not permitted");
							goto skip;
						} else
							dcc_log = (op->flg & OP_LOG);
					} else
					if (!strcmp(op->name,"xmit") && !strncasecmp(dcc_cmd,"xmit",4)) {
						if (op->flg & OP_DENY) {
							syslog(LLEV,"deny host=%.512s/%.20s operation DCC XMIT",rladdr,riaddr);
							notice(0,nick,"*** DCC XMIT is not permitted");
							goto skip;
						} else
							dcc_log = (op->flg & OP_LOG);
					} else
					if (!strcmp(op->name,"unknown") && strncasecmp(dcc_cmd,"send",4) && 
					    strncasecmp(dcc_cmd,"chat",4) && strncasecmp(dcc_cmd,"xmit",4)) {
						if (op->flg & OP_DENY) {
							syslog(LLEV,"deny host=%.512s/%.20s operation DCC %s",rladdr,riaddr,dcc_cmd);
							notice(0,nick,"*** unknown dcc commands are not permitted");
							goto skip;
						} else
							dcc_log = (op->flg & OP_LOG);
					}
				}
				if (dcc_addr && dcc_port) {
					if (dcc_port>MINPORT) {
						strcpy(addr,proxy_dcc(dcc_addr,dcc_port,proxy_ip,dcc_cmd,isclient,dcc_log));
						sprintf(ctcp_out,"\01DCC %.4s %.128s %.512s %d\01",dcc_cmd,dcc_fname,addr,dcc_size);
						strcat(out_string,ctcp_out);
						strcat(out_string,"\r\n");
					} else {
						dcc_addr=htonl(dcc_addr);
						syslog(LLEV,"deny dcc connection to %d.%d.%d.%d restricted port %d direction=%d",
						((char *)&dcc_addr)[0]&0xff,
						((char *)&dcc_addr)[1]&0xff,
						((char *)&dcc_addr)[2]&0xff,
						((char *)&dcc_addr)[3]&0xff,
						dcc_port,isclient);
					}
				} else {
baddcc:						syslog(LLEV,"malformed dcc request, cmd=%s direction=%d",dcc_cmd,isclient);
				}
skip:				;
			} else {
				strcat(out_string,"\01");
				strncat(out_string,ctcp_string,sizeof(out_string)-strlen(out_string)-4);
				strcat(out_string,"\01");
			}
		}
	}

parsed: 
/*	syslog(LLEV,"out: %s",out_string); */
	return(write(to_sock,out_string,strlen(out_string)));
}

iostats generic_tunnel(int sock1, int sock2)
{
	int		maxfd;
	fd_set		rset;
	char		buffer[16384 + 1];
	int		x;
	iostats 	iodata;
	struct timeval	timo;

	iodata.inbytes  = 0;
	iodata.outbytes = 0;

	if(sock1 > sock2)
		maxfd = sock1 + 1;
	else
		maxfd = sock2 + 1;

	while(1)
	{
		FD_ZERO(&rset);
		FD_SET(sock1,&rset);
		FD_SET(sock2,&rset);
		timo.tv_sec = timeout;
		timo.tv_usec = 0;
		if((x = select(maxfd, &rset, NULL, NULL, &timo)) < 0) {
			if (errno != EBADF && errno != EINTR)
				syslog(LLEV,"fwtksyserr: select: %m");
			exit(1);
		} else {
			if(x == 0) {
				syslog(LLEV,"Network timeout signal on dcc tunnel after %d seconds", timeout);
				break;
			}
		}

		if (FD_ISSET(sock1,&rset))
		{
			if((x = read(sock1,buffer,sizeof(buffer))) <= 0)
			{
				break;
			}
			if (write(sock2,buffer,x)!=x) 
				break;
			iodata.outbytes+=x;
		}
		
		if (FD_ISSET(sock2,&rset))
		{
			if((x = read(sock2,buffer,sizeof(buffer))) <= 0)
			{
				break;
			}
			if (write(sock1,buffer,x)!=x)
				break;
			iodata.inbytes+=x;
		}
	}
	close(sock1);
	close(sock2);
	return(iodata);
}


/*      
 * Catch alarm signals and exit.
 */

static void dcc_timeout(int sig)
{
	syslog(LLEV,"Network timeout signal on dcc tunnel after %d seconds", timeout);
	exit(1);
}

static void init_timeout(int sig)
{
	syslog(LLEV, "Network timeout signal after %d seconds",timeout);
	time(&offtime);
	syslog(LLEV, "exit host=%.512s/%.20s dccs=%d in=%d out=%d duration=%d",rladdr,riaddr,0,0,0,(int)offtime - ontime);
	exit(1);
}

static char *proxy_dcc(int destaddr, int destport, ipaddr_t proxy_ip, char *cmd, int isclient, int dcc_log)
{
   	static char				retvalue[512];
	int					listen_sock;
	int					sock;
	int					c_sock;
	struct sockaddr_in			addr;
	struct sockaddr_in			to_addr;
	int					len;
	iostats					iodata;
	static char				sender_riaddr[512];
	static char				sender_rladdr[512];
	static char				receiver_riaddr[512];
	static char				receiver_rladdr[512];
	static	time_t				dccontime;
	static	time_t				dccofftime;

	listen_sock = bind_to_port(0, 1);

	/*
	 * Ignore dead servers so no zombies should be left hanging.
	 */

	signal(SIGCHLD, waitwaitwait);

   	if (fork()) {
		signal(SIGCHLD, waitwaitwait);
       		len = sizeof(to_addr);
		if (getsockname(listen_sock, (struct sockaddr *)&to_addr, &len) < 0)
		{
			syslog(LOG_ERR, "getsockname(): %m");
		   	*retvalue = '\0' ;
			return retvalue;
		}

	       	/*
	       	 * Close the socket as the child does the handling.
	       	 */

	       	close(listen_sock);
	   	sprintf(retvalue,"%lu %u",
				proxy_ip,
				ntohs(to_addr.sin_port)) ;
	   	return retvalue ;
 	}
   
   	/* Give people three minutes to accept the call..
	 */

	signal(SIGALRM, dcc_timeout);
   	alarm(timeout) ;
   
       	/*
       	 * Accept an incoming connection.
       	 */
       	len = sizeof(addr);
       	if ((sock = accept(listen_sock, (struct sockaddr *)&addr, &len)) < 0)
       	{
       		/*
       		 * Connection resets are common enough to log them as debug only.
       		 */
       		syslog((errno == ECONNRESET ? LOG_DEBUG : LOG_ERR), "accept(): %m");
	   	exit (0) ;
       	}
   	close(listen_sock) ;
/*	syslog(LLEV,"accept passed"); */
   
	/*
	 * Ok, here we go!  Disable timer and set the address to connect to.
	 */
	alarm(0) ;
	signal(SIGALRM, SIG_IGN);
	to_addr.sin_family = AF_INET;
	to_addr.sin_port = htons(destport);
	to_addr.sin_addr.s_addr = htonl(destaddr);
	c_sock = socket(AF_INET, SOCK_STREAM, 0);
   
/*	syslog(LLEV,"connecting to %s port %d",inet_ntoa(to_addr.sin_addr),destport); */
	if (connect(c_sock, (struct sockaddr *)&to_addr, sizeof(to_addr)) < 0) {
		syslog(LLEV,"dcc connect failed");
		iodata.inbytes  = 0;
		iodata.outbytes = 0;
	} else {
		if (dcc_log) {
			if (peername(c_sock,sender_rladdr,sender_riaddr,sizeof(sender_riaddr))) {
				syslog(LLEV,"fwtksyserr: cannot get peer name");
				close(sock); 
				close(c_sock);
				exit(1);
			}
			if (peername(sock,receiver_rladdr,receiver_riaddr,sizeof(receiver_riaddr))) {
				syslog(LLEV,"fwtksyserr: cannot get peer name");
				close(sock); 
				close(c_sock);
				exit(1);
			}
			syslog(LLEV,"dcc session connected cmd=%.4s src=%.256s/%.128s dst=%.256s/%.128s",
			       cmd,sender_rladdr,sender_riaddr,receiver_rladdr,receiver_riaddr);
		}
					
/*		syslog(LLEV,"tunneling %d and %d",sock,c_sock); */
		time(&dccontime);
		iodata = generic_tunnel(c_sock,sock);
		time(&dccofftime);
	}
		
       	close(sock);
	close(c_sock);
	if (!isclient) {
		write(iostat_fd[1],&iodata.inbytes,sizeof(int));
		write(iostat_fd[1],&iodata.outbytes,sizeof(int));
	} else {
		write(iostat_fd[1],&iodata.outbytes,sizeof(int));
		write(iostat_fd[1],&iodata.inbytes,sizeof(int));
	}
	syslog(LLEV,"dcc session exit cmd=%.4s src=%.256s/%.128s dst=%.256s/%.128s in=%d out=%d duration=%d",
	       cmd,sender_rladdr,sender_riaddr,receiver_rladdr,receiver_riaddr,iodata.inbytes,iodata.outbytes,
               (int)dccofftime - dccontime);

	close(iostat_fd[0]);
	close(iostat_fd[1]);
       	closelog();
     	exit(0);
}

void server(int s)
 {
	Cfg	*cf;
	int 	sock_c = -1;
	int 	maxfd  = s+1;
	char 	buffer[1024];
	char 	buffer1[1024];
	char 	user[1024];
	char 	nick[100];
	int 	u,n,p;
	fd_set 	rset;
	u_short	myport;
	char 	*cmd, *server, *nck,*pass,*uiname;
	char 	*ulname, *uhost, *userver;
	char	*port = NULL;
	char 	outname[12], outhost[512];
	char 	myserver[1024];
	char 	ufname[64];
	int 	ib = 0;
	int 	ob = 0;
	int	dccs = 0;
	iostats dcc_stat;

	struct   sockaddr_in	sadr;
	int			lsadr;

	u=1;
	n=1;
	p=0;

	if(po)
	  p=1;

	cu++;
	signal(SIGALRM, init_timeout);
	alarm(timeout);

	while((u+n+p)&&3){
		memset(buffer, 0, 1023);
		if(sogets(s, buffer, 1024) <= 0)
		{
			close(s);
			cu--;
			time(&offtime);
			syslog(LLEV,"exit host=%.512s/%.20s dccs=%d in=%d out=%d duration=%d",rladdr,riaddr,dccs,ib,ob,(int)offtime - ontime);
			return;
		}
		if(!strncasecmp(buffer, "USER ", 5)){
			u=0;
			strcpy(user,buffer);
			ulname=NULL;
			strtok(user," ");
			ulname=strtok(NULL," ");
			uhost=strtok(NULL," ");
			userver=strtok(NULL," ");
			uiname=strtok(NULL," \n\r"); 
			strncpy(ufname,uiname,64);

			if (!(*fakeduser==*""))
				strncpy(outname,fakeduser,12);
			else
				strncpy(outname,ulname,12); 

			if (!(*fakedhost==*""))
				strncpy(outhost,fakedhost,512);
			else
				strncpy(outhost,uhost,512);
		}
		else
		if(!strncasecmp(buffer, "NICK ", 5)){
			strcpy(nick,buffer);
			nck=NULL;
			strtok(nick," ");
			nck=strtok(NULL," \n\r");
			if(nck) {
				if (p) 
				  notice(s,nck,"This gateway requires authorization. Please use /QUOTE PASS command.");
			  	n=0;
			}
		}
		else                                       
	        if(!strncasecmp(buffer, "PASS ", 5)){
	                strcpy(myserver,buffer);
	                pass=NULL;
	                strtok(myserver," ");
	                pass=strtok(NULL," \n\r");
	                if(pass) {
	                	if(!strncmp(pass,ps,strlen(ps)))
	                		p=0;
				else {
					syslog(LLEV,"Authorization failed: bad password %.128s",pass);
				}
	                }
	        }
		else
	        if(!strncasecmp(buffer, "QUIT", 4)){
			close(s);
                       	cu--;
			time(&offtime);
			syslog(LLEV,"exit host=%.512s/%.20s dccs=%d in=%d out=%d duration=%d",rladdr,riaddr,dccs,ib,ob,(int)offtime - ontime);
                       	return;
		}
		else
			syslog(LLEV,"unexpected command %s",buffer);
        }

	if((cf = cfg_get("welcome-msg",confp)) != (Cfg *)0) {
		if(cf->argc != 1) {
			syslog(LLEV,"fwtkcfgerr: welcome-msg must have one parameter, line %d",cf->ln);
			exit(1);
		}
		if(showfile(s,nck,cf->argv[0])) {
			syslog(LLEV,"fwtksyserr: cannot display welcome %.512s: %m",cf->argv[0]);
			exit(1);
		}
	} else {
        notice(s,nck,"*** Welcome to IRC Jumperbot- Your in the temperal limbo that requires you connect to a real server"); 
	}

proxy_ready:

	pipe(iostat_fd);
	fcntl(iostat_fd[0],F_SETFL,O_NONBLOCK);

	server=NULL;
	if (*plug_server==*"") 
		if (transparent)
			if (strrchr(userver,'"'))
			/*
 			 *  mirc workaround - thanks to dimon@lim.ru
 			 */
				server=strtok(userver,"\"");
			else 
				server=userver;
		else while (!server) {
			notice(s,nck,"type /quote conn <server> [port] to connect");
			sotimeout(timeout);
			memset(buffer, 0, 1024);
			if (sogets(s, buffer, 1024) <= 0) {
				close(s);
                        	cu--;
				time(&offtime);
				syslog(LLEV,"exit host=%.512s/%.20s dccs=%d in=%d out=%d duration=%d",rladdr,riaddr,dccs,ib,ob,(int)offtime - ontime);
                        	return;
                	}
			cmd = NULL;
                        server = NULL;
                        port = NULL;
                        cmd = strtok(buffer, " ");
                        if(!cmd)
                                continue;
                        if(strcasecmp(cmd, "conn"))
                                continue;
                        server = strtok(NULL, " \n\r");
                        port = strtok(NULL, " \n\r");
		}
	else
		server=plug_server;

	strncpy(myserver,server,512);

	if(!port)         
        	myport=dp;        
        else  
                myport = atoi(port);  

        if(validests != (char **)0) {
        	char    **xp;

                for(xp = validests; *xp != (char *)0; xp++) {
                	if(**xp == '!' && hostmatch(*xp + 1,server)) {
                        	baddest(s,nick,server);
                                goto proxy_ready;
                        } else if(hostmatch(*xp,server))
	                        break;
                 }
                 if(*xp == (char *)0) {
                        baddest(s,nick,server);
                        goto proxy_ready;
                 }
        }
	syslog(LLEV,"permit host=%.512s/%.20s connect to %.512s",rladdr,riaddr,server);
	sprintf(user,"USER %.12s %.256s %.512s %.64s\n",outname,outhost,myserver,ufname);
	sprintf(buffer1,"*** Connecting to port %d of server %.512s",myport,myserver);
	notice(s,nck,buffer1);
	sock_c = do_connect(myserver, myport);
	if(sock_c < 0) {
		sock_c = -1;
		notice(s,nck,"*** Connection failed");
		goto proxy_ready;
	}
	sprintf(buffer1,"*** GATEWAY CONNECTED TO %.512s:%d",myserver,myport);
	notice(s,nck,buffer1);
	sprintf(buffer,"NICK %s\n",nck);
	ob += write(sock_c, buffer, strlen(buffer)); 
	ob += write(sock_c, user, strlen(user)); 

	lsadr = sizeof(sadr);

	if (!cside_ip) {
		getsockname(s,(struct sockaddr *)&sadr,&lsadr);
		cside_ip = ntohl(sadr.sin_addr.s_addr);
	}
	if (!sside_ip) {
		getsockname(sock_c,(struct sockaddr *)&sadr,&lsadr);
		sside_ip = ntohl(sadr.sin_addr.s_addr);
	}

	alarm(0);
	signal(SIGALRM, SIG_IGN);
	sotimeout(DGETS_TIMEOUT);

	while(1)
	{
		FD_ZERO(&rset);
		FD_SET(s, &rset);
		FD_SET(sock_c, &rset);
		if(sock_c > s)
			maxfd = sock_c + 1;
		else
			maxfd = s + 1;
		select(maxfd, &rset, NULL, NULL, NULL);
		memset(buffer, 0, 1024);
		if(FD_ISSET(s, &rset))
		{

			switch(sogets(s, buffer, 1023)) {
			case 0:
				close(s);
				close(sock_c);
				time(&offtime);

				while (read(iostat_fd[0],&dcc_stat,sizeof(iostats)) > 0) {
				ib += dcc_stat.inbytes;
				ob += dcc_stat.outbytes;
				dccs++;
				}
				syslog(LLEV,"exit host=%.512s/%.20s dccs=%d in=%d out=%d duration=%d",rladdr,riaddr,dccs,ib,ob,(int)offtime - ontime);
				return;
			case -1:
				notice(s,nck,"timeout on dgets, continuing"); 
			}

/*
 *
 *  We can put outgoing message stream routines handling here
 *  (like rewriting/blocking CTCP and so on, proxying DCC SEND..)
 *
 */
/*			ob += write(sock_c, buffer, strlen(buffer)); */
			ob += filtered_write(sock_c, buffer, strlen(buffer),sside_ip,nick,0); 
		}
		if(FD_ISSET(sock_c, &rset))
		{
			memset(buffer, 0, 1024);
			switch(sogets(sock_c, buffer, 1023)) {
			case 0:
				close(s);
				close(sock_c);
				cu--;
				time(&offtime);
				while (read(iostat_fd[0],&dcc_stat,sizeof(iostats)) > 0) {
				ib += dcc_stat.inbytes;
				ob += dcc_stat.outbytes;
				dccs++;
				}
				syslog(LLEV,"exit host=%.512s/%.20s dccs=%d in=%d out=%d duration=%d",rladdr,riaddr,dccs,ib,ob,(int)offtime - ontime);
				return;
			case -1:
				notice(s,nck,"timeout on dgets, continuing"); 
			}
/*
 *
 *  We can put incoming message stream handling here
 *  (like responding/blocking CTCP requests and so on, proxying incoming DCC..)
 *
 */
/*			ib += write(s, buffer, strlen(buffer)); */
			ib += filtered_write(s, buffer, strlen(buffer), cside_ip,nick, 1);
		}
	}
}

int
main(int argc, char *argv[])
{
	Cfg		*cf;
        
        mu=100;
        dp=6667;
        cu=0;
        po=0;
#ifndef LOG_DAEMON
	openlog("irc-gw",LOG_PID);
#else
	openlog("irc-gw",LOG_PID|LOG_NDELAY,LFAC);
#endif



	if((confp = cfg_read("irc-gw")) == (Cfg *)-1)
 	  exit(1);

	if ((cf = cfg_get("default-port",confp)) != (Cfg *)0) {
		if(cf->argc != 1) {
			syslog(LLEV,"fwtkcfgerr: default-port must have one parameter, line %d",cf->ln);
			exit(1);
		}
		dp = str_to_port(cf->argv[0]);
	}

	if (peername(0,rladdr,riaddr,sizeof(rladdr))!= 0)
	  {
	    syslog (LLEV, "fwtksyserr: cannot get peer name: %m");
	    exit(1);
	  }

	strcpy(ps,"-NONE-");

	if (!oktotalkto()) {
		if((cf = cfg_get("denial-msg",confp)) != (Cfg *)0) {
			if(cf->argc != 1) {
				syslog(LLEV,"fwtkcfgerr: denial-msg must have one parameter, line %d",cf->ln);
				exit(1);
			}
			if(showfile(0,"*",cf->argv[0])) {
				syslog(LLEV,"fwtksyserr: cannot display denial %.512s:%m",cf->argv[0]);
			}
		}
		exit(1);
	}

	if((cf = cfg_get("timeout",confp)) != (Cfg *)0) {
		if(cf->argc != 1) {
			syslog(LLEV,"fwtkcfgerr: timeout must have one parameter, line %d",cf->ln);
			exit(1);
		}
		if((timeout = atoi(cf->argv[0])) <= 0) {
			syslog(LLEV,"fwtkcfgerr: timeout %.100s invalid, line %d",cf->argv[0],cf->ln);
			exit(1);
		}
	} else
		timeout = PROXY_TIMEOUT;

/*	write(0, buffer, strlen(buffer)); */
	time(&ontime);
	server(0);
	close(iostat_fd[0]);
	close(iostat_fd[1]);
	exit(0);
}

