/*
 * Copyright 2009 Security Technology Research
 * All rights reserved. 
 */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <syslog.h>
#include <sys/time.h>

#ifdef  OpenBSD
#include        <ssl/x509.h>
#include        <ssl/err.h>
#include        <ssl/ssl.h>
#else
#include        <openssl/x509.h>
#include        <openssl/err.h>
#include        <openssl/ssl.h>
#endif

#include "firewall.h"
#include "sslfunc.h"
#include "libemfw.h"
#include "emio.h"

void
emio_sslaccept(emi, emo, rladdr, riaddr, cvrfy, cert, key)
EMI_FILE 	*emi;
EMO_FILE 	*emo;
char 		*rladdr;
char 		*riaddr;
char 		*cvrfy;
X509		*cert;
EVP_PKEY	*key; 
{
	if (emi->fd != emo->fd) {
		syslog(LLEV,"fwtksyserr: internal logic error in emio_sslaccept()");
		exit(1);
	}

	emi->sslio = emo->sslio  = proxy_SSLaccept(emi->fd, rladdr, riaddr, cvrfy, cert, key);

	/* Discard anything left in the buffer */

	if (emi->ptr != emi->end) {
		syslog(LLEV,"securityalert: stray data in input buffer before TLS handshake (accept)");
		emi->ptr = emi->end = emi->buf;
	}

	emo_flush(emo);

	if (!emi->sslio) {
      		emi->flags |= EMI_FLAG_ERROR; 
      		emo->flags |= EMO_FLAG_ERROR; 
      		emi->error = errno;
      		emo->error = errno;
	}

	return;
}	

void
emio_sslconnect(emi, emo, dst, svrfy, cert, key)
EMI_FILE 	*emi;
EMO_FILE 	*emo;
char 		*dst;
char 		*svrfy;
X509		*cert;
EVP_PKEY	*key; 
{
	if (emi->fd != emo->fd) {
		syslog(LLEV,"fwtksyserr: internal logic error in emio_sslaccept()");
		exit(1);
	}

	/* Discard anything left in the buffer */

	if (emi->ptr != emi->end) {
		syslog(LLEV,"securityalert: stray data in input buffer before TLS handshake (connect)");
		emi->ptr = emi->end = emi->buf;
	}

	emo_flush(emo);

	emi->sslio = emo->sslio  = proxy_SSLconnect(emi->fd, dst, svrfy, cert, key);

	if (!emi->sslio) {
      		emi->flags |= EMI_FLAG_ERROR; 
      		emo->flags |= EMO_FLAG_ERROR; 
      		emi->error = errno;
      		emo->error = errno;
	}

	return;
}	

