/*
 * Copyright 2010 Security Technology Research
 * All rights reserved. 
 */

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <syslog.h>
#include <sys/time.h>
#include "firewall.h"
#include "sslfunc.h"
#include "libemfw.h"
#include "emio.h"
#include "firewall2.h"

/*
 * Basic functions to send a file to socket.
 * The "macro" functionality is very simple and designed to
 * work with administrator provided data (error templates etc);
 * on no circumstance should this function ever be used 
 * with user-supplied data stream. Thus loop detection and data
 * sanity checks are *very* basic and never meant to provide adequate
 * protection.
 */

int emo_putfile(EMO_FILE *emo, const char *fname)
{
	char buf[MAX_STR];
	FILE *f = NULL;
	if (!(f = fopen(fname, "r")))
		return(-1);
	while(fgets(buf, sizeof(buf) -1, f) != (char *)0)
		emo_puts (emo, buf);
	fclose(f);
	return(emo_flush(emo));
}

int emo_putfile_withmacros(EMO_FILE *emo, const char *fname, text_macros *m)
{
	char 		buf[MAX_STR +1];
	FILE 		*f = NULL;
	text_macros 	*p;
	char		*c;
	if (!(f = fopen(fname, "r")))
		return(-1);
	while(fgets(buf, sizeof(buf) -1, f) != (char *)0) {
		int	i = 0;
		for (p = m; p->name; p++) {
			char	nbuf[MAX_STR +1];
			if ((c = strstr(buf, p->name))) {
				if ((i++) >= 256) {
					syslog(LLEV, "fwtksyserr: too many macro substitutions in file %.256s", fname);
					exit(1);
				}
				*c = '\0';
				c += strlen(p->name);
				snprintf(nbuf, MAX_STR, "%s%s%s", buf, p->value, c);
				strlcpy(buf,nbuf, MAX_STR);
			}
		}
		emo_puts (emo, buf);
	}
	fclose(f);
	return(emo_flush(emo));
}

