/*
 * Copyright 2010 Security Technology Research
 * All rights reserved. 
 */


/* Please note that at the moment this code checks if the certificate
 * in question *pretends* to be EV according to policies list.
 * It does not perform any actual issuer checks (as we do not even have
 * the issuer certificate handy), nor does it check if it even validates.
 * So the prupose is to impement more *restrictive* policies on certificates
 * pretending to be EV (strict hostname match with no wildacrds), 
 * not permissive ones.
 */

#ifdef	OpenBSD
#include 	<ssl/x509.h>
#include 	<ssl/x509v3.h>
#else
#include 	<openssl/x509.h>
#include 	<openssl/x509v3.h>
#endif

#include <string.h>
struct trustedEVinfo {
  	const char *policyoid;
  	const char *CAname; 
  	const char *EVroot_sha1_fingerprint;
  	const char *issuer;
  	const char *serial;
};


static const struct trustedEVinfo trustedEVlist[] = {
	{
    		"2.16.840.1.114171.500.9",
    		"WellsSecure EV",
    		"E7:B4:F6:9D:61:EC:90:69:DB:7E:90:A7:40:1A:3C:F4:7D:4F:E8:EE",
		"CN=WellsSecure Public Root Certificate Authority,OU=Wells Fargo Bank NA,O=Wells Fargo WellsSecure,C=US",
    		"AQ=="
  	},
  	{
    		"1.2.392.200091.100.721.1",
    		"SECOM EV",
    		"FE:B8:C4:32:DC:F9:76:9A:CE:AE:3D:D8:90:8F:FD:28:86:65:64:7D",
		"OU=Security Communication EV RootCA1,O=\"SECOM Trust Systems CO.,LTD.\",C=JP",
    		"AA=="
  	},
  	{
    		"1.3.6.1.4.1.6334.1.100.1",
    		"Cybertrust EV",
    		"5F:43:E5:B1:BF:F8:78:8C:AC:1C:C7:CA:4A:9A:C6:22:2B:CC:34:C6",
		"CN=Cybertrust Global Root,O=Cybertrust, Inc",
    		"BAAAAAABD4WqLUg="
  	},
  	{
    		"2.16.528.1.1001.1.1.1.12.6.1.1.1",
    		"DigiNotar EV",
    		"C0:60:ED:44:CB:D8:81:BD:0E:F8:6C:0B:A2:87:DD:CF:81:67:47:8C",
		"E=info@diginotar.nl,CN=DigiNotar Root CA,O=DigiNotar,C=NL",
    		"DHbanJEMTiye/hXQWJM8TA=="
  	},
  	{
    		"2.16.756.1.89.1.2.1.1",
    		"SwissSign EV",
    		"D8:C5:38:8A:B7:30:1B:1B:6E:D4:7A:E6:45:25:3A:6F:9F:1A:27:61",
		"CN=SwissSign Gold CA - G2,O=SwissSign AG,C=CH",
    		"ALtAHEP1Xk+w"
  	},
  	{
    		"1.3.6.1.4.1.23223.2",
    		"StartCom EV",
    		"3E:2B:F7:F2:03:1B:96:F3:8C:E6:C4:D8:A8:5D:3E:2D:58:47:6A:0F",
		"CN=StartCom Certification Authority,OU=Secure Digital Certificate Signing,O=StartCom Ltd.,C=IL",
    		"AQ=="
  	},
  	{
    		"2.16.840.1.113733.1.7.23.6",
    		"VeriSign EV",
    		"4E:B6:D5:78:49:9B:1C:CF:5F:58:1E:AD:56:BE:3D:9B:67:44:A5:E5",
		"/C=US/O=VeriSign, Inc./OU=VeriSign Trust Network/OU=Terms of use at https://www.verisign.com/rpa (c)06/CN=VeriSign Class 3 Extended Validation SSL CA",
    		"GNrRniZ96LtKIVjNzGs7Sg=="
  	},
  	{
    		"1.3.6.1.4.1.14370.1.6",
    		"GeoTrust EV",
    		"32:3C:11:8E:1B:F7:B8:B6:52:54:E2:E2:10:0D:D6:02:90:37:F0:96",
		"CN=GeoTrust Primary Certification Authority,O=GeoTrust Inc.,C=US",
    		"GKy1av1pthU6Y2yv2vrEoQ=="
  	},
  	{
    		"2.16.840.1.113733.1.7.48.1",
    		"Thawte EV",
    		"91:C6:D6:EE:3E:8A:C8:63:84:E5:48:C2:99:29:5C:75:6C:81:7B:81",
		"CN=thawte Primary Root CA,OU=\"(c) 2006 thawte, Inc. - For authorized use only\",OU=Certification Services Division,O=\"thawte, Inc.\",C=US",
    		"NE7VVyDV7exJ9C/ON9srbQ=="
  	},
  	{
    		"2.16.840.1.114404.1.1.2.4.1",
    		"Trustwave EV",
    		"B8:01:86:D1:EB:9C:86:A5:41:04:CF:30:54:F3:4C:52:B7:E5:58:C6",
		"CN=XRamp Global Certification Authority,O=XRamp Security Services Inc,OU=www.xrampsecurity.com,C=US",
    		"UJRs7Bjq1ZxN1ZfvdY+grQ=="
  	},
  	{
    		"2.16.840.1.114404.1.1.2.4.1",
    		"Trustwave EV",
    		"87:82:C6:C3:04:35:3B:CF:D2:96:92:D2:59:3E:7D:44:D9:34:FF:11",
		"CN=SecureTrust CA,O=SecureTrust Corporation,C=US",
    		"DPCOXAgWpa1Cf/DrJxhZ0A=="
  	},
  	{
    		"2.16.840.1.114404.1.1.2.4.1",
    		"Trustwave EV",
    		"3A:44:73:5A:E5:81:90:1F:24:86:61:46:1E:3B:9C:C4:5F:F5:3A:1B",
		"CN=Secure Global CA,O=SecureTrust Corporation,C=US",
    		"B1YipOjUiolN9BPI8PjqpQ=="
  	},
  	{
    		"1.3.6.1.4.1.6449.1.2.1.5.1",
    		"Comodo EV",
    		"9F:74:4E:9F:2B:4D:BA:EC:0F:31:2C:50:B6:56:3B:8E:2D:93:C3:11",
		"CN=COMODO ECC Certification Authority,O=COMODO CA Limited,L=Salford,ST=Greater Manchester,C=GB",
    		"H0evqmIAcFBUTAGem2OZKg=="
  	},
  	{
    		"1.3.6.1.4.1.6449.1.2.1.5.1",
    		"Comodo EV",
    		"66:31:BF:9E:F7:4F:9E:B6:C9:D5:A6:0C:BA:6A:BE:D1:F7:BD:EF:7B",
		"CN=COMODO Certification Authority,O=COMODO CA Limited,L=Salford,ST=Greater Manchester,C=GB",
    		"ToEtioJl4AsC7j41AkblPQ=="
  	},
  	{
    		"1.3.6.1.4.1.6449.1.2.1.5.1",
    		"Comodo EV",
    		"02:FA:F3:E2:91:43:54:68:60:78:57:69:4D:F5:E4:5B:68:85:18:68",
		"CN=AddTrust External CA Root,OU=AddTrust External TTP Network,O=AddTrust AB,C=SE",
    		"AQ=="
  	},
  	{
    		"1.3.6.1.4.1.6449.1.2.1.5.1",
    		"Comodo EV",
    		"58:11:9F:0E:12:82:87:EA:50:FD:D9:87:45:6F:4F:78:DC:FA:D6:D4",
		"CN=UTN - DATACorp SGC,OU=http://www.usertrust.com,O=The USERTRUST Network,L=Salt Lake City,ST=UT,C=US",
    		"RL4Mi1AAIbQR0ypoBqmtaQ=="
  	},
  	{
    		"1.3.6.1.4.1.6449.1.2.1.5.1",
    		"Comodo EV",
    		"04:83:ED:33:99:AC:36:08:05:87:22:ED:BC:5E:46:00:E3:BE:F9:D7",
		"CN=UTN-USERFirst-Hardware,OU=http://www.usertrust.com,O=The USERTRUST Network,L=Salt Lake City,ST=UT,C=US",
    		"RL4Mi1AAJLQR0zYq/mUK/Q=="
  	},
  	{
    		"2.16.840.1.114413.1.7.23.3",
    		"Go Daddy EV OID a",
    		"27:96:BA:E6:3F:18:01:E2:77:26:1B:A0:D7:77:70:02:8F:20:EE:E4",
		"OU=Go Daddy Class 2 Certification Authority,O=\"The Go Daddy Group, Inc.\",C=US",
    		"AA=="
 	},
  	{
    		"2.16.840.1.114413.1.7.23.3",
    		"Go Daddy EV OID a",
    		"31:7A:2A:D0:7F:2B:33:5E:F5:A1:C3:4E:4B:57:E8:B7:D8:F1:FC:A6",
		"E=info@valicert.com,CN=http://www.valicert.com/,OU=ValiCert Class 2 Policy Validation Authority,O=\"ValiCert, Inc.\",L=ValiCert Validation Network",
    		"AQ=="
  	},
  	{
    		"2.16.840.1.114414.1.7.23.3",
    		"Go Daddy EV OID b",
    		"AD:7E:1C:28:B0:64:EF:8F:60:03:40:20:14:C3:D0:E3:37:0E:B5:8A",
		"OU=Starfield Class 2 Certification Authority,O=\"Starfield Technologies, Inc.\",C=US",
    		"AA=="
  	},
  	{
    		"2.16.840.1.114412.2.1",
    		"DigiCert EV",
    		"5F:B7:EE:06:33:E2:59:DB:AD:0C:4C:9A:E6:D3:8F:1A:61:C7:DC:25",
		"CN=DigiCert High Assurance EV Root CA,OU=www.digicert.com,O=DigiCert Inc,C=US",
    		"AqxcJmoLQJuPC3nyrkYldw=="
  	},
  	{
    		"1.3.6.1.4.1.8024.0.2.100.1.2",
    		"Quo Vadis EV",
    		"CA:3A:FB:CF:12:40:36:4B:44:B2:16:20:88:80:48:39:19:93:7C:F7",
		"CN=QuoVadis Root CA 2,O=QuoVadis Limited,C=BM",
    		"BQk="
  	},
  	{
    		"1.3.6.1.4.1.782.1.2.1.8.1",
    		"Network Solutions EV",
    		"74:F8:A3:C3:EF:E7:B3:90:06:4B:83:90:3C:21:64:60:20:E5:DF:CE",
		"CN=Network Solutions Certificate Authority,O=Network Solutions L.L.C.,C=US",
    		"V8szb8JcFuZHFhfjkDFo4A=="
  	},
  	{
    		"2.16.840.1.114028.10.1.2",
    		"Entrust EV",
    		"B3:1E:B1:B7:40:E3:6C:84:02:DA:DC:37:D4:4D:F5:D4:67:49:52:F9",
		"CN=Entrust Root Certification Authority,OU=\"(c) 2006 Entrust, Inc.\",OU=www.entrust.net/CPS is incorporated by reference,O=\"Entrust, Inc.\",C=US",
    		"RWtQVA=="
  	},
  	{
    		"1.3.6.1.4.1.4146.1.1",
    		"GlobalSign EV",
    		"B1:BC:96:8B:D4:F4:9D:62:2A:A8:9A:81:F2:15:01:52:A4:1D:82:9C",
		"CN=GlobalSign Root CA,OU=Root CA,O=GlobalSign nv-sa,C=BE",
    		"BAAAAAABFUtaw5Q="
  	},
  	{
    		"1.3.6.1.4.1.4146.1.1",
    		"GlobalSign EV",
    		"75:E0:AB:B6:13:85:12:27:1C:04:F8:5F:DD:DE:38:E4:B7:24:2E:FE",
		"CN=GlobalSign,O=GlobalSign,OU=GlobalSign Root CA - R2",
    		"BAAAAAABD4Ym5g0="
  	},
  	{
    		"2.16.578.1.26.1.3.3",
    		"Buypass Class 3 CA 1", 
    		"61:57:3A:11:DF:0E:D8:7E:D5:92:65:22:EA:D0:56:D7:44:B3:23:71",
		"CN=Buypass Class 3 CA 1,O=Buypass AS-983163327,C=NO",
    		"Ag=="
  	},

  	{
    		"",
    		"",
		"00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00:00", 
    		"",
		""	
  	}
};

int
isEVcert(X509 *mycert)
{
	POLICYINFO		*policy;
	STACK_OF(POLICYINFO)	*cert_policies;
	int 			i;

	if( !(cert_policies = (STACK_OF(POLICYINFO) *)X509_get_ext_d2i( mycert, 
			NID_certificate_policies, NULL, NULL)) ||
	    !(sk_POLICYINFO_num(cert_policies)) )
		return(0);

	for (i = 0; i < sk_POLICYINFO_num(cert_policies) ; i++) {
		char	policybuf[64] = "";
		int	j;
		if (!(policy = sk_POLICYINFO_value(cert_policies, i)))
			continue;
		OBJ_obj2txt(policybuf, sizeof(policybuf), policy->policyid, 1);
		for (j = 0; *(trustedEVlist[j].policyoid) ; j++) {
			if (!strcmp(policybuf,trustedEVlist[j].policyoid) 
			/* We cannot enforce this check at the moment as i do not
  			   have accurate collection of EV issuer NAME_onelines */
  			/*						   &&
			    !strcmp(X509_NAME_oneline(X509_get_issuer_name(mycert),
						      NULL, 1024),
				    trustedEVlist[j].issuer) */) {
				sk_POLICYINFO_free(cert_policies);
				return(1);
			}
		}

	}
	return(0);
	sk_POLICYINFO_free(cert_policies);
	
}

#ifdef TEST

#include <stdio.h>
#ifdef	OpenBSD
#include 	<ssl/pem.h>
#else
#include 	<openssl/pem.h>
#endif

int main(int argc, char *argv[])
{
	FILE	*f;
	X509	*mycert = NULL;
	int	isEV;

	if (argc != 3) {
		printf("Syntax: %s {e|n} certfile.pem\n", argv[0]);
		exit(1);
	}
	if ((f = fopen(argv[2], "r")) && 
	    (mycert = PEM_read_X509(f, NULL, NULL, NULL))) {
		isEV = isEVcert(mycert);
		fclose(f);
		printf("%s is%s EV certificate\n", argv[2], isEV ? "" : " NOT");
		exit ((*argv[1] == 'e') ^ isEV);
	}
	printf("Cannot read certificate %s\n", argv[2]);
	exit(1);
}
#endif



