/*
 * Copyright (c) 2002-2009, OpenFWTK Development Group
 * All rights reserved. See LICENSE.
 */

/*
 * OpenFWTK SSL wrapper functions
 *
 * (C) Copyright 2002 ArkanoiD <ark@eltex.ru>
 *
 */

#include	<stdio.h>
#include	<string.h>
#ifdef SOLARIS
#include	<strings.h>
#endif
#include	<syslog.h>
#include	<sys/types.h>
#include	<sys/ioctl.h>
#include	<sys/time.h>
#include	<sys/socket.h>
#include	<unistd.h>
#include	<stdlib.h>
#include	<netdb.h>
#include	<netinet/in.h>
#include	<errno.h>
#include	<time.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<fcntl.h>
#include	<ctype.h>

#ifdef	OpenBSD
#include 	<ssl/x509.h>
#include 	<ssl/x509v3.h>
#include 	<ssl/err.h>
#include 	<ssl/ssl.h>
#else
#include 	<openssl/x509.h>
#include 	<openssl/x509v3.h>
#include 	<openssl/err.h>
#include 	<openssl/ssl.h>
#include 	<openssl/rand.h>
#include 	<openssl/bn.h>
#endif

#include	"firewall.h"
#include	"auth.h"

#include	"fwfunc.h"
#include	"firewall2.h"


SSL_CTX			*proxy_SSLCTX = NULL;
STACK_OF(X509_NAME) 	*proxy_SSLcertnames = NULL;
EVP_PKEY		*proxy_SSLkey = NULL;
char			*proxy_SSLcert_cache_path = NULL;
char			*proxy_SSLcert_fp_path = NULL;
char			*proxy_SSLciphers = NULL;

/*
 * Hostname match better (not fully, as it does not always make sense ;-) 
 * conforming to Saint-Andre & Hodges tls-server-id draft
 * http://tools.ietf.org/html/draft-saintandre-tls-server-id-check-03
 */

extern 	int		hostmatch_tls(const char *,const char *, int);

/*
 * See notes inside evchk.c - this function does not check if the certificate
 * *is* EV, it just checks if it pretends to be one.
 */

extern	int 		isEVcert(X509 *);

/*
 *	NID copy macro. Note it requires idx to be defined and available.
 */

#define COPY_NID(f,t,x) if ((idx = X509_NAME_get_index_by_NID(f,x,-1)) != -1)\
	X509_NAME_add_entry(t, X509_NAME_get_entry(f, idx), -1, -1)

/* Handles, callbacks and functions for dynamically loaded 
 * Carillon Pathfinder library
 */

#ifdef PATHFINDER_SHLIB

#include	<dlfcn.h>

static 	void		*pflib_handle = NULL;
static	int		(*pathfinder_cb)(X509_STORE_CTX *, void*) = NULL;
static	int		(*pathfinder_dbus)(const char*, const char *,
					   const int, const int,
					   char **) = NULL;

#define POLICYOID_ANY "2.5.29.32.0"

char			*proxy_pf_policy = POLICYOID_ANY;


		
/* IMPORTANT: buf should be at least 512 bytes long */

static	int
pathfinder_verify(cert, buf)
X509	*cert;
char	*buf;
{
	const char 	*hex = "0123456789ABCDEF";
	size_t		size;
	unsigned char 	*keybuf, *iend, *cp;
	char 		*certdata_str, *certdata_str_i;
	char		*errmsg;

	size = i2d_X509(cert, NULL);
	iend = keybuf = malloc(size);
	certdata_str = malloc(size * 2 + 1);
	i2d_X509(cert, &iend);
	cp = keybuf;
	certdata_str_i = certdata_str;
	while (cp < iend) {
		unsigned char ch = *cp++;
		*certdata_str_i++ = hex[( ch >> 4 ) & 0xf];
		*certdata_str_i++ = hex[ch & 0xf];
	}
	*certdata_str_i = 0;
	free(keybuf);

	if (!((*pathfinder_dbus)(certdata_str, proxy_pf_policy,
			        strcmp(proxy_pf_policy, POLICYOID_ANY) ? 1 : 0,
			        0, &errmsg))) {
		if (buf)
			strlcpy(buf, errmsg, 512 - 1);
		free(errmsg);	
		free(certdata_str);	
		return(-1);
	} else {
		free(errmsg);	
		free(certdata_str);
		return(X509_V_OK);
	}
}

/*

static	int
pathfinder_verify_cert_cb(ctx, dummy)
X509_STORE_CTX	*ctx;
void		*dummy;
{
	return (pathfinder_verify(ctx->cert, NULL) == X509_V_OK);

}

*/

/*	
 *	Parse pathfinder error message to determine if there was a path building
 *	error which may indicate self-signed certificates or unkown issuers;
 *	since OpenSSL already provides more detailed diagnostics, there is
 *	no reason to override it with less informative pathfinder failure.
 *	Skip "Pathfinder certificate validation failed: ", this gives 42.
 *	Ugly, but we do not have a better method yet.
 */

static int
pathfinder_may_ignore(buf)
char 	*buf;
{
	if (!strncmp("Couldn't build path.", buf, 20)) {
		return(1);
	}
	return(0);
}

#endif

int
proxy_SSLcheck_weak(ssl)
SSL		*ssl;
{
	STACK_OF(X509) *sk;
	int i;

	/* 	TODO:
   		SSL_get_current_cipher(ssl) and check if it is weak;
		check for less common digest/cipher combos
	*/

	if (!(sk=SSL_get_peer_cert_chain(ssl)))
		return(-1);

	for (i=0; i<sk_X509_num(sk); i++) {
		int 	alg = OBJ_obj2nid(sk_X509_value(sk,i)->sig_alg->algorithm);
		if (alg == NID_md5WithRSAEncryption || alg == NID_md2WithRSAEncryption)
			return(1);
	}
	return(0);
}

/* Certificate fingerprint helper functions 	*/

#define HASHSIZE 20				/* We use SHA1, 160bit 	*/
#define HASHSIZE_HEX (HASHSIZE * 3) - 1		/* aa:bb:cc format	*/

static void
tohex(raw,hex,len)
char    *raw;
char    *hex; 
int     len;
{
        char *hdig="0123456789ABCDEF";

        while(len > 0){
                len--;
                *hex++ = hdig[((*raw)>>4)&0xf];
                *hex++ = hdig[(*raw++)&0xf];
		*hex++ = ':';
        }
        *(--hex) = '\0';
        return;
}

int
proxy_SSLcheck_cert_fingerprint(cert, host, buf)
const	X509	*cert;
const	char	*host;
	char	*buf;
{
	char	filename[MAXPATHLEN + 1];
	char	cert_fp[HASHSIZE + 1];
	char	cert_hex[HASHSIZE_HEX + 1];
	char	known_hex[HASHSIZE_HEX + 1];
	int	f;
	unsigned int hashsize;

	if (!proxy_SSLcert_fp_path)
		return(X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT);
	
	if (strchr(host, '/') || (strlen(host) > (MAXPATHLEN - 32))) {
		syslog(LLEV,"securityalert: malformed host name %.512s (may be truncated)", host);
		if (buf)
			snprintf(buf, 512 - 1, "Malformed host name (may be truncated): %s", host);
		return(X509_V_ERR_APPLICATION_VERIFICATION);
	}

	if (!X509_digest(cert, EVP_sha1(), (unsigned char*) cert_fp, &hashsize) ||
	    hashsize != HASHSIZE) {
		syslog(LLEV,"fwtksyserr: cannot compute certificate fingerprint");
		exit(1);
	}

	tohex(cert_fp, cert_hex, HASHSIZE);

	snprintf(filename, MAXPATHLEN - 1, "%s/%s.sha1", 
			 	 proxy_SSLcert_cache_path, host);

	if ((f = open(filename, O_RDONLY)) != -1) {
		if ((read(f, known_hex, HASHSIZE_HEX) != HASHSIZE_HEX) ||
		     close(f)) {
			syslog(LLEV,"fwtksyserr: cannot read certificate fingerprint %.128s:%.128s",
					filename, strerror(errno));
			exit(1);
		}
		if (memcmp(cert_hex, known_hex, HASHSIZE_HEX)) {
			if (buf)
				snprintf(buf, 512 - 1, "REMOTE HOST IDENTIFICATION FOR %s HAS CHANGED!\n"
					"Known certificate fingerprint (sha1): %s\n"
			   		"Remote host fingerprint:              %s",
					host, known_hex, cert_hex);
			return(X509_V_ERR_CERT_SIGNATURE_FAILURE); 
		}
	} else if (!(((f = open(filename, O_CREAT|O_TRUNC|O_WRONLY, 0644)) != -1) &&
		   (write(f, cert_hex, HASHSIZE_HEX) == HASHSIZE_HEX) &&
		   !close(f))) {
		syslog(LLEV,"fwtksyserr: cannot read certificate fingerprint %.128s:%.128s",
				filename, strerror(errno));
		exit(1);
	} 

	if (buf)
		snprintf(buf, 512 - 1, "SSL certificate does not verify, %s",
			X509_verify_cert_error_string(X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT));
	return(X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT);
	
}

int
proxy_SSLcheck_cert(ssl,host,vrfy,buf)
SSL		*ssl;
const char	*host;
const char	*vrfy;
char		*buf;
{
	X509		*peer;
	X509_NAME 	*peersubj;
	int		err;
        int     	dnsname_found = 0;
        int     	i, r;
        STACK_OF(GENERAL_NAME) *gens;
	X509_NAME_ENTRY *n;

	if (buf)
		*buf = '\0';

	if (!strcasecmp(vrfy,"none"))
		return(X509_V_OK);

	if (((err = SSL_get_verify_result(ssl)) != X509_V_OK) &&
	    !(proxy_SSLcert_fp_path &&
		(err == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT))
#ifdef PATHFINDER_SHLIB
	/*  Pathfinder may override these errors */
	    && !(pathfinder_dbus && 
		 ((err == X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN) ||
		  (err == X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY) ||
		  (err == X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE) ||
		  (err == X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT)))
#endif
	) {
		if (buf)
			snprintf(buf, 512 - 1, "SSL certificate does not verify, %s",
				X509_verify_cert_error_string(err));
		syslog (LLEV,"SSL certificate does not verify, %s",
			X509_verify_cert_error_string(err));
		return(err);
	}

	if (!(peer = SSL_get_peer_certificate(ssl))) {
		if (buf)
			snprintf(buf, 512 - 1, "Unable to get peer certificate");
		syslog(LLEV,"Unable to get peer certificate");
		SSL_set_verify_result(ssl, X509_V_ERR_APPLICATION_VERIFICATION);
		return(-1);
	}
	if (proxy_SSLcert_fp_path &&
	    (err == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT)) 
			return(proxy_SSLcheck_cert_fingerprint(peer, host, buf));

	
	if (!(peersubj = X509_get_subject_name(peer))) {
		if (buf)
			snprintf(buf, 512 - 1, "Unable to obtain peer certificate subject");
		syslog(LLEV,"Unable to obtain peer certificate subject");
		SSL_set_verify_result(ssl, X509_V_ERR_APPLICATION_VERIFICATION);
		return(-1);
	}

	/* rfc3280 states that "Certificate users MUST be able to
	   handle serialNumber values up to 20 octets.  Conformant CAs MUST NOT
	   use serialNumber values longer than 20 octets." */

	if (ASN1_STRING_length((ASN1_STRING*)X509_get_serialNumber(peer)) > 20) {
		if (buf)
			snprintf(buf, 512 - 1, "Certificate serial number too long");
		syslog(LLEV,"securityalert: certificate serial number too long");
		SSL_set_verify_result(ssl, X509_V_ERR_APPLICATION_VERIFICATION);
		return(-1);
	}

#ifdef PATHFINDER_SHLIB
	if (pathfinder_dbus) {
		char	pbuf[512 + 1];
		if (pathfinder_verify(peer, pbuf) == X509_V_OK) 
			SSL_set_verify_result(ssl, X509_V_OK);
		else {
			/*	We may ignore certain errors if the chain does not verify anyways,
			 	but we should not, if it does but pathfinder returns error - this may
		    	 	indicate someone messing with CRLs or other nasty things.  */
			 
			if ((err != X509_V_OK) && pathfinder_may_ignore(pbuf)) {
				if (buf) 
					snprintf(buf, 512 - 1, "SSL certificate does not verify, %s and Pathfinder cannot build path either (%s).",
						 X509_verify_cert_error_string(err), pbuf);
				syslog (LLEV, "%s", buf); 
				return(err);
			}
			if (buf) {
				snprintf(buf, 512 - 1, "Pathfinder certificate validation failed: %s", pbuf);
				syslog(LLEV, "%s", buf);
			}
			SSL_set_verify_result(ssl, X509_V_ERR_APPLICATION_VERIFICATION);
			return(-1);
		}
	}
#endif

	if (strcasecmp(vrfy,"checkaddr"))
		return(X509_V_OK);

	/*
         * Verify the name(s) in the peer certificate against the peer
         * hostname. Log peer hostname/certificate mis-matches. If a match is
         * required but fails, bail out with a verification error.
         */

	gens = X509_get_ext_d2i(peer, NID_subject_alt_name, 0, 0);
	if (gens) {
		if ((r = sk_GENERAL_NAME_num(gens)) == 0) {
			if (buf)
				snprintf(buf, 512 - 1, "Malformed altname extension");
			SSL_set_verify_result(ssl, X509_V_ERR_APPLICATION_VERIFICATION);
			return (-1);
		}
		for (i = 0, r; i < r; ++i) {
			const GENERAL_NAME *gn = sk_GENERAL_NAME_value(gens, i);
			if (gn->type == GEN_DNS) {
				if (strlen((char*) ASN1_STRING_data(gn->d.ia5)) !=
				    ASN1_STRING_length(gn->d.ia5)) {
					syslog(LLEV,"securityalert: certificate contains embedded NULL in altname DNS extension");
					if (buf)
						snprintf(buf, 512 - 1, "Malformed altname extension");
					SSL_set_verify_result(ssl, X509_V_ERR_APPLICATION_VERIFICATION);
					return (-1);
				}
				if (isspace(*((char*) gn->d.ia5->data))) {
					if (buf)
						snprintf(buf, 512 - 1, "Malformed altname extension");
					SSL_set_verify_result(ssl, X509_V_ERR_APPLICATION_VERIFICATION);
					return (-1);
				}
				dnsname_found++;
				if (hostmatch_tls((char*) gn->d.ia5->data, host,
						  isEVcert(peer))) {
					sk_GENERAL_NAME_free(gens);
					return(X509_V_OK);
				}
			}
		}
		sk_GENERAL_NAME_free(gens);
	}

	/* If there *is* DNS in altnames, at least one should match host name */

	if (dnsname_found) {
		syslog(LLEV,"Peer provided altname extension, but none of DNS records match the host name");
		if (buf)
			snprintf(buf, 512 - 1, "Peer provided altname extension, but none of DNS records match the host name");
		return(-1);
	}

	if (((i = X509_NAME_get_index_by_NID(peersubj,NID_commonName, -1)) == -1) ||
	    ((n = X509_NAME_get_entry(peersubj ,i)) == NULL)) {
		if (buf)
			snprintf(buf, 512 - 1, "CN not found in peer certificate");
		syslog(LLEV,"CN not found in peer certificate");
		SSL_set_verify_result(ssl, X509_V_ERR_APPLICATION_VERIFICATION);
		return(-1);
	}
	if (strlen((char*) ASN1_STRING_data(n->value)) != ASN1_STRING_length(n->value)) {
		syslog(LLEV,"securityalert: certificate contains embedded NULL in CN");
		if (buf)
			snprintf(buf, 512 - 1, "Malformed CN");
		SSL_set_verify_result(ssl, X509_V_ERR_APPLICATION_VERIFICATION);
		return (-1);
	}
	if (hostmatch_tls((char*) ASN1_STRING_data(n->value), host, 
			  isEVcert(peer)))
		return(X509_V_OK);
	SSL_set_verify_result(ssl, X509_V_ERR_APPLICATION_VERIFICATION);
	if (buf)
		snprintf(buf, 512 - 1, "Host name mismatch (certificate issued for %s != %s)", (char*) ASN1_STRING_data(n->value), host);
	return(-1);
}

X509*
proxy_SSLread_cert(certfile)
const char *certfile;
{
	FILE	*f;
	X509	*mycert;

	if (!(f = fopen(certfile, "r"))) {
		syslog(LLEV,"fwtksyserr: cannot open certificate file %.128s:%.128s",
			certfile, strerror(errno));
			exit(1);
	}
	if (!(mycert = PEM_read_X509(f, NULL, NULL, NULL))) {
		syslog(LLEV,"fwtksyserr: cannot read certificate file %.128s:%.128s",
			certfile,ERR_error_string(ERR_get_error(), NULL));
		exit(1);
	}
	fclose(f);
	return(mycert);

}


EVP_PKEY*
proxy_SSLread_key(keyfile)
const 	char	*keyfile;
{
	FILE		*f;
	EVP_PKEY	*mykey;

	if (!(f = fopen(keyfile, "r"))) {
		syslog(LLEV,"fwtksyserr: cannot open private key file %.128s:%.128s",
			keyfile, strerror(errno));
			exit(1);
	}
	if (!(mykey = PEM_read_PrivateKey(f, NULL, NULL, NULL))) {
		syslog(LLEV,"fwtksyserr: cannot read private key file %.128s:%.128s",
			keyfile,ERR_error_string(ERR_get_error(), NULL));
		exit(1);
	}
	fclose(f);
	return(mykey);
}

X509*
proxy_SSLconf_cert(sslcertname)
const char *sslcertname;
{
	char	*sslcertfile;
	sslcertfile = proxy_conf_string(proxy_confp, sslcertname);
	if (!sslcertfile) {
		syslog(LLEV,"fwtkcfgerr: certifcate file not specified");
		exit(1);
	}
	return(proxy_SSLread_cert(sslcertfile));
}

EVP_PKEY*
proxy_SSLconf_key(keyname)
const 	char	*keyname;
{
	char	*keyfile;
	keyfile = proxy_conf_string(proxy_confp, keyname);
	if (!keyfile) {
		syslog(LLEV,"fwtkcfgerr: key file %.128s not specified", keyname);
		exit(1);
	}
	return(proxy_SSLread_key(keyfile));
}

/*
 *	Common SSL init and configuration functions are called here.
 *	Verify location priority (only one used at time) is:
 *	1) Explicitly configured as trustedCA configuration parameter;
 *	2) Carillon Pathfinder (if available)
 *	3) Default OpenSSL system-wide verify locations.
 */


void
proxy_SSLinit()
{
	char		*sslcertfile = NULL;
	char		*vrfyfile = NULL;
#ifdef PATHFINDER_SHLIB
	char		*pferr;
	char		*pfmode;

	if (!(pfmode = proxy_conf_string(proxy_confp,"pathfinder")))
		pfmode = "default";

	if ((strcmp(pfmode,"disable")) && 
	    (pflib_handle || (pflib_handle = dlopen(PATHFINDER_SHLIB, RTLD_NOW)))) {
		dlerror();
		pathfinder_cb = dlsym(pflib_handle, "openssl_verify_cb");
		pathfinder_dbus = dlsym(pflib_handle, "pathfinder_dbus_verify");
		if ((pferr = dlerror())) {
			syslog(LLEV,"fwtksyserr: cannot load pathfiner runtime %s, %s", PATHFINDER_SHLIB, pferr);
			exit(1);
		}
	}
	if ((!strcmp(pfmode,"enable")) && (!pflib_handle)) {
		syslog(LLEV,"fwtksyserr: cannot load pathfiner runtime %s", PATHFINDER_SHLIB);
		exit(1);
	}
	proxy_pf_policy = proxy_conf_string(proxy_confp,"policyOID");
	if (!proxy_pf_policy)
		proxy_pf_policy = POLICYOID_ANY;
#endif

	sslcertfile = proxy_conf_string(proxy_confp,"certificate");
	vrfyfile = proxy_conf_string(proxy_confp,"trustedCA");
	proxy_SSLcert_cache_path = proxy_conf_string(proxy_confp, "certificate-cache");
	proxy_SSLcert_fp_path = proxy_conf_string(proxy_confp, "certificate-fingerprints");
	proxy_SSLciphers = proxy_conf_string(proxy_confp, "ciphers");

	OPENSSL_config(NULL);
	SSL_library_init();
	SSL_load_error_strings();
	ERR_load_BIO_strings();
	ERR_load_crypto_strings();
	OpenSSL_add_all_algorithms();

	if (!sslcertfile) {
		syslog(LLEV,"fwtkcfgerr: certifcate file not specified");
		exit(1);
	}
	proxy_SSLkey = proxy_SSLconf_key("private-key");
	if (!(proxy_SSLCTX = SSL_CTX_new(SSLv23_method()))) {
		syslog(LLEV,"fwtksyserr: cannot create SSL context, %s",
				ERR_error_string(ERR_get_error(), NULL));
		exit(1);
	}
	if (!(SSL_CTX_use_certificate_chain_file(proxy_SSLCTX, sslcertfile))) {
		syslog(LLEV,"fwtksyserr: cannot read certificate chain file %s:%s",
		sslcertfile,ERR_error_string(ERR_get_error(), NULL));
		exit(1);
	}
	if (!(SSL_CTX_use_PrivateKey(proxy_SSLCTX, proxy_SSLkey)) ||
	    !(SSL_CTX_check_private_key(proxy_SSLCTX))) {
		syslog(LLEV,"fwtksyserr: error loading private key %s:%s",
				proxy_conf_string(proxy_confp, "private-key"),
				ERR_error_string(ERR_get_error(), NULL));
		exit(1);
	}

	if (vrfyfile) {
		if (!(SSL_CTX_load_verify_locations(proxy_SSLCTX,
						    vrfyfile,NULL))) {
			syslog(LLEV,"fwtksyserr: error setting verify locations %.128s:%.128s",
					vrfyfile,
					ERR_error_string(ERR_get_error(), NULL));
			exit(1);
		}
		if (!(proxy_SSLcertnames =SSL_load_client_CA_file(vrfyfile))) {
			syslog(LLEV,"fwtksyserr: error loading CA verify file %s:%s",
					vrfyfile,
					ERR_error_string(ERR_get_error(), NULL));
			exit(1);
		}
		SSL_CTX_set_client_CA_list(proxy_SSLCTX, proxy_SSLcertnames);
	} else {

#ifdef PATHFINDER_SHLIB
/*	Even if we have VERIFY_NONE, the callback overrides internal
	verification result. So i disable it for now. No application uses it anyways yet.
	This code is likey to be replaced with SSL_set_verify callbacks.
*/ 
/*
		if (pathfinder_cb) 
			SSL_CTX_set_cert_verify_callback(proxy_SSLCTX, pathfinder_verify_cert_cb, NULL);
		else
*/
#endif

		if (!SSL_CTX_set_default_verify_paths(proxy_SSLCTX)) {
			syslog(LLEV,"fwtksyserr: error setting default certificate verify path: %.128s", ERR_error_string(ERR_get_error(), NULL));
			exit(1);
		}
	}

	if (!proxy_SSLciphers)
		proxy_SSLciphers = "HIGH:-MD5";

	if (!SSL_CTX_set_cipher_list(proxy_SSLCTX, proxy_SSLciphers)) {
		syslog(LLEV,"fwtksyserr: error selecting cipher list %s: %.128s", 
			proxy_SSLciphers, ERR_error_string(ERR_get_error(), NULL));
		exit(1);
	}
	if (!cfg_get("sslv2", proxy_confp))
		SSL_CTX_set_options(proxy_SSLCTX, SSL_OP_NO_SSLv2);
}

/* 
 *	Free SSL structures to allow subsequent re-init
 */

void
proxy_SSLdeinit()
{
	if (proxy_SSLkey)
		EVP_PKEY_free(proxy_SSLkey);
	if (proxy_SSLCTX)
		SSL_CTX_free(proxy_SSLCTX);
	if (proxy_SSLcertnames)
		sk_X509_NAME_free(proxy_SSLcertnames);
	EVP_cleanup();
	ERR_free_strings();
#ifdef PATHFINDER_SHLIB
	if (pflib_handle) {
		dlclose(pflib_handle);
		pflib_handle = NULL;
		pathfinder_cb = NULL;
		pathfinder_dbus = NULL;
	}
#endif
}


SSL*
proxy_SSLaccept(int fd, char *rladdr, char *riaddr, char *cvrfy, X509 *cert, EVP_PKEY *key)
{
	SSL	*cconn;
	cconn = SSL_new(proxy_SSLCTX);
	SSL_set_fd(cconn,fd);
	SSL_set_client_CA_list(cconn,proxy_SSLcertnames);  

	if (cert && (SSL_use_certificate(cconn, cert) != 1)) {
		syslog(LLEV,"fwtksyserr: SSL_use_certificate(): %.128s", 
			ERR_error_string(ERR_get_error(), NULL));
		exit(1);
	}
	if (key && (SSL_use_PrivateKey(cconn, key) != 1)) {
		syslog(LLEV,"fwtksyserr: SSL_use_PrivateKey(): %.128s", 
			ERR_error_string(ERR_get_error(), NULL));
		exit(1);
	}

	if (strcasecmp(cvrfy,"none")) 
		SSL_set_verify(cconn,SSL_VERIFY_PEER,NULL);
	else
		SSL_set_verify(cconn,SSL_VERIFY_NONE,NULL);

	if (SSL_accept(cconn) == -1) {
		syslog(LLEV,"Error accepting SSL connection: %.128s", 
				ERR_error_string(ERR_get_error(), NULL));
		SSL_free(cconn);
		return(NULL);
	}

	if (proxy_SSLcheck_cert(cconn,riaddr,cvrfy, (char*) NULL)) {
		syslog(LLEV,"deny host=%s/%s certificate does not match host identification",
		rladdr,riaddr);
		SSL_shutdown(cconn);
		SSL_free(cconn);
		return(NULL);
	}
	return(cconn);
}
	

SSL*
proxy_SSLconnect(int fd, char *dst, char *svrfy, X509 *cert, EVP_PKEY *key)
{
	SSL	*sconn;
	sconn = SSL_new(proxy_SSLCTX);

	if (cert && (SSL_use_certificate(sconn, cert) != 1)) {
		syslog(LLEV,"fwtksyserr: SSL_use_certificate(): %.128s", 
			ERR_error_string(ERR_get_error(), NULL));
		exit(1);
	}
	if (key && (SSL_use_PrivateKey(sconn, key) != 1)) {
		syslog(LLEV,"fwtksyserr: SSL_use_PrivateKey(): %.128s", 
			ERR_error_string(ERR_get_error(), NULL));
		exit(1);
	}

	SSL_set_fd(sconn,fd);

	if (!strcasecmp(svrfy,"none"))
		SSL_set_verify(sconn,SSL_VERIFY_NONE,NULL);
	else 
		SSL_set_verify(sconn,SSL_VERIFY_PEER, NULL);

#ifdef SSL_CTRL_SET_TLSEXT_HOSTNAME
	SSL_set_tlsext_host_name(sconn, dst);
#endif

	if (SSL_connect(sconn) == -1) {
		syslog(LLEV,"Error initiating SSL connection: %s", 
			ERR_error_string(ERR_get_error(), NULL));
		SSL_free(sconn);
		return(NULL);
	}

	if (proxy_SSLcheck_cert(sconn,dst,svrfy, (char*) NULL)) {
		syslog(LLEV,"deny host=%s/%s destination=%s certificate does not match destination identification",
		       proxy_stats.rladdr,proxy_stats.riaddr,dst);
		SSL_shutdown(sconn);
		SSL_free(sconn);
		return(NULL);
	}
	return(sconn);
}

int
proxy_SSLread(SSL *conn, char *buf, int buflen)
{
	int	x, e;
	if (((x = SSL_read(conn,buf,buflen)) > 0)) {
		return(x);
	}
	switch (e = SSL_get_error(conn,x)) {
	case SSL_ERROR_NONE:
	case SSL_ERROR_ZERO_RETURN:
	case SSL_ERROR_WANT_WRITE:
	case SSL_ERROR_WANT_READ:
	case SSL_ERROR_WANT_X509_LOOKUP:
		return(-1);
	case SSL_ERROR_SYSCALL:
		if ((x == 0) && ((errno == 0) || (errno == ECONNRESET)))
			return(-1);
	case SSL_ERROR_SSL:
		syslog(LLEV,"SSL read error %d: %.128s, system error %.128s", e,
				ERR_error_string(ERR_get_error(), NULL), 
				strerror(errno));
		return(-1);
	default:
		return(-1);
	}
}

int
proxy_SSLwrite(SSL *conn, const char *buf, int buflen)
{
	unsigned int 	nbytes = 0;
	int		x, e;

	while (nbytes < buflen) {
		x = SSL_write(conn,buf,buflen-nbytes);
		if (x <= 0) {
			switch (e = SSL_get_error(conn,x)) {
			case SSL_ERROR_NONE:
			case SSL_ERROR_ZERO_RETURN:
				return (nbytes ? nbytes : x);
			case SSL_ERROR_WANT_WRITE:
			case SSL_ERROR_WANT_READ:
			case SSL_ERROR_WANT_X509_LOOKUP:
				continue;
			case SSL_ERROR_SYSCALL:
			case SSL_ERROR_SSL:
				syslog(LLEV,"SSL write error %d: %.128s,"
					    " system error %.128s", e,
				       ERR_error_string(ERR_get_error(), NULL), 
				       strerror(errno));
				return (nbytes ? nbytes : x);
			}
		}
		buf += x;
		nbytes += x;
	}
	return nbytes;
}

static int
add_extstr_by_NID(X509 *cert, X509V3_CTX* ctx, int NID, char *value)
{
	X509_EXTENSION	*ext = NULL;
	ext = X509V3_EXT_conf_nid(NULL, ctx, NID, value);
	X509_add_ext(cert, ext, -1); 
	X509_EXTENSION_free(ext);
	return(0);
}

X509*
proxy_SSLspoof_server_cert(srvsession)
SSL		*srvsession;
{
	X509		*peercert, *newcert, *CAcert;
	X509_NAME	*srvname;
	X509_NAME	*issuer;
	EVP_PKEY	*signkey;
	BIGNUM		*serialBN = NULL;
	char		*cakeyname;
	int		loc;
	int		rc;
	X509_EXTENSION	*altnames = NULL;
	unsigned char	buf[1024];
	int		idx;
	X509V3_CTX 	v3ctx; 
	char		filename[MAXPATHLEN + 1] = "";

	cakeyname = proxy_conf_string(proxy_confp,"builtinCA-key");

	signkey = cakeyname ? proxy_SSLread_key(cakeyname) : proxy_SSLkey;

	peercert = SSL_get_peer_certificate(srvsession);
	srvname = X509_get_subject_name(peercert);

	serialBN = ASN1_INTEGER_to_BN(X509_get_serialNumber(peercert),NULL);

	/*
	 *	Check for a certificate cache hit. If we miss, we keep the filename
	 *	for future use. Check for valid certificates only to prevent cache
	 *	poisoning (we could handle that different way, but we won't).
	 *	No malformed filenames check, but directory traversal
	 *	prevention seems reasonable.
	 *
	 *	The filenames are human readable and the format is PEM.
	 *	DER might be slightly faster, but not enough for us to care.
	 */

	if (((rc = SSL_get_verify_result(srvsession)) == X509_V_OK) &&
	    proxy_SSLcert_cache_path) {
		char	*serial;
		FILE	*f;

		/*	We do not cache malformed certificates	*/

		if ((X509_NAME_get_text_by_NID(srvname, NID_commonName, 
					       (char*) buf, sizeof(buf)) != -1) &&
		    (serial = BN_bn2hex(serialBN)) &&
		    !strchr((char*) buf,'/')) {
			snprintf(filename, MAXPATHLEN - 1, "%s/%s.%s.pem", 
			 	 proxy_SSLcert_cache_path, buf, serial);
			OPENSSL_free(serial);
			if ((f = fopen(filename,"r"))) {
				if (!(newcert = PEM_read_X509(f, NULL, NULL, NULL))) {
					syslog(LLEV,"fwtksyserr: cannot read cached certificate %.128s:%.128s",
						filename, ERR_error_string(ERR_get_error(), NULL));
					exit(1);
				} 
				fclose(f);
				return(newcert);
			}
		}		
	}

	newcert = X509_new();
	X509_set_version(newcert, 0x2);
	X509_set_subject_name(newcert, srvname);

	CAcert = proxy_SSLconf_cert("builtinCA");
	X509V3_set_ctx(&v3ctx, CAcert , newcert, NULL, NULL, 0);

/*
 *	X509_gmtime_adj(X509_get_notBefore(newcert), -365);
 *	X509_gmtime_adj(X509_get_notAfter(newcert), (long)60*60*24*365);
 */
/*	Better just keep it
 */
	X509_set_notBefore(newcert, X509_get_notBefore(peercert));
	X509_set_notAfter(newcert, X509_get_notAfter(peercert));
	X509_set_pubkey(newcert, proxy_SSLkey);

/*	We do that once per connection. If certificate validation
 *	succeeds we just discard the modified serial, so this may
 *	be optimized a bit.
 */

/*
 *	Do a simple transformation to ensure mobile users will get
 *	different serials while browsing directly or via proxy
 */
	BN_add_word(serialBN, (BN_ULONG) 0x10000000UL);


	switch (rc) {
	case X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT:
	/* Self-sign self-signed			*/
		issuer = X509_get_subject_name(peercert);
		signkey = proxy_SSLkey;
		break;
	case X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN:
	case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY:
	case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT:
	/* Dummy issuer for unverifiable certificates, 	
	   preserve other informational fields 		*/
	
		issuer = X509_NAME_new();
		COPY_NID(srvname,issuer,NID_localityName);
		COPY_NID(srvname,issuer,NID_stateOrProvinceName);
		COPY_NID(srvname,issuer,NID_organizationalUnitName);
		COPY_NID(srvname,issuer,NID_organizationName);
		COPY_NID(srvname,issuer,NID_countryName);

		if (X509_NAME_get_text_by_NID(X509_get_issuer_name(peercert), NID_commonName, 
						(char*) buf, sizeof(buf)) != -1) {
			unsigned char newissuer[1024];
			snprintf((char*) newissuer,sizeof(newissuer)-1,"UNVERIFIED: %s", buf);
			X509_NAME_add_entry_by_txt(issuer, "CN", MBSTRING_ASC, newissuer, -1, -1, 0); 
			snprintf((char*) buf, sizeof(buf), "WARNING: unable to verify certificate chain: %s", X509_verify_cert_error_string(rc));
			add_extstr_by_NID(newcert, &v3ctx, NID_netscape_comment, (char*) buf);
		}
		break;
	case X509_V_ERR_OUT_OF_MEM:
	/* Too bad, won't continue			*/
		syslog(LLEV,"fwtksyserr: out of memory while checking server certificate");
		exit(1);
	case X509_V_OK:
		BN_pseudo_rand(serialBN, 64, -1, 0);
	case X509_V_ERR_CERT_NOT_YET_VALID:
	case X509_V_ERR_CERT_HAS_EXPIRED:
	/* Should we handle it any special way?		*/
		issuer = X509_get_subject_name(CAcert);
		break;
	}

	X509_set_issuer_name(newcert, issuer);
	BN_to_ASN1_INTEGER(serialBN, X509_get_serialNumber(newcert));
	BN_free(serialBN);

	add_extstr_by_NID(newcert, &v3ctx, NID_basic_constraints, "critical,CA:FALSE");
	add_extstr_by_NID(newcert, &v3ctx, NID_key_usage, "nonRepudiation,digitalSignature,keyEncipherment");

	if (((loc = X509_get_ext_by_NID(peercert, NID_subject_alt_name, -1)) >= 0) &&
	    (altnames = X509_get_ext(peercert, loc))) {
		X509_add_ext(newcert, altnames, -1);
		X509_EXTENSION_free(altnames);
	}

	if (((loc = X509_get_ext_by_NID(peercert, NID_issuer_alt_name, -1)) >= 0)  &&
	    (altnames = X509_get_ext(peercert, loc))) {
		X509_add_ext(newcert, altnames, -1);
		X509_EXTENSION_free(altnames);
	}

	add_extstr_by_NID(newcert, &v3ctx, NID_subject_key_identifier, "hash");
	add_extstr_by_NID(newcert, &v3ctx, NID_authority_key_identifier, "keyid,issuer:always");
	add_extstr_by_NID(newcert, &v3ctx, NID_ext_key_usage, "1.3.6.1.5.5.7.3.1");



	add_extstr_by_NID(newcert, &v3ctx, NID_netscape_cert_type, "server");

	if (X509_NAME_get_text_by_NID(srvname, NID_commonName, 
					(char*) buf, sizeof(buf)) != -1) 
		add_extstr_by_NID(newcert, &v3ctx, NID_netscape_ssl_server_name, (char*) buf); 
	X509_sign(newcert, signkey, EVP_sha1());
/*
	SSL_CTX_add_extra_chain_cert(proxy_SSLCTX,read_certificate_file(cacertname));
*/
	if ((rc == X509_V_OK) && *filename) {
		FILE	*f;
		if (!(f = fopen(filename,"w")) || !PEM_write_X509(f, newcert)) {
			syslog(LLEV,"fwtksyserr: cannot write certificate file %.128s: %128s %.128s",
				filename, ERR_error_string(ERR_get_error(), NULL), strerror(errno));
			exit(1);
		}
		fclose(f);
	}
	return(newcert);

}


X509*
proxy_SSL_makeup_error_cert(const char* srvname)
{
	X509		*newcert, *CAcert;
	X509_NAME	*subject, *issuer;
	unsigned int	serial;
	EVP_PKEY	*signkey;
	char		*cakeyname;
	int		idx;

	cakeyname = proxy_conf_string(proxy_confp,"builtinCA-key");
	signkey = cakeyname ? proxy_SSLread_key(cakeyname) : proxy_SSLkey;

	newcert = X509_new();

	X509_set_version(newcert, 0x2);
	X509_gmtime_adj(X509_get_notBefore(newcert),-365);
	X509_gmtime_adj(X509_get_notAfter(newcert), (long)60*60*24*30);
	X509_set_pubkey(newcert, proxy_SSLkey);
	subject = X509_NAME_new();
	issuer = X509_get_subject_name(CAcert = proxy_SSLconf_cert("builtinCA"));
	COPY_NID(issuer,subject,NID_localityName);
	COPY_NID(issuer,subject,NID_stateOrProvinceName);
	COPY_NID(issuer,subject,NID_organizationalUnitName);
/*	X509_NAME_add_entry_by_txt(name, "OU", MBSTRING_ASC, "Error page service", -1, -1, 0); */
	COPY_NID(issuer,subject,NID_organizationName);
	COPY_NID(issuer,subject,NID_countryName);
	X509_NAME_add_entry_by_txt(subject, "CN", MBSTRING_ASC, (unsigned char*) srvname, -1, -1, 0);
	X509_set_issuer_name(newcert, issuer);
	X509_set_subject_name(newcert, subject);
	RAND_bytes((unsigned char*)&serial, sizeof(serial));
   	ASN1_INTEGER_set(X509_get_serialNumber(newcert), serial);
	X509_sign(newcert, signkey, EVP_sha1());
	return(newcert);
}

