/*
 * Copyright 2010 Security Technology Research
 * All rights reserved. 
 */

#include <string.h>
#include <syslog.h>
#ifdef SOLARIS
#include <strings.h>
#endif
#include "firewall.h"

int
hostmatch_tls(const char *pattern, const char *name, const int isEV)
{
	const char*	ppos;
	const char*   npos;

	ppos = pattern;
	npos = name;

	if (!isEV && !strncmp(pattern, "*.", 2)) {
		ppos += 2;
		if (!index(ppos,'.')) {
#ifndef TEST
			syslog(LLEV,"securityalert: 2LD is not allowed in wildcards: %.200s", pattern);
#endif
			return(0);
		}
		if (!(npos = strchr(name, '.')) || !(*(++npos))) {
#ifndef TEST
			syslog(LLEV,"non-fqdn name: %.200s", name);
			return(0);
#endif
		}
	}
	return(!strcasecmp(ppos, npos));
}

#ifdef TEST

#include <stdio.h>
#include <stdlib.h>

int
main() {
	printf ("Testing hostmatch_tls()..\n");
	if ((hostmatch_tls("*.com","paypal.com", 0) != 0) ||	
	    (hostmatch_tls("*","paypal.com", 0) != 0) ||	
	    (hostmatch_tls("*.mydomain.com","mydomain.com", 0) != 0) ||	
	    (hostmatch_tls("*.mydomain.com","foo.bar.mydomain.com", 0) != 0) ||
	    (hostmatch_tls("*.mydomain.com","*.*.mydomain.com", 0) != 0) ||
	    (hostmatch_tls("foo*.mydomain.com","foobar.mydomain.com", 0) != 0) ||
	    (hostmatch_tls("foo.*.mydomain.com","foo.bar.mydomain.com", 0)!= 0) ||
	    (hostmatch_tls("*.mydomain.com","foo.mydomain.com", 1) != 0) ||
	    (hostmatch_tls("*.mydomain.com","foo.mydomain.com", 0) == 0)) {
		printf("failed.\n");
		exit(1);
	} else {
		printf("passed.\n");
		exit(0);
	}
}
#endif
	
