/*
 * Copyright 2011 Security Technology Research
 * All rights reserved. 
 */

#include <unistd.h>
#include <string.h>
#include <strings.h>
#include <limits.h>
#include <inttypes.h>
#include <errno.h>
#include <syslog.h>
#include <time.h>

#include "firewall.h"
#include "fwfunc.h"
#include "libemfw.h"
#include "firewall2.h"
#include "emio.h"
#include "headers.h"
#include "http.h"
#include "message.h"

/*
 *      Parse rfc2822 address simple way and get address itself.
 *	It sucks and ignores all complicated stuff, use with care.
 */
/*
 *	Warning: this function returns static buffer.
 */

char*
message_addr_strip_brackets(const char *address)
{
	char	*p, *p1;
	static	char	mybuf[MAXLINE];

	memset(mybuf,0,sizeof(mybuf));
	
	if (!address || !(*address))
		return(NULL);

	strncpy(mybuf,address,sizeof(mybuf)-1);

	if (!(p=index(mybuf,'<')))
		return(mybuf);

	if (!(*(++p)) || !(p1=rindex(p,'>')))
		return(NULL);
	*(p1) = '\0';
	return(p);
}

/* 
 *	0 OK, 1 size exceeded, 2 input error, 3 output error
 *	We need to treat EOF and exceeded size differently from http.
 */ 

int
message_copy(EMO_FILE *dst, EMI_FILE *src, long long size)
{
	char	buffer[MAXLINE * 10];
	int	n;

	if (size == 0) 
		return(1);
	if (size == -1LL)
		size = MAX_CONTENT_LENGTH;
	
	while (size > 0) {
		char	*p;
                n = emi_gets (src, buffer, sizeof(buffer));
                if ((n < 0) || emi_error(src)) 
			return(2);

		if (!strcmp(buffer,"."))
			return(0);
		if (n > size - 2)
			break;
		p = buffer;
		while(*p) {
			emo_write(dst, p, MIN(MAXLINE, n));
			emo_puts(dst, "\r\n");
			if (emo_error(dst))
				return(3);
			size -= MIN(MAXLINE, n) + 2;
			p += MIN(MAXLINE,n);
			n -= MAXLINE;
		}
	}
	return(1);
}

char*
message_savebody(char *template, EMI_FILE *src, long long size, int *result)
{
	char		*qf;
	int		tfd;
	EMO_FILE	*tqf;
	int		rc;

	qf = xstrdup(template);

	if (((tfd = mkstemp(qf)) == -1) || !(tqf = emo_fdopen(tfd, 4096))) {
		syslog(LLEV,"fwtksyserr: cannot create temporary file, %s",
			strerror(errno));
		exit(1);
	}

	size = (size == -1LL) ? MAX_CONTENT_LENGTH : size;
	rc = message_copy(tqf , src, size);
	emo_close(tqf);
	if (result)
		*result = rc;
	if (rc) {
		unlink(qf);
		return(NULL);
	}
	return(qf);
}

