/* testcap.c */

/* Copyright 2009 Security Technology Research */


#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <syslog.h>
#include "firewall.h"
#include "libemfw.h"
#include "firewall2.h"
#include "emio.h"
#include "ctype.h"
#include "headers.h"
#include "http.h"
#include "url.h"
#include "icap.h"
#define SELECT_CTYPE_HTTP
#include "tables.h"
#include "ci_milter.h"
#include "filters.h"

#define MAXFILTERS 64

#define HTTP_CHAR_TYPE(c,x) (http_char_type[(unsigned char)c] & (x))

#define LOALPHA(c)	HTTP_CHAR_TYPE (c, CTYPE_LOALPHA)
#define UPALPHA(c)	HTTP_CHAR_TYPE (c, CTYPE_UPALPHA)
#define ALPHA(c)	HTTP_CHAR_TYPE (c, CTYPE_ALPHA)
#define DIGIT(c)	HTTP_CHAR_TYPE (c, CTYPE_DIGIT)
#define WHITE(c)	HTTP_CHAR_TYPE (c, CTYPE_WHITE)

extern const unsigned char http_char_type[UCHAR_MAX+1];

#define MAX_HDR_SIZE             (128 * 1024)

/* These variables must be global. */

static void usage (void) ATTR_NORETURN;

static void usage (void)
{
  fputs ("Usage: testfilters {-reqmod|respmod} [-encode] [-headers <headerfile>] -body <bodyfile> [-request \"<http-request>\"] -filters <filter> [<filter..]\n", stderr);
  exit (1);
}

void quit (int rc)
{
  	exit (rc);
}

static int recursion = 0;

void vsyslog (int prio, const char *fmt, va_list arg_ptr)
{
  	if (recursion++ != 0)
    		abort ();
  	vfprintf (stderr, fmt, arg_ptr);
  	fprintf (stderr, "\r\n");
  	--recursion;
}

void syslog (int prio, const char *fmt, ...)
{
  	va_list arg_ptr;

  	va_start (arg_ptr, fmt);
  	vsyslog (prio, fmt, arg_ptr);
  	va_end (arg_ptr);
}

/* In case all the syslog()-related functions are in one module... */

void openlog (const char *ident, int logopt, int facility)
{
}

void closelog (void)
{
}

int setlogmask (int maskpri)
{
  	return 0;
}



int main (int argc, char *argv[])
{

  	int 		i;
  	icap_req 	reqtype;
	char		*name_headers 	= NULL;
	char		*name_body 	= NULL;
	char		*request	= NULL;
	char		*filters[MAXFILTERS + 1];
	struct app_data	adata;
	EMI_FILE	*headers_in	= NULL;
	int64_t		bodysize =  -1LL;
	char		*errmsg;
	int		rc;

  	if (argc < 2)
    		usage ();

	adata.encode = 0;
	adata.headers = NULL;
	filters[0] = NULL;

	for (i=1 ; i < argc; i++) {
  		if (!strcmp (argv[i], "-reqmod")) {
      			reqtype = REQMOD; 
    		} else if (!strcmp (argv[i], "-respmod")) {
      			reqtype = RESPMOD;
		} else if (!strcmp (argv[i], "-encode")) {
			adata.encode = 1;
    		} else if ((!strcmp (argv[i], "-headers")) && (argc - i >= 1)) {
			name_headers = xstrdup(argv[++i]);
    		} else if ((!strcmp (argv[i], "-body")) && (argc - i >= 1)) {
			name_body = xstrdup(argv[++i]);
    		} else if ((!strcmp (argv[i], "-request")) && (argc - i >= 1)) {
			request = xstrdup(argv[++i]);
    		} else if ((!strcmp (argv[i], "-filters")) && (argc - i >= 1)) {
			int	j;
			for (j = 1; ((i + j < argc) && (j < MAXFILTERS) && 
				     (argv[i+j][0] != '-')) ; j++)
				     filters[j-1] = xstrdup(argv[i+j]);
			filters[j] = NULL;
			i += j;
		}
	}

  	if (!name_body || !filters[0])
    		usage ();

	sprintf(proxy_stats.riaddr, "192.168.1.1");

	if (name_headers) {
		adata.smtp_sender = xstrdup("ark@eltex.net");
		adata.smtp_rcpts = xmalloc(sizeof(char*)*2);
		adata.smtp_rcpts[0] = xstrdup("somebody@nowhere.net");
		adata.smtp_rcpts[1] = NULL;
		adata.target_host = proxy_stats.dst;

		if (!(headers_in = emi_open(name_headers, 4096))) {
			perror("cannot open headers file");
			exit(1);
		}
			
		adata.headers = header_read(headers_in, NULL);
		emi_close(headers_in);
	}

	if (!request)
		request = xstrdup("GET / HTTP/1.0");

	rc = filter_inspect_object(name_body, &bodysize, request, reqtype,
			           filters, adata.headers ? &adata : NULL, 
				   &errmsg);

	printf("filter completed with code %x\n", rc);
	exit(0);
}
