/* testcap.c */

/* Copyright 2009 Security Technology Research */


#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <syslog.h>
#include "firewall.h"
#include "libemfw.h"
#include "firewall2.h"
#include "emio.h"
#include "ctype.h"
#include "headers.h"
#include "http.h"
#include "url.h"
#include "icap.h"
#define SELECT_CTYPE_HTTP
#include "tables.h"

#define HTTP_CHAR_TYPE(c,x) (http_char_type[(unsigned char)c] & (x))

#define LOALPHA(c)	HTTP_CHAR_TYPE (c, CTYPE_LOALPHA)
#define UPALPHA(c)	HTTP_CHAR_TYPE (c, CTYPE_UPALPHA)
#define ALPHA(c)	HTTP_CHAR_TYPE (c, CTYPE_ALPHA)
#define DIGIT(c)	HTTP_CHAR_TYPE (c, CTYPE_DIGIT)
#define WHITE(c)	HTTP_CHAR_TYPE (c, CTYPE_WHITE)

extern const unsigned char http_char_type[UCHAR_MAX+1];

#define MAX_HDR_SIZE             (128 * 1024)

/* These variables must be global. */

EMI_FILE *c_in, *s_in;		/* Input from client, from server */
EMO_FILE *c_out, *s_out;	/* Output to client, to server */

static void usage (void) ATTR_NORETURN;

static void usage (void)
{
  fputs ("Usage: testicap -reqmod [-has-body] -icap-in <infile> -icap-out <outfile>\n", stderr);
  exit (1);
}

void quit (int rc)
{
  	exit (rc);
}

static int recursion = 0;

void vsyslog (int prio, const char *fmt, va_list arg_ptr)
{
  	if (recursion++ != 0)
    		abort ();
  	emo_flush (s_out); emo_flush (c_out);
  	emo_vprintf (s_out, fmt, arg_ptr);
  	emo_write (s_out, "\r\n", 2);
  	emo_flush (s_out); emo_flush (c_out);
  	--recursion;
}

void syslog (int prio, const char *fmt, ...)
{
  	va_list arg_ptr;

  	va_start (arg_ptr, fmt);
  	vsyslog (prio, fmt, arg_ptr);
  	va_end (arg_ptr);
}

/* In case all the syslog()-related functions are in one module... */

void openlog (const char *ident, int logopt, int facility)
{
}

void closelog (void)
{
}

int setlogmask (int maskpri)
{
  	return 0;
}



int main (int argc, char *argv[])
{

  	int 		i;
  	icap_req 	reqtype;
	char		*name_icapin 	= NULL;
	char		*name_icapout 	= NULL;
	EMI_FILE	*icap_in	= NULL;
	EMO_FILE	*icap_out	= NULL;
	struct hdr_line *req_hdr	= NULL;
	struct hdr_line *res_hdr	= NULL;
	struct hdr_line *icap_hdr	= NULL;
	char		*cachefile	= NULL;
	int    		has_body = 0;
	int 		bodysize = -1;
	int		rc;

  	i = 1;
  	if (argc < 2)
    		usage ();
  	if (!strcmp (argv[i], "-reqmod")) {
      		reqtype = REQMOD; ++i;
    	}
  	if (!strcmp (argv[i], "-respmod")) {
      		reqtype = RESPMOD; ++i;
    	}
  	if (!strcmp (argv[i], "-has-body")) {
      		has_body = 1; ++i;
    	}
	if ((!strcmp (argv[i], "-icap-in")) && (argc - i >= 1)) {
		i++;
		name_icapin = xstrdup(argv[i++]);
	}
	if ((!strcmp (argv[i], "-icap-out")) && (argc - i >= 1)) {
		i++;
		name_icapout = xstrdup(argv[i++]);
	}

  	if (i != argc || !name_icapin || !name_icapout)
    		usage ();

	sprintf(proxy_stats.riaddr, "192.168.1.1");

  	if (reqtype == REQMOD) {
      		c_in = emi_fdopen (0, 4096);
      		c_out = emo_fdopen (3, 4096);
      		s_in = NULL;
      		s_out = emo_fdopen (1, 4096);
      		icap_in = emi_open(name_icapin, 4096 + 1024);
      		icap_out = emo_open(name_icapout, 4096 + 1024);

      		if (!c_in || !c_out || !s_out || !icap_in || !icap_out) {
	  		perror ("emi_fdopen()");
	  		exit (2);
		}

      		req_hdr = header_read (c_in, NULL);
		if (has_body)
			bodysize = atoi(header_field_value(header_field_by_name(req_hdr,"Content-length"),NULL));

		switch (rc = icap_request(icap_in, icap_out, "icap://localhost/test",  
					  REQMOD, req_hdr, NULL , has_body ?  c_in : NULL, 0, bodysize, &cachefile, 4096, &icap_hdr)) {
		case 204:
			header_write(s_out, req_hdr);
			emo_printf(s_out,"\r\n");
			if (has_body) { 
				if (cachefile) 
					c_in = emi_open(cachefile, 4096);
				http_copy (s_out, 0, c_in, 0, -1, OUT);
			}
			break;
		case 200:
			if (icap_is_response(icap_hdr))
				emo_printf(s_out,"**this is repsonse sent to client**\r\n");
      			req_hdr = header_read (icap_in, NULL);
			header_write(s_out, req_hdr);
			emo_printf(s_out,"\r\n");
			if (has_body || (icap_is_response(icap_hdr)))
				http_copy (s_out, 0, icap_in, 1, -1, OUT);
			break;
		default:
			syslog(LLEV,"icap returned code %d", rc);
			exit(0);
		}
	} else {
      		c_in = NULL;
      		c_out = emo_fdopen (1, 4096);
      		s_in = emi_fdopen (0, 4096);;
      		s_out = c_out;
      		icap_in = emi_open(name_icapin, 4096 + 1024);
      		icap_out = emo_open(name_icapout, 4096 + 1024);

      		if (!s_in || !c_out || !s_out || !icap_in || !icap_out) {
	  		perror ("emi_fdopen()");
	  		exit (2);
		}
		
		header_add_line (&req_hdr, "GET /origin-resource HTTP/1.1");
		header_add_line (&req_hdr, "Host: www.origin-server.com");
		header_add_line (&req_hdr, "Accept: text/html, text/plain, image/gif");
		header_add_line (&req_hdr, "Accept-Encoding: gzip, compress");

      		res_hdr = header_read (s_in, NULL);
		if (has_body && header_field_by_name(req_hdr,"Content-length"))
			bodysize = atoi(header_field_value(header_field_by_name(req_hdr,"Content-length"),NULL));
		switch (rc = icap_request(icap_in, icap_out, "icap://localhost/test",  
					  RESPMOD, req_hdr, res_hdr , has_body ?  s_in : NULL, 0, -1, &cachefile, 4096, &icap_hdr)) {
		case 204:
			http_write_header(c_out, req_hdr);
			emo_printf(c_out,"\r\n");
			if (has_body) { 
				if (cachefile) 
					s_in = emi_open(cachefile, 4096);
				http_copy (c_out, 0, s_in, 0, -1, IN);
			}
			break;
		case 200:
      			req_hdr = header_read (icap_in, NULL);
			http_write_header(c_out, req_hdr);
			emo_printf(c_out,"\r\n");
			if (has_body)
				http_copy (c_out, 0, icap_in, 1, -1, OUT);
			break;
		default:
			syslog(LLEV,"icap returned code %d", rc);
			exit(0);
		}
	}

  /* Don't close the files as we might want to write to file
     descriptor 2! */

  	if (emo_flush (c_out) != 0 || 
	    (s_out != c_out && emo_flush (s_out) != 0) ||
	    (emo_flush(icap_out))) {
      		perror ("emo_flush()");
      		exit (2);
    	}
  	return 0;
}
