/* testmessage.c */

/* Copyright 2011 Security Technology Research */


#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <syslog.h>
#include "firewall.h"
#include "libemfw.h"
#include "firewall2.h"
#include "emio.h"
#include "ctype.h"
#include "headers.h"
#include "http.h"
#include "message.h"

/* These variables must be global. */

EMI_FILE *c_in, *s_in;		/* Input from client, from server */
EMO_FILE *c_out, *s_out;	/* Output to client, to server */

static void usage (void) ATTR_NORETURN;

static void usage (void)
{
  fputs ("Usage: testmessage [-size maxsize] [-drop headername] [-insert headerstring] [-add headerstring]\n", stderr);
  exit (1);
}

void quit (int rc)
{
  	exit (rc);
}

static int recursion = 0;

void vsyslog (int prio, const char *fmt, va_list arg_ptr)
{
  	if (recursion++ != 0)
    		abort ();
  	emo_flush (s_out); emo_flush (c_out);
  	emo_vprintf (s_out, fmt, arg_ptr);
  	emo_write (s_out, "\r\n", 2);
  	emo_flush (s_out); emo_flush (c_out);
  	--recursion;
}

void syslog (int prio, const char *fmt, ...)
{
  	va_list arg_ptr;

  	va_start (arg_ptr, fmt);
  	vsyslog (prio, fmt, arg_ptr);
  	va_end (arg_ptr);
}

/* In case all the syslog()-related functions are in one module... */

void openlog (const char *ident, int logopt, int facility)
{
}

void closelog (void)
{
}

int setlogmask (int maskpri)
{
  	return 0;
}



int main (int argc, char *argv[])
{

  	int 		i;
	long long 	maxsize = -1LL;
	struct hdr_line	*h;
	char		*dropfield = NULL;
	char		*insertstring = NULL;
	char		*addstring = NULL;

  	i = 1;
	for (i = 1; i < argc ; i++) {
		if ((!strcmp (argv[i], "-size")) && (argc - i >= 1)) 
			maxsize = atoi(argv[++i]);
		else if ((!strcmp (argv[i], "-drop")) && (argc - i >= 1)) 
			dropfield = xstrdup(argv[++i]);
		else if ((!strcmp (argv[i], "-insert")) && (argc - i >= 1)) 
			insertstring = xstrdup(argv[++i]);
		else if ((!strcmp (argv[i], "-add")) && (argc - i >= 1)) 
			addstring = xstrdup(argv[++i]);
		else {
			fprintf(stderr,"argc %d argv %s",argc, argv[1]);
			usage ();
		}
	} 

	sprintf(proxy_stats.riaddr, "192.168.1.1");

      	c_in = emi_fdopen (0, 4096);
      	c_out = emo_fdopen (3, 4096);
      	s_in = NULL;
      	s_out = emo_fdopen (1, 4096);

	if (!c_in || !c_out || !s_out) {
	  	perror ("emi_fdopen()");
	  	exit (2);
	}

	h = header_read(c_in, "testmessage");
	if (addstring)
		header_add_line(&h, addstring);
	if (insertstring)
		header_insert_line(&h, insertstring);
	if (dropfield)
		header_drop(&h, header_field_by_name(h, dropfield));
	emo_printf(s_out,"---HEADER---\r\n");
	header_write(s_out, h);
	emo_printf(s_out,"---BODY---\r\n");
	switch(message_copy(s_out, c_in, maxsize)) {
	case 0:	break;
	case 1: emo_printf(s_out, "Size exceeded\r\n");
		break;
	case 2: emo_printf(s_out, "Premature EOF\r\n");
		break;
	case 3: emo_printf(s_out, "Output error!\r\n");
		break;
	default:emo_printf(s_out, "Internal logic error!\r\n");
	}	

  /* Don't close the files as we might want to write to file
     descriptor 2! */

  	if (emo_flush (c_out) != 0 || 
	    (s_out != c_out && emo_flush (s_out) != 0)) {
      		perror ("emo_flush()");
      		exit (2);
    	}
  	return 0;
}
