/*
 * Copyright (c) 1999-2007, OpenFWTK Development Group
 * All rights reserved. See LICENSE.
 */


#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <syslog.h>
#include <stdlib.h>
#include <stdarg.h>
#include "firewall.h"
#include "fwfunc.h"

#define TESTBUF 46
#define TESTAV 8

int testenargv(char* rawstring, int count, ...)
{
	char	buf[TESTBUF];
	char	*av[TESTAV];
	char	*token;
	int	i,t;
	va_list	ap;

	t = enargv(rawstring, av, TESTAV, buf, TESTBUF);
	printf("enargv(\"%s\"): %d tokens: ", rawstring, t);
	if (t != count) {
		printf("count mismatch (expected %d) FAILED\n", count);
		return(1);
	}
	if (t <= 0) {
		printf("PASSED\n");
		return(0); /* We are expected to fail here */
	}
	va_start(ap, count);
	for (i = 0; i < t; i++) {
		token = va_arg(ap, char *);
		if (strcmp(av[i], token)) {
			printf("|%s| != |%s| FAILED\n", av[i], token);
			return(1);
		} else
			printf("|%s", av[i]);
	}
	va_end(ap);
	printf("| PASSED\n");
	return(0);
}

int testenargv_n(char* rawstring, int count, ...)
{
	char	buf[TESTBUF];
	char	*av[TESTAV];
	char	*token;
	int	i,t;
	va_list	ap;

	t = enargv_n(rawstring, av, TESTAV, buf, TESTBUF);
	printf("enargv_n(\"%s\"): %d tokens: ", rawstring, t);
	if (t != count) {
		printf("count mismatch (expected %d) FAILED\n", count);
		return(1);
	}
	if (t <= 0) {
		printf("PASSED\n");
		return(0); /* We are expected to fail here */
	}
	va_start(ap, count);
	for (i = 0; i < t; i++) {
		token = va_arg(ap, char *);
		if (strcmp(av[i], token)) {
			printf("|%s| != |%s| FAILED\n", av[i], token);
			return(1);
		} else
			printf("|%s", av[i]);
	}
	va_end(ap);
	printf("| PASSED\n");
	return(0);
}


int 
main(int argc,char* argv[])
{
	int rc = 0;

	printf("Testing enargv() (buffer size is %d, max count is %d)..\n", TESTBUF, TESTAV);

	rc |= testenargv("aa bbb   ccc ddd", 4,  "aa", "bbb", "ccc", "ddd" );
	rc |= testenargv("aa bbb    \"ccc ddd\"", 3,  "aa", "bbb", "ccc ddd" );
	rc |= testenargv_n("aa bbb    \"ccc ddd\"", 3,  "aa", "bbb", "\"ccc ddd\"" );
	rc |= testenargv("aa bbb 'ccc ddd'", 3,  "aa", "bbb", "ccc ddd" );
	rc |= testenargv_n("aa bbb 'ccc ddd'", 4,  "aa", "bbb", "'ccc", "ddd'" );
	rc |= testenargv("aa bbb \"'ccc ddd\"' eee fff'", 3,  "aa", "bbb", "'ccc ddd eee fff");
	rc |= testenargv_n("aa bbb \"'ccc ddd\"' eee fff'", 5,"aa","bbb","\"'ccc ddd\"'","eee","fff'");
	rc |= testenargv("aa \"bbb\\t12345\"", 2,  "aa", "bbb\t12345");
	rc |= testenargv_n("aa \"bbb\\t12345\"", 2,  "aa", "\"bbb\\t12345\"");
	rc |= testenargv("aa bbb\\t12345", 2,  "aa", "bbb\t12345");
	rc |= testenargv_n("aa bbb\\t12345", 2,  "aa", "bbb\\t12345");
	rc |= testenargv("aa bbb\t12345", 3,  "aa", "bbb","12345");
	/* Try token count overflow */
	rc |= testenargv("aa bbb ccc ddd eee ffff ggg", 7, "aa","bbb","ccc","ddd","eee","ffff","ggg"); 
	rc |= testenargv("aa bbb ccc ddd eee ffff ggg hhh", -1); 
	rc |= testenargv("aa bbbcccdddeeeffffggghhhiiijjjzzzzzzzzzzzzzzzzzzzzzzzzzzzzzzz", -1); 
	if (rc) {
		printf("Tests failed.\n");
		exit(1);
	} else {
		printf("Tests complete.\n");
		exit(0);
	} 
}
