/*
 * Copyright (c) 1999-2007, OpenFWTK Development Group
 * All rights reserved. See LICENSE.
 */


#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <syslog.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include "firewall.h"
#include "firewall2.h"
#include "fwfunc.h"

#define GOOGLE_DNS "8.8.8.8"
#define GOOGLE_DNS_NAME "google-public-dns-a.google.com"
#define NOREVERSE_HOST "1.1.1.1"

extern char* ip2name(struct in_addr);

int have_sane_dns()
{
	in_addr_t	norev;
	struct	in_addr	gdns;

	norev = inet_addr(NOREVERSE_HOST);
	gdns.s_addr = inet_addr(GOOGLE_DNS);
	return (!strcmp(ip2name(gdns), GOOGLE_DNS_NAME)) &&
	        !gethostbyaddr(&norev,sizeof(in_addr_t), AF_INET);
}

int testhostmatch(char* pattern, char* host, int expected)
{
	printf("hostmatch(\"%s\", \"%s\") (%d)..",pattern, host, expected);
	if (hostmatch(pattern, host) == expected) {
		printf("PASSED\n");
		return(0);
	}
	printf("FAILED\n");
	return(1);
}

int testhostmatchp(char* pattern, char* host, int defport, int port, 
		    int expected)
{
	printf("hostmatchp(\"%s\", \"%s\", %d, %d) (%d)..",pattern, host,
		defport, port, expected);
	if (hostmatchp(pattern, host, defport, port) == expected) {
		printf("PASSED\n");
		return(0);
	}
	printf("FAILED\n");
	return(1);
}
	

int 
main(int argc,char* argv[])
{
	int rc = 0;

	printf("Performing offline tests..\n");

	rc |= testhostmatch("*","192.168.0.1",1);
	rc |= testhostmatch("*","101.101.101.101",1);
	rc |= testhostmatch("192.168.0.*","192.168.0.1",1);
	rc |= testhostmatch("192.168.0.0/255.255.255.0","192.168.0.1",1);
	rc |= testhostmatch("192.168.0.0/24","192.168.0.1",1);
	rc |= testhostmatch("192.168.1.0/24","192.168.0.1",0);
	rc |= testhostmatch("192.168.0.0/0xffffff00","192.168.0.1",1);
	rc |= testhostmatch("*.tis.com","www.tis.com",1);
	rc |= testhostmatch("*.tis.com","tis.com",0);
	rc |= testhostmatch(".tis.com","tis.com",0);
	rc |= testhostmatch("google.*","google.ru",1);
	rc |= testhostmatch("google.*","google.com.ru",1);
	rc |= testhostmatch("*.google.com","8.8.8.8",1);
	rc |= testhostmatchp("*","192.168.0.1",443,8080,0);
	rc |= testhostmatchp("*","192.168.0.1",443,443,1);
	rc |= testhostmatchp("*:*","192.168.0.1",443,8080,1);
	rc |= testhostmatchp("*:8080","192.168.0.1",443,8080,1);

	if (have_sane_dns()) {
		printf("We have internet name resolution working, performing online tests..\n");
		rc |= testhostmatch("*.tis.com",GOOGLE_DNS,0);
		rc |= testhostmatch(GOOGLE_DNS,GOOGLE_DNS_NAME,1);
		rc |= testhostmatch("8.8.0.0/16",GOOGLE_DNS_NAME,1);
		rc |= testhostmatch("unknown",GOOGLE_DNS,0);
		rc |= testhostmatch("unknown",NOREVERSE_HOST,1);
		rc |= testhostmatch("unknown","unknown",1);
		printf("Now disabling reverse address lookup..\n");
		proxy_nodns = 1;
		rc |= testhostmatch("unknown",GOOGLE_DNS,1);
		proxy_nodns = 0;
	} else
		printf("DNS unavailable, online tests skipped\n");
	if (rc) {
		printf("Tests failed.\n");
		exit(1);
	} else {
		printf("Tests complete.\n");
		exit(0);
	} 
}
