/*
 * Copyright (c) 1996-2010, OpenFWTK Development Group
 * All rights reserved. See LICENSE.
 */

#include <stdio.h>
#include <stdlib.h>  
#include <string.h>
#include <strings.h>
#include <syslog.h>
#include <signal.h>
#include <unistd.h> 
#include <sys/types.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <net/if.h>
#include "firewall.h"
#include "fwfunc.h"

#ifdef IPFILTER
#ifdef PF
#error "You cannot compile both IPFilter and PF support, choose one."
#endif
#include "ip_compat.h"
#include "ip_fil.h"
#include "ip_nat.h"
#endif

#ifdef PF
#include "net/pfvar.h"
#endif

void
proxy_get_transparent_dst(dsthost,port)
char *dsthost;
int  *port;
{
	socklen_t		len;
	struct sockaddr_in	to_addr;
#ifdef IPFILTER
	struct sockaddr_in	from_addr;
	int			nat_fd;
	struct	natlookup	natlookup;
#if defined(IPFILTER_VERSION) && (IPFILTER_VERSION >= 4000027)
    	struct ipfobj obj;

    	obj.ipfo_rev = IPFILTER_VERSION;
    	obj.ipfo_size = sizeof(natLookup);
    	obj.ipfo_ptr = &natLookup;
    	obj.ipfo_type = IPFOBJ_NATLOOKUP;
    	obj.ipfo_offset = 0;
#endif /* IPFILTER_VERSION */
#else
#ifdef PF
	struct pfioc_natlook 	natlookup;
#endif
#endif /* IPFILTER */

	bzero((char *)&to_addr, sizeof(to_addr));

	len = sizeof(to_addr);
	if (getsockname(0, (struct sockaddr *)&to_addr, &len) < 0) {
		syslog(LLEV,"fwtksyserr: getsockname() failed: %s", strerror(errno));
		exit(1);
	}

#if (!defined(IPFILTER) && !defined(PF))

	strncpy(dsthost,inet_ntoa(to_addr.sin_addr),sizeof(dsthost));
	return;

#else

	bzero((char *)&from_addr, sizeof(from_addr));
	if (getpeername(0, (struct sockaddr *)&from_addr, &len) < 0) {
		syslog(LLEV,"fwtksyserr: getpeername() failed: %s", strerror(errno));
		exit(1);
	}

	bzero(natlookup, sizeof(natlookup));

#ifdef  IPFILTER
	natlookup.nl_inport = to_addr.sin_port;
	natlookup.nl_outport = from_addr.sin_port;
	natlookup.nl_inip = to_addr.sin_addr;
	natlookup.nl_outip = from_addr.sin_addr;
	natlookup.nl_flags = IPN_TCP;
#endif

	if ((nat_fd = open(
#ifdef	IPNAT_NAME
			      IPNAT_NAME, O_RDONLY
#else
#ifdef  IPFILTER
			      IPL_NAT, O_RDONLY
#else
#ifdef 	PF
			      "/dev/pf", O_RDWR		      
#else
#error "Unknown transparency option"
#endif /* PF */
#endif /* IPFILTER */
#endif /* IPNAT_NAME (implies IPFILTER) */
					           )) < 0) {
		syslog(LLEV,"fwtksyserr: cannot open NAT deivce: %s", strerror(errno));
		exit(1);
	}

#ifdef	IPFILTER
	if (ioctl(nat_fd, SIOCGNATL, 
#if defined(IPFILTER_VERSION) && (IPFILTER_VERSION >= 4000027)
    					&obj
#else
					&natlookup
#endif
						  ) == -1) {
		syslog(LLEV,"fwtksyserr: SIOCGNATL failed: %s", strerror(errno));
		close(nat_fd);
		exit(1);
	}

	strlcpy(dsthost,inet_ntoa(natlookup.nl_realip),16);
	*port = ntohs(natlookup.nl_realport);
	return;

#endif

#ifdef PF
	natlookup.saddr.v4.s_addr = from_addr.sin_addr.s_addr;
	natlookup.sport = from_addr.sin_port;
	natlookup.daddr.v4.s_addr = to_addr.sin_addr.s_addr;
	natlookup.dport = to_addr.sin_port;
	natlookup.af = AF_INET;
	natlookup.proto = IPPROTO_TCP;
	natlookup.direction = PF_OUT;

	if (ioctl(nat_fd, DIOCNATLOOK, &natlookup) < 0) {
		syslog(LLEV,"fwtksyserr: DIOCNATLOOK failed: %s", strerror(errno));
		close(nat_fd);
		exit(1);
	}
	close(nat_fd);
	strlcpy(dsthost,inet_ntoa(natlookup.rdaddr.v4),16);
	*port = ntohs(natlookup.rdport);
	return;
#endif /* PF */
#endif /* IPFILTER || PF */
}



