/*
 * Copyright (c) 1998-2011, OpenFWTK Development Group
 * All rights reserved. See LICENSE.
 */

/*
 * pop3-gw.c, a Post Office Protocol version 3 proxy designed to work
 *                with TIS fwtk
 *
 * (C) Copyright 1998-2004 by ArkanoiD <ark@eltex.ru>
 *
 */

#include <stdio.h>
#include <stdlib.h>  
#include <ctype.h>  
#include <string.h>
#include <strings.h>
#include <syslog.h>
#include <signal.h>
#include <unistd.h> 
#include <sys/types.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <netdb.h>
#ifdef	OpenBSD
#include <ssl/md5.h> 
#else
#include <openssl/md5.h> 
#endif
#include <assert.h> 
#include <net/if.h>
#ifdef SOLARIS
#include <fcntl.h>
#endif
#include "firewall.h"
#include "sslfunc.h"
#include "libemfw.h"
#include "emio.h"
#include "firewall2.h"
#include "fwfunc.h"
#include "auth.h"
#include "ci_milter.h"
#include "headers.h"
#include "message.h"
#include "icap.h"
#include "filters.h"

static  char*   moduleId ATTR_UNUSED = "$Id: pop3-gw.c,v 1.46 2011/04/11 18:39:36 arkenoi Exp $";

#define MLINE	0400		/* hardcoded	   */

#define	QTEMPLATE	"/tmp/pop3-gw.XXXXXXXXXX"

#define TIMESTAMPLEN 64
#include "transaction.h"


#define	BUFSIZE	4096		/* should not be less than any definitions
				   below				    */	
#define MAXCMD	256		/* Maximum length of pop3 command with args */
#define MAXRESP	256		/* Longest single-line responce		    */
#define MAXLINE 998		/* Longest message line, rfc2822	    */

#define	POP3S_PORT	995	/* May be missing from services file	    */

static	int	fallback = 0;
static	int	transparent = 0;
static	int	secureauth = 0;
static	int	extendperm = 0;
static	int	client_ssl = 0;
static	int	server_ssl = 0;
static	int	client_stls = 0;
static	int	server_stls = 0;

static	Cfg	*confp;
static	char	**validests = (char **)0;
static	char	**validusers = (char **)0;
static	char	**filters = (char **)0;
static  char	separator[2] = "@";
static	int	user_limit = AUTH_USIZ;
static	int	pass_limit = AUTH_PWSIZ;
static	int	cdscp = 0;
static	int	sdscp = 0;

static EMI_FILE	*c_in = NULL;
static EMO_FILE	*c_out = NULL;
static EMI_FILE	*s_in = NULL;
static EMO_FILE	*s_out = NULL;

static	char	username[AUTH_USIZ+1];

#define STAT_SYNC() proxy_stats.inbytes = emi_amount(s_in,0); \
		    proxy_stats.outbytes = emo_amount(s_out,0);

static	char	*authtype;

static	struct capabilities {
		int	capa;
		int	top;
		int	user;
		int	uidl;
		int	stls;
		int	sasl_login;
		int	sasl_plain;
		} server_cap = { 0, 1, 1, 1, 0, 0, 0 } ;

#if (MAXCMD >= MAX_STR) || (MAXRESP >= MAX_STR) || (MAXLINE >= MAX_STR)
#error MAX_STR is too small!
#endif

static fwparm	options[] = {	
	{	FWPARM_BOOL,	"-apop-only",	(char*)	&secureauth	},
	{	FWPARM_BOOL,	"-secure-auth",	(char*)	&secureauth	},
	{	FWPARM_BOOL,	"-fallback",	(char*)	&fallback	},
 	{	FWPARM_BOOL,	"-transparent",	(char*)	&transparent	},
	{	FWPARM_BOOL,	"-extnd",	(char*)	&extendperm	},
	{	FWPARM_CHAR,	"-separator",		separator	},
	{	FWPARM_LIST,	"-dest",	(char*)	&validests	},
	{	FWPARM_LIST,	"-user",	(char*) &validusers	},
	{	FWPARM_LIST,	"-filter",	(char*) &filters	},
	{	FWPARM_STRING,	"-plug-to",	proxy_stats.dst		},
	{	FWPARM_PORT,	"-port",	(char*)&proxy_stats.port},
	{	FWPARM_STRING,	"-authuser",	proxy_stats.authuser	},
	{	FWPARM_BOOL,	"-ssl-client",	(char*) &client_ssl	},
	{	FWPARM_BOOL,	"-ssl-server",	(char*) &server_ssl	},
	{	FWPARM_BOOL,	"-stls-server",	(char*) &server_stls	},
	{	FWPARM_OPLOG,	"-log",	(char*) &transaction_descr	},
	{	FWPARM_OPDENY,	"-deny",(char*) &transaction_descr	},
	{       FWPARM_DSCP,    "-client-dscp", (char*) &cdscp		},
	{       FWPARM_DSCP,    "-server-dscp", (char*) &sdscp		},
	{	0, 		0,		0 			} };

extern	int auth_perm(Cfg*,char*,char*,char*,char**); 
static  int get_response(int);
static  int get_multiline();
static void send_error(const char *) ATTR_NORETURN;

static int 
trim_eol(s)
char	*s;
{
	if (!s || strcmp(s+strlen(s)-2,"\r\n"))
		return(1);
	else
		s[strlen(s)-2] = '\0';
	return(0);
}

static int
read_server(buffer,maxsize)
char*	buffer;
int	maxsize;
{
	int	x;

	memset(buffer,0,maxsize);
	if (emi_gets(s_in, buffer, maxsize-1) < 0) {
		if (((x = strlen(buffer))== (maxsize-1))) {
			syslog(LLEV,"securityalert: server line too long, truncated");
			return(x);
		}
		syslog(LLEV,"error reading server data");
		emo_puts(c_out,"-ERR error reading server data");
		emo_flush(c_out);
		proxy_exit();
	} else
		return(x);
	return(-1); 
}

#define CHECK_SERVER() if (emo_error(s_out) || emi_error(s_in)) { \
			syslog(LLEV,"error writing server data, %s", strerror(errno)); \
			emo_puts(c_out,"-ERR error writing server data"); \
			emo_flush(c_out);	\
			proxy_exit(); \
		     }

#define CHECK_CLIENT() if (emo_error(c_out) || emi_error(c_in)) { \
			if (errno) \
				syslog(LLEV,"error reading client data, %s", strerror(errno)); \
			proxy_exit(); \
		     }


static void
puts_qf(fd,line)
int	fd;
char	*line;
{
	int	x;
	x = strlen(line);
	if ((write(fd,line,x) != x) || (write(fd,"\r\n",2) != 2)) {
		close(fd);
		syslog(LLEV,"fwtksyserr: error writing quarantine file, %s", strerror(errno));
		proxy_exit();
	}
}

static void
reject_qf(char* qf)
{
	int	tfd;

	if (unlink(qf)) {
		syslog(LLEV,"fwtksyserr: error unlinking quarantine file, %s", strerror(errno));
		proxy_exit();
	}
	if ((tfd = open(qf,S_IRUSR|S_IWUSR,O_CREAT|O_TRUNC|O_WRONLY)) == -1) {
		syslog(LLEV,"fwtksyserr: cannot create temporary file");
		proxy_exit();
	}
        puts_qf(tfd, "Return-Path: <quarantine@localhost>");
	puts_qf(tfd, "From: \"Firewall content inspection service\" <quarantine@localhost>");
	puts_qf(tfd, "Subject: Message was rejected by content inspection service");
	puts_qf(tfd, "X-Reject: YES");
	puts_qf(tfd, "Lines: 1\r\n");
	puts_qf(tfd, "");
	puts_qf(tfd, "The message does not meet filtering criteria and cannot be displayed");
	close(tfd);
}

/*
 *	Transaction state functions 
 */


/*
 *	Handle a pop3 command that has no arguments
 */

int
do_noargs(command,flags,argc,argv)
char*	command;
int	flags;
int	argc;
char	**argv;
{
	if (argc != 1) {
		syslog(LLEV,"%.100s command syntax error %d",command,argc);
		emo_puts(c_out,"-ERR syntax error\r\n");
		return(0);
	} else {
		emo_puts(s_out,argv[0]); emo_puts(s_out,"\r\n");
		get_response(flags);
	}
	return(0);
}

/*
 *	Handle pop3 commands that take single message number as argument
 */

int
do_message(command,flags,argc,argv)
char*	command;
int	flags;
int	argc;
char	**argv;
{
	char			buf[1024];
	unsigned	int	msgnum;

	memset(buf,0,sizeof(buf));

	if (argc != 2) {
		syslog(LLEV,"%.100s command syntax error",command);
		emo_puts(c_out,"-ERR syntax error\r\n");
		return(0);
	} else {
		if (!(msgnum = atoi(argv[1]))) {
			emo_puts(c_out,"-ERR message number invalid\r\n");
			return(0);
		}
		emo_printf(s_out,"%.4s %u",argv[0],msgnum);
		get_response(flags);
	}
	return(0);
}

/*
 *	Handle generic pop3 command - no advanced parsing here (wrong?)
 */

static int
do_generic(command,flags,argc,argv)
char*	command;
int	flags;
int	argc;
char	**argv;
{
	emo_puts(s_out,command); emo_puts(s_out,"\r\n");
	get_response(flags);
	return(0);
}


/*
 *	Handle RETR and TOP commands (special case)
 */

int
do_retr_top(command,flags,argc,argv)
char*	command;
int	flags;
int	argc;
char	**argv;
{
	char			buf[BUFSIZE];
	char			request[MAXCMD+1];
	char			okmesg[MAXRESP+1];
	unsigned	int	msgnum;
	unsigned	int	numlines;
	uint64_t		savedcount;
	int			tfd;
	char*			qf;
	struct 	hdr_line	*msghdr;
	struct 	hdr_line	*h;
	FILE			*tf;
	char			*cmd;
	struct	app_data	adata;
	char			*err_msg;
	char*			reject_policy;
	char*			request_style;

	memset(buf,0,sizeof(buf));
	memset(okmesg,0,sizeof(okmesg));
	numlines = 99999999;

	if (argc < 2) {
		syslog(LLEV,"%.100s command syntax error",command);
		emo_puts(c_out,"-ERR syntax error\r\n");
		return(0);
	}

	if (!(msgnum = atoi(argv[1]))) {
		emo_puts(c_out,"-ERR message number invalid\r\n");
		return(0);
	}

	if (!(strcasecmp(argv[0],"RETR"))) {
		if (argc != 2) {
			syslog(LLEV,"%.100s command syntax error",command);
			emo_puts(c_out,"-ERR syntax error\r\n");
			return(0);
		}
		snprintf(buf,MAXCMD,"RETR %u\r\n",msgnum);
		cmd = "RETR";
	} else if (!(strcasecmp(argv[0],"TOP"))) {
		if (argc != 3) {
			syslog(LLEV,"%.100s command syntax error",command);
			emo_puts(c_out,"-ERR syntax error\r\n");
			return(0);
		}
		numlines = atoi(argv[2]);
		snprintf(buf,MAXCMD,"TOP %u %u\r\n",msgnum,numlines);
		cmd = "TOP";
	} else {
		syslog(LLEV,"fwtksyserr: internal logic error, cannot handle %.128s",command);
		proxy_exit();
	}

	if (!numlines)
		return(do_generic(command,flags,argc,argv));

	emo_puts(s_out,buf); emo_flush(s_out); CHECK_SERVER();
	read_server(buf,MAXRESP);
	if (strncmp(buf,"+OK",3)) {
		emo_puts(c_out,buf);
		emo_puts(c_out,"\r\n");
		return(0);
	}

	qf = xstrdup(QTEMPLATE);
	if ((tfd = mkstemp(qf)) == -1) {
		syslog(LLEV,"fwtksyserr: cannot create temporary file");
		proxy_exit();
	}
	savedcount = proxy_stats.inbytes;
	adata.smtp_sender = NULL;
/* 
 *	XXX the way we get from address is ugly and definitely needs fixing
 */

/*
 *	header processing starts here
 */

	msghdr = header_read(s_in, "pop3");
	STAT_SYNC();


	if ((h = header_field_by_name(msghdr,"return-path"))) {
		if (!(adata.smtp_sender = message_addr_strip_brackets(header_field_value(h, NULL))))
			syslog(LLEV,"warning: malformed Return-Path");
	} else if ((h = header_field_by_name(msghdr,"from"))) {
		if (!(adata.smtp_sender = message_addr_strip_brackets(header_field_value(h, NULL))))
			syslog(LLEV,"warning: malformed From");
	}

	if (header_is_truncated(msghdr))
			goto qfdone;
		
/*
 *	Now we continue with message body
 */
	while(1) {
		read_server(buf,MAXLINE); 
		STAT_SYNC();
		if (!strcmp(buf,"."))
			break;
		puts_qf(tfd,buf);
	}
qfdone:
	close(tfd);

	if (!filters)
		goto resp;

	if (!adata.smtp_sender)
		adata.smtp_sender = "unknown";

	adata.smtp_rcpts = xmalloc(sizeof(char*)*2);
	snprintf(buf,BUFSIZE-1,"%s@%s", username, proxy_stats.dst); 
	adata.smtp_rcpts[0] = xstrdup(buf);
	adata.smtp_rcpts[1] = NULL;
	adata.target_host = proxy_stats.dst;
	adata.encode = 0;
	adata.headers = msghdr; 


/*	Fake a pop3 "http-like" request, somewhat resembling rfc2384 */

	if (!(request_style = proxy_conf_string(proxy_confp,
				       		"filter-request-style")) ||
	    !(strcmp(request_style,"http"))) 
		snprintf(request, sizeof(request)-1,
			 "GET pop://%s;auth=%s@%s:%u/?%u HTTP/1.0", username, 
			 authtype, proxy_stats.dst, proxy_stats.port, msgnum );
	else if (!(strcmp(request_style,"native"))) 
		snprintf(request, sizeof(request)-1,"%s pop://%s;auth=%s@%s:%u/?%u POP/3.0", cmd, username, authtype, proxy_stats.dst, proxy_stats.port, msgnum ); 
	else {
		syslog(LLEV,"fwtkcfgerr: invalid request style \"%s\"",
			request_style);
		exit(1);
	}

	switch (filter_inspect_object(qf, NULL, request, RESPMOD, filters, 
				      &adata, &err_msg)) {
	case -1:
		syslog(LLEV,"fwtksyserr: content inspection failed: %s",err_msg);
		emo_puts(c_out,"-ERR content inspection failed\r\n");
		emo_flush(c_out);
	 	proxy_exit();
	case 1:
		/*
	 	 *	no policy or "replace" - replace message with  
		 * 	error template
		 *	no special action taken - the mail cleint will decide
	 	 */
		if (!(reject_policy = proxy_conf_string(proxy_confp,
					       "filter-reject-policy")) ||
		    !(strcmp(reject_policy,"replace"))) {
			reject_qf(qf);
			goto resp;
		}
		/*
		 *	"delete" policy - delete message and send -ERR 
		 */	
		if (!(strcmp(reject_policy,"delete"))) {
			emo_printf(c_out,"DELE %u\r\n",msgnum);
			read_server(buf,MAXRESP);
			if (strncmp(buf,"+OK",3)) {
				emo_puts(c_out,"-ERR message rejected by "
					       "content filter; "
					       "further errors occured while "
					       "deleting message\r\n");
				return(0);
			}
		}
		/*
		 *	"keep" policy - send -ERR to client but do not delete
		 */
		emo_puts(c_out,"-ERR message rejected by content filter\r\n");
		return(0);
	case 0:
		break;
	default:
		syslog(LLEV,"fwtksyserr: internal logic error, filter_inspect_object return value is invalid");
		proxy_exit();
	}
resp:
	emo_puts(c_out,"+OK\r\n");
	header_write(c_out, msghdr);
	emo_puts(c_out,"\r\n");

	if (!(tf = fopen(qf,"r"))) {
		syslog(LLEV,"fwtksyserr: cannot open quarantine file");
		proxy_exit();
	}
	memset(buf,0,sizeof(buf));
	while(fgets(buf,sizeof(buf),tf)) {
		if (trim_eol(buf))
			syslog(LLEV,"warning: unterminated line");
		emo_puts(c_out,buf);
		emo_puts(c_out,"\r\n");
		memset(buf,0,sizeof(buf));
	}
	emo_puts(c_out,".\r\n");
	emo_flush(c_out);
	fclose(tf);
	unlink(qf); 
	header_free(&msghdr); 
	free(adata.smtp_rcpts[0]);
	free(adata.smtp_rcpts);
	free(qf);

	syslog(LLEV,"mail from=<%.128s> to=<%.40s@%.80s> bytes=%llu",
		     adata.smtp_sender,username,proxy_stats.dst,
		     proxy_stats.inbytes-savedcount);
	return(0);
}

/*
 *	Commands that are multiline if no arg is given and provide single
 *	responce if message number is ised as argument (LIST and UIDL)
 */

int
do_list(command,flags,argc,argv)
char*   command;
int     flags;
int     argc;
char    **argv;
{
	switch (argc) {
	case 1:
		return(do_noargs(command,MLINE,argc,argv));
	case 2:
		return(do_message(command,0,argc,argv));
	default:
		syslog(LLEV,"%.100s command syntax error",command);
		emo_puts(c_out,"-ERR syntax error\r\n");
		return(0);
	}
}

/*
 *	Handle restricted and unimplemented commands
 */

int
do_restrict(command,flags,argc,argv)
char*	command;
int	flags;
int	argc;
char	**argv;
{
	emo_puts(c_out,"-ERR not implemented\r\n");
	return(0);
}

int
do_nostate(command,flags,argc,argv)
char*	command;
int	flags;
int	argc;
char	**argv;
{
	emo_puts(c_out,"-ERR not allowed in transaction state\r\n");
	return(0);
}

/*
 *	Handle CAPA
 */

int
do_capa(command,flags,argc,argv)
char*	command;
int	flags;
int	argc;
char	**argv;
{
	emo_puts(c_out,"+OK Capabilities list follows\r\n");
	if (server_cap.top)
		emo_puts(c_out, "TOP\r\n"); 
	if (!secureauth || c_out->sslio)
		emo_puts(c_out, "USER\r\n"); 
	if (server_cap.uidl)
		emo_puts(c_out, "UIDL\r\n"); 
	if (client_stls)
		emo_puts(c_out, "STLS\r\n"); 
	emo_puts(c_out,".\r\n");
	return(0);
}

/*
 *	Handle QUIT command and termination state
 */

int
do_quit(command,flags,argc,argv)
char*	command;
int	flags;
int	argc;
char	**argv;
{
	emo_puts(s_out,"QUIT\r\n");
	get_response(0);
	proxy_exit();
	/*NOTREACHED*/
	return(0);
}

static int
get_multiline()
{
	char		buf[1024];

	while(1) {
		read_server(buf,MAXLINE); 
		emo_puts(c_out,buf); emo_puts(c_out,"\r\n"); emo_flush(c_out);
		CHECK_CLIENT();
		if (!strcmp(buf,"."))
			return(0);
	}
}

static int
get_response(int flags)
{
	char		buf[BUFSIZE];

	emo_flush(s_out);
	CHECK_SERVER();
	read_server(buf,MAXRESP);
	emo_puts(c_out,buf); emo_puts(c_out,"\r\n"); emo_flush(c_out);

	if (!strncmp(buf,"+OK",3) && (flags & MLINE)) 
		return(get_multiline());
	return(0);
}

static int
parse_user_cmd(buffer,user)
char	*buffer;
char	*user;
{
	char	*p;

	if ((p = strtok(buffer+5,separator)) != NULL)
		strlcpy(user,p,AUTH_USIZ);
	if ((p = strtok(NULL,":")) != NULL)
		strlcpy(proxy_stats.dst,p,MAX_HOSTNAME);
	if ((p = strtok(NULL,"\r\n")) != NULL)
		proxy_stats.port = atoi(p);
	if (!proxy_stats.port)
		proxy_stats.port = str_to_port("pop3");
	if ((strlen(user)==0) || (strlen(proxy_stats.dst)==0) || 
	    (strlen(user) > user_limit))
		return(1);
	return(0);
}

static int
parse_apop_cmd(buffer,user,digest)
char	*buffer;
char	*user;
char	*digest;
{
	char	*p;
	if ((p = strtok(buffer+5," ")) != NULL)
		strlcpy(user,p,AUTH_USIZ);
	if ((p = strtok(NULL,"\r\n")) != NULL)
		strncpy(digest,p,32);
	return ((strlen(user)==0) || 
		(strlen(user) > user_limit) || (strlen(digest)!=32));
}

static int
parse_server_greeting(buffer,timestamp)
char	*buffer;
char	*timestamp;
{
	char	*p;

	if (strncasecmp(buffer,"+OK",3)) 
		return(1);
	
	strlcpy(timestamp,"none",TIMESTAMPLEN);

	if ((p = strchr(buffer,'<'))) {
		if (!strchr(p,'>')) {
			syslog(LLEV,"securityalert: server timestamp %.128s is malformed", p);
			return(1);
		} else {
			int i;

			for (i = 0; i < strlen(p); i++)
				if (!isascii(p[i]) || iscntrl(p[i])) {
					syslog(LLEV,"securityalert: server timestamp %.128s is malformed", p);
					return(1);
				}

			strlcpy(timestamp,p,TIMESTAMPLEN);
			syslog(LLEV,"received timestamp from server: %.64s",timestamp);
		}
	}
	return(0);

}

static void
tohex(hex,raw,len)
unsigned char *hex; 
unsigned char *raw;
int	len;
{
	char *hdig="0123456789abcdef";

	while(len > 0){
		len--;
		*hex++ = hdig[((*raw)>>4)&0xf];
		*hex++ = hdig[(*raw++)&0xf];
	}
	*hex++ = '\0';
	return;
}

/* 
 *	Authentication functions.
 *	return zero on success and present +OK to user.
 *	return non-zero on failure and let the calling code explain the
 *	situation (sometimes we fail silently)
 */

static int
do_server_auth_raw(user,digest)
char	*user;
char	*digest;
{
	char buf[BUFSIZE];

	emo_printf(s_out,"APOP %s %s\r\n",user,digest);
	emo_flush(s_out);

	read_server(buf,MAXRESP);

	if (strncasecmp(buf,"+OK",3)) 
		return(1);

	emo_puts(c_out,"+OK authenticated to server via APOP");
	if (s_out->sslio)
		emo_puts(c_out," (TLS protected)");
	emo_puts(c_out,"\r\n");
	emo_flush(c_out);
	return(0);
}

static int
do_server_auth(user,password,timestamp)
char	*user;
char	*password;
char	*timestamp;
{
	MD5_CTX	ctx;
	unsigned char digest_raw[17];
	unsigned char digest_hex[35];
	char buf[BUFSIZE];

	MD5_Init(&ctx);
	MD5_Update(&ctx,(unsigned char*) timestamp,strlen(timestamp));
	MD5_Update(&ctx,(unsigned char*) password,strlen(password));
	MD5_Final(digest_raw,&ctx);
	tohex(digest_hex,digest_raw,16);
	emo_printf(s_out,"APOP %s %s\r\n",user,digest_hex);
	emo_flush(s_out);

	read_server(buf,MAXRESP);

	if (strncasecmp(buf,"+OK",3)) 
		return(1);

	emo_puts(c_out,"+OK authenticated to server via APOP");
	if (s_out->sslio)
		emo_puts(c_out," (TLS protected)");
	emo_puts(c_out,"\r\n");
	emo_flush(c_out);
	return(0);

}

static int
do_server_pauth(user,password)
char	*user;
char	*password;
{
	char buf[MAXRESP];

	emo_printf(s_out,"USER %.128s\r\n",user);
	emo_flush(s_out);
	read_server(buf,MAXRESP);
	if (strncasecmp(buf,"+OK",3)) {
		syslog(LLEV,"username rejected by server, %.256s", buf);
		return(1);
	}
	emo_printf(s_out,"PASS %.128s\r\n",password);
	emo_flush(s_out);
	read_server(buf,MAXRESP);
	if (strncasecmp(buf,"+OK",3)) 
		return(1);

	emo_puts(c_out,"+OK authenticated to server via USER/PASS");
	if (s_out->sslio)
		emo_puts(c_out," (TLS protected)");
	emo_puts(c_out,"\r\n");
	emo_flush(c_out);
	return(0);
	
}

static void
check_server_caps()
{
	char buf[MAXRESP];

	emo_printf(s_out,"CAPA\r\n");
	emo_flush(s_out);
	read_server(buf,MAXRESP);
	if (strncasecmp(buf,"+OK",3)) {
		server_cap.capa = 0;
		return;
	}

	server_cap.capa = 1;
	server_cap.top = 0;
	server_cap.user = 0;
	server_cap.uidl = 0;
	server_cap.stls = 0;
	server_cap.sasl_login = 0;
	server_cap.sasl_plain = 0;

	do {
		read_server(buf,MAXRESP);
		if (!strcasecmp(buf,"TOP"))
			server_cap.top = 1; 
		else if (!strcasecmp(buf,"USER"))
			server_cap.user = 1; 
		else if (!strcasecmp(buf,"UIDL"))
			server_cap.uidl = 1; 
		else if (!strcasecmp(buf,"STLS"))
			server_cap.stls = 1; 
		else if (!strcasecmp(buf,"SASL")) {
			char	tokbuf[MAX_STR];
			char	*tokav[MAX_ARG];
			int	tokac;
			int	i;

			tokac = enargv(buf,tokav,56,tokbuf,sizeof(tokbuf));

			for (i = 1; i < tokac; i++) {
				if (!strcasecmp(tokav[i],"LOGIN"))
					server_cap.sasl_login = 1;
				else if (!strcasecmp(tokav[i],"PLAIN"))
					server_cap.sasl_plain = 1;
			}
		}
	} while (strcmp(buf,"."));
}	

static void
send_error(const char *errtext)
{
	if (!c_in->sslio && client_ssl) {
		emio_sslaccept(c_in, c_out, 
			       proxy_stats.rladdr, proxy_stats.riaddr, "none", 
			       proxy_SSL_makeup_error_cert(proxy_stats.dst), 
		               NULL);
		if (!c_in->sslio) {
			syslog(LLEV, "Cannot negotiate client SSL conenction");
			exit(1);
		}
	}
	emo_printf(c_out,"-ERR %s\r\n",errtext);
	emo_flush(c_out);
	exit(1);
}

static void
negotiate_server_ssl()
{
	int	r;
	char	buffer[MAX_STR];
	char	*server_vrfy = proxy_conf_string(proxy_confp,
						 "server-verify");

	if (!server_vrfy)
		server_vrfy = "basic";

	emio_sslconnect(s_in, s_out, proxy_stats.dst, "none",
			NULL, NULL);
	if (!s_in->sslio)
		send_error("Unable to estabilish server SSL connection");	
	DEBUG_ASSERT(sizeof(buffer) > 512);
	r = proxy_SSLcheck_cert(s_out->sslio,proxy_stats.dst,
				strcmp(server_vrfy,"strict") ? 
				"basic" : "checkaddr", buffer);
	if (PROXYSSL_CHECK_VRFY(server_vrfy, r)) 
		send_error("Server certificate validaion falied");
}

/*
 * Open pop3 server. No return value: we quit if no success.
 */

static void
pop3_open_server(char *timestamp)
{
	char	buffer[MAX_STR];
	int	server_fd = -1;

	DEBUG_ASSERT(MAX_STR > MAXRESP);

	proxy_update_operation("CONNECTING");

	if (proxy_stats.port == POP3S_PORT)
		server_ssl = 1; /* Are we just being too smart here? */

	if ((server_fd = conn_server(proxy_stats.dst,proxy_stats.port,0,buffer)) < 0) {
		syslog(LLEV,"cannot connect to server %s/%d: %s", proxy_stats.dst,proxy_stats.port, buffer);
		send_error("POP3 gateway destination connect failed");
		proxy_exit();
	}
	if (sdscp)
		proxy_set_dscp(server_fd,sdscp);
	s_in = emi_fdopen (server_fd, 4096);
	s_out = emo_fdopen (server_fd, 4096);
	if (s_in == NULL || s_out == NULL) {
		syslog (LLEV, "fwtksyserr: fdopen() failed: %s", strerror(errno));
		exit(1);
	}
	emi_timeout (s_in, PROXY_TIMEOUT);
	emo_timeout (s_out, PROXY_TIMEOUT);

	if (server_ssl) 
		negotiate_server_ssl();

	if (emi_gets(s_in,buffer,MAXRESP) == -1) {
		syslog(LLEV,"error reading server data");
		send_error("error reading server data");
	}
	if (parse_server_greeting(buffer,timestamp) != 0) {
		close(server_fd);
		syslog(LLEV,"server greeting message is invalid");
		send_error("server greeting message is invalid");
	}
	check_server_caps();
	if (!s_in->sslio && server_stls && server_cap.stls) {
		emo_puts(s_out,"STLS\r\n");
		emo_flush(s_out);
		read_server(buffer,MAXRESP);
		if (strncasecmp(buffer,"+OK",3)) {
			syslog(LLEV,"Server declared STLS capability, but rejected encryption request");
			send_error("Server declared STLS capability, but rejected encryption request");
		} else {
			negotiate_server_ssl();
		}
	}
	/*
	 *	Client SSL connection may be already estabilished at this point.
	 *	If it is not, we may spoof the certificate.
	 */
	if (client_ssl) {
		if (!c_in->sslio)
			emio_sslaccept(c_in, c_out,
			       proxy_stats.rladdr, proxy_stats.riaddr, "none", 
			       s_in->sslio ? 
				proxy_SSLspoof_server_cert(s_in->sslio) : NULL,
			       NULL);
		if (!c_in->sslio) {
			syslog(LLEV, "Cannot negotiate client SSL conenction");
			exit(1);
		}
	}

}

int
main(int argc, char *argv[])
{
	char			buffer[MAX_STR];
	char			pass[AUTH_PWSIZ];
	char			timestamp[TIMESTAMPLEN];
	char			digest_hex[35];
	int			server_fd = -1;
	int			stage_complete;
	char			tokbuf[MAX_STR];
	char			*tokav[MAX_ARG];
	int			tokac;
	Cfg			*cf;
	int			defport;
	const struct hash_entry	*stls_he;

	proxy_init(argc,argv);
	proxy_SSLinit();
	proxy_chroot_setugid();

	if (!strcmp(proxy_name,"pop3s-gw")) {
		proxy_stats.port = defport = POP3S_PORT;
		client_ssl = server_ssl = 1;
	} else
		proxy_stats.port = defport = str_to_port("pop3");



	c_in = emi_fdopen (0, 4096);
	c_out = emo_fdopen (0, 4096);
	if (c_in == NULL || c_out == NULL) {
		syslog (LLEV, "fwtksyserr: fdopen() failed: %s", strerror(errno));
		exit (1);
	}

	emi_timeout (c_in, PROXY_TIMEOUT);
	emo_timeout (c_out, PROXY_TIMEOUT);

/*
 *	Configure global filters first
 */

	if ((cf=cfg_get("milter", confp))) {
		syslog(LLEV,"fwtkcfgerr: 'milter' configuration tag is deprecated, use 'filter' instead");
		exit(1);
	}

	cf=cfg_get("filter", confp);

	while (cf != (Cfg *)0) {
		int	i;
		if((cf->argc < 1) || (cf->argv[0][0] == '-')) {
			syslog(LLEV,"fwtkcfgerr: missing parameter, line %d",cf->ln);
		}
		for (i = 0; i < cf->argc; i++)
			addlist(cf->argv[i], &filters);
		cf=cfg_get("filter",(Cfg *)0);
	}
	
	if ((confp=proxy_conf_hosts(proxy_confp,proxy_stats.rladdr,						       proxy_stats.riaddr))) 
		proxy_parse_options(confp,options);
	else {
		if (!client_ssl)
			emo_puts(c_out, "-ERR access denied\r\n");
		proxy_exit();
	}

	if (cdscp)
		 proxy_set_dscp(0,cdscp);

	user_limit = proxy_conf_int(proxy_confp,"user-limit",1,sizeof(username)-1,user_limit);
	pass_limit = proxy_conf_int(proxy_confp,"pass-limit",1,sizeof(pass)-1,pass_limit);

	memset(buffer, 0, sizeof(buffer));
	memset(username, 0, sizeof(username));
	memset(pass, 0, sizeof(pass));

	/* Check flags */
	if (!(stls_he = find_he2(&transaction_descr,"STLS",4))) {
		syslog(LLEV,"fwtksyserr: function handler not found");
		exit(1);
	}
	client_stls = !(stls_he->flags & OP_DENY);

/* 	
 *	Connect/auth part may seem tricky, actually it is not.
 *	If we know the destination (transparent or configured), we connect
 *	immediately, to get timestamp and let APOP fall through.
 *	If we do not, we parse USER command to find out the server name,
 *	the connect and try APOP if it is available.
 *	If APOP seems to be supported, fallback is permitted, but it 
 *	fails, we reconnect to the server and authenticate via USER/PASS 
 *	(yuck!). If it is not supported, we try USER/PASS immediately.
 *	Please note that fallback to USER/PASS is highly insecure and
 *	exposes the password in plaintext, so it is safer to deny it.
 */

#define CHECK_USERS() if (searchlist(username,validusers)) { \
			emo_puts(c_out,"-ERR access denied\r\n"); \
                        syslog(LLEV,"deny host=%.512s/%.20s access for user %s",proxy_stats.rladdr,proxy_stats.riaddr,username); \
			proxy_exit(); \
		    }

#define CHECK_DEST()  if (proxy_check_dest_port(validests,extendperm,defport)) { \
			emo_puts(c_out,"-ERR POP3 gateway denied destination access\r\n"); \
			proxy_exit(); \
		    }

#define	CHECK_SSLPORT() if (proxy_stats.port == POP3S_PORT) \
			{ defport = POP3S_PORT;  server_ssl = 1; }

	if (!(*proxy_stats.dst) && !transparent) {
	/*
	 *	We do not need timestamp, as APOP is not available in this mode.
	 */
		*timestamp = '\0';
	/*	
	 *	For non-transparent operation, we do not spoof any certificates.
	 */
		if (client_ssl) {
			emio_sslaccept(c_in, c_out,
			       proxy_stats.rladdr, proxy_stats.riaddr,
			       "none", NULL, NULL);
			if (!c_in->sslio) {
				syslog(LLEV, "Cannot negotiate client SSL conenction");
				exit(1);
			}
		}
	} else {
		if (transparent) {
/*
 *	We do not check if port number is permitted because it comes
 *	from redirection rule we defined manually anyways.
 */
			proxy_get_transparent_dst(proxy_stats.dst,&proxy_stats.port);
			if (proxy_stats.port == POP3S_PORT)
				client_ssl = 1;

			CHECK_SSLPORT();
			CHECK_DEST();
		} 

		CHECK_SSLPORT();
		pop3_open_server(timestamp);
	}
	

	emo_printf(c_out, "+OK POP3 gateway ready %.128s\r\n",timestamp);
	emo_flush(c_out);

	stage_complete = 0;
	proxy_update_operation("AUTHENTICATION");

/* 
 *	USER, APOP or QUIT 
 */


	while(!stage_complete) {
		if(emi_gets(c_in, buffer, 512) <= 0)
			proxy_exit();
		if (client_stls && !strcasecmp(buffer,"STLS")) {
			if (c_in->sslio) {
				emo_puts(c_out,"-ERR TLS connection already estabilished\r\n");
				emo_flush(c_out);
				continue;
			}
			emio_sslaccept(c_in, c_out,
			       proxy_stats.rladdr, proxy_stats.riaddr, "none", 
			       s_in->sslio ? 
				proxy_SSLspoof_server_cert(s_in->sslio) : NULL,
			       NULL);
			if (!c_in->sslio) {
				emo_puts(c_out,"-ERR Cannot negotiate client SSL conenction\r\n");
				syslog(LLEV, "Cannot negotiate client SSL conenction");
				emo_flush(c_out);
				exit(1);
			}
			emo_puts(c_out,"+OK SSL connection negotiated\r\n");
			emo_flush(c_out);
		} else if (!strncasecmp(buffer,"USER ",5)) {
			if (secureauth && !c_in->sslio) {
				emo_puts(c_out,"-ERR insecure protocol is not permitted\r\n");
				emo_flush(c_out);
				goto skip_user;
			}
			if (!(*proxy_stats.dst)) {
				if (parse_user_cmd(buffer,username) != 0) {
					emo_puts(c_out,"-ERR incorrect syntax, use user@host[:port] to specify destination\r\n"); 
					emo_flush(c_out);
				} else {
					CHECK_USERS();
					CHECK_SSLPORT();
					CHECK_DEST();
	
					pop3_open_server(timestamp);
					if (!strcmp(timestamp,"none") && !fallback && !s_out->sslio) {
						syslog(LLEV,"fallback to USER/PASS authentication is not permitted");
						emo_puts(c_out,"-ERR server does not support APOP\r\n");
						emo_flush(c_out);
						proxy_exit();
					} else 
						stage_complete = 1;
				}  
			} else {
				if ((buffer+5) && *(char*)(buffer+5) && 
				    (strlen(buffer+5) <= user_limit)) {
					strncpy(username,buffer+5,sizeof(username));
					CHECK_USERS();
					stage_complete = 1;
				} else {
					emo_puts(c_out,"-ERR syntax error\r\n");
					emo_flush(c_out);
				}
			}

skip_user:		;

		} else if (!strncasecmp(buffer,"APOP ",5)) { 
			if (!strcmp(timestamp,"none") || !*proxy_stats.dst) {
				emo_puts(c_out,"-ERR APOP is unsupported on client side\r\n");	
			} else {
				if (parse_apop_cmd(buffer,username,digest_hex)) {
					emo_puts(c_out,"-ERR syntax error\r\n");
				} else {
					CHECK_USERS();
					authtype = "APOP";
					if (do_server_auth_raw(username,digest_hex)) {
						emo_puts(c_out,"-ERR APOP authentication failed\r\n");
						proxy_exit();
					}
					else goto transaction_state;
				}
			}
						

		} else if (!strncasecmp(buffer,"CAPA",4)) 
			do_capa((char*) NULL, 0, 0, (char **) NULL);
		else if (!strncasecmp(buffer,"QUIT",4)) { 
			emo_puts(c_out,"+OK proxy session terminated\r\n");	
			proxy_exit();
		} else if (!strncasecmp(buffer,"CAPA",4) || 
          		   !strncasecmp(buffer,"AUTH",4)) {
			emo_puts(c_out,"-ERR not implemented\r\n");
		} else {
			emo_puts(c_out,"-ERR invalid command\r\n");
			syslog(LLEV,"unknown authorization state command %.100s",buffer);
		}
	}

	emo_puts(c_out,"+OK we are connected to server\r\n");
	emo_flush(c_out);

	stage_complete = 0;
/*
 *	PASS or QUIT 
 */

	while (!stage_complete) {
		if (emi_gets(c_in,buffer, 512) == -1) {
			CHECK_CLIENT();
			proxy_exit();
		} else if (!strncasecmp(buffer,"QUIT",4)) { 
			emo_puts(c_out,"+OK proxy session terminated\r\n");	
			emo_flush(c_out);
			proxy_exit();
		} else if (!strncasecmp(buffer,"PASS ",5)) {
			strncpy(pass,buffer+5,sizeof(pass));
			if (strlen(pass) <= pass_limit)
				stage_complete = 1;
			else {
				emo_puts(c_out,"-ERR password too long\r\n");
				emo_flush(c_out);
			}
		} else {
			emo_puts(c_out,"-ERR PASS command required to complete AUTHORIZATION state\r\n");
			emo_flush(c_out);
		} 
	}
/*
 *	Got user credentials, authenticate 
 */
	if (!s_out->sslio && !strcmp(timestamp,"none"))
		syslog(LLEV,"warning: no secure authentication methods available");
	if ( !strcmp(timestamp,"none") ||
	      (do_server_auth(username,pass,timestamp) != 0)) {
/*
 *		APOP authentication failed or not available
 */
		if (!s_out->sslio && !fallback) {
			close(server_fd);
			syslog(LLEV,"user failed to authenticate to server");
			emo_puts(c_out,"-ERR server authentication failed\r\n");
			emo_flush(c_out);
			proxy_exit();
		} else if (strcmp(timestamp,"none")) {
/*
 *		APOP authentication failed so we have to reconnect to server
 */
			emi_close(s_in);
			emo_close(s_out);
			syslog(LLEV,"warning: APOP authentication failed, reverting to insecure mode"); 
			pop3_open_server(timestamp);
		}
/*
 *		Now trying USER/PASS
 */
	        if (do_server_pauth(username,pass) != 0) {
			emo_puts(c_out,"-ERR authentication failed\r\n");
			emo_flush(c_out);
			syslog(LLEV,"user failed to authenticate to server");
			proxy_exit();
		}
		authtype = "USER";
	} else
		authtype = "APOP";

/*
 *	Proceed to TRANSACTION state
 */
transaction_state:


	while(1) {
		int				msg_int;
		const	struct	hash_entry	*he;
		int				(*handler)(char*,int,int,char**);

		proxy_update_operation("IDLE");
		if(emi_gets(c_in, buffer, 512) <= 0) {
			CHECK_CLIENT();
			continue;
		}
		tokac = enargv(buffer,tokav,56,tokbuf,sizeof(tokbuf));
		if(!tokav[0])
			continue;
		if (!(he = find_he2(&transaction_descr,tokav[0],strlen(tokav[0])))) {
			syslog(LLEV,"unknown pop3 command %.100s",buffer);
			emo_puts(c_out,"-ERR command unknown or out of state\r\n");
			emo_flush(c_out);
			continue;
		} else {
			proxy_update_operation(buffer);
			if (he->flags & OP_LOG)
				syslog(LLEV,"%.512s/%.20s: %.512s",proxy_stats.rladdr,proxy_stats.riaddr,buffer);
			if (he->flags & OP_DENY) {
				syslog(LLEV,"deny host=%.512s/%.20s operation %.100s",proxy_stats.rladdr,proxy_stats.riaddr,buffer);
				emo_puts(c_out,"-ERR operation denied\r\n");
				emo_flush(c_out);
				continue;
			}
			if(extendperm && (he->flags & OP_XTND)) {
				msg_int = auth_perm(proxy_confp,proxy_stats.authuser,"pop3-gw",proxy_stats.riaddr,tokav);
				if(msg_int == 1 || msg_int == -1) {
					syslog(LLEV,"deny host=%.512s/%.20s operation %.100s - extended permissions",proxy_stats.rladdr,proxy_stats.riaddr,buffer);
					emo_puts(c_out,"-ERR operation denied\r\n");
					emo_flush(c_out);
					continue;
				}
			}
			handler = (int(*)(char*,int,int,char**))he->handler;
			assert (TRANSACTION_HANDLER ((int(*)())handler));
			handler(buffer,he->flags,tokac,tokav);
			emo_flush(c_out);
			emo_flush(s_out);
			STAT_SYNC();
			continue;
		}
	}
/*
 *	Unreached
 */					
	proxy_exit();
}


