/*
 * Copyright (c) 1996-2011, OpenFWTK Development Group
 * All rights reserved. See LICENSE.
 */

/*
 * This is independant code branch modified to fit fwtk configuration and
 * logging style by ArkanoiD, 2001
 *
 * Original copyright info follows.
 */
	
/*
 * smtpd, Obtuse SMTP daemon, storing agent. does simple collection of
 * mail messages, for later forwarding by smtpfwdd.
 *
 * Copyright (c) 1996, 1997 Obtuse Systems Corporation. All rights
 * reserved.
 *   
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Obtuse Systems 
 *      Corporation and its contributors.
 * 4. Neither the name of the Obtuse Systems Corporation nor the names
 *    of its contributors may be used to endorse or promote products 
 *    derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY OBTUSE SYSTEMS CORPORATION AND
 * CONTRIBUTORS ``AS IS''AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL OBTUSE SYSTEMS CORPORATION OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

char *obtuse_copyright =
"Copyright 1996 - Obtuse Systems Corporation - All rights reserved.";

#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <pwd.h>
#include <grp.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <arpa/nameser.h>
#include <sys/wait.h>
#include <unistd.h>
#include <time.h>
#include <assert.h>

#ifndef NS_MAXDNAME
#define NS_MAXDNAME MAXDNAME
#endif

#ifndef MIN
# define MIN(a, b) (((a) > (b)) ? (b) : (a))
#endif

#define SMTPS_PORT 465

#include <syslog.h>
#include <errno.h>
#include <limits.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/utsname.h>
#include <string.h>
#include <sysexits.h>
#include <ctype.h>
#include <string.h>
#include <strings.h>
#include <fcntl.h>
#include <sys/select.h>

#ifdef	OpenBSD
#include <ssl/md5.h> 
#else
#include <openssl/md5.h> 
#endif
#ifdef SOLARIS
#include <fcntl.h>
#endif


#include "smtp.h"
#include "firewall.h"
#include "sslfunc.h"
#include "libemfw.h"
#include "emio.h"
#include "firewall2.h"
#include "fwfunc.h"
#include "auth.h"
#include "ci_milter.h"
#include "headers.h"
#include "icap.h"
#include "base64.h"
#include "filters.h"
#include "http.h"
#include "message.h"
#include "smtpd.h"

/*
 * handler state flags
 */

#define	NEED_HELO	000400
#define	NEED_TLS	001000
#define	NEED_MAIL	002000
#define	NEED_RCPT	004000

#define STATE_MASK	07400

/* 
 * state
 */

#define	OK_HELO		000400
#define	OK_TLS		001000
#define	OK_MAIL		002000
#define	OK_RCPT		004000
#define	OK_DATA		010000
#define	OK_EHLO		020000
#define	OK_AUTH		040000

#include "esmtp.h"

static char* moduleId ATTR_UNUSED = "$Id: smtpd.c,v 1.16 2011/04/06 12:34:27 arkenoi Exp $";

#ifndef SPOOLSUBDIR
#define SPOOLSUBDIR "."
#endif

static	char	myname[MAXHOSTNAMELEN];
static	char *client_claimed_name = "UNKNOWN";
static	char *spoolfile = NULL;
static	char *bodyfile = NULL;
static	char *spooldir = SPOOLSUBDIR;	/* this is relative to our chroot. */
static	char	*auth_username = NULL;
static	char	*auth_password = NULL;
static	char	*auth_type = NULL;
static	long long maxsize = 0;
static	int maxheader;
static	int maxrecip;
static	int currecip = 0;
EMI_FILE	*c_in = NULL;
EMI_FILE	*s_in = NULL;
EMO_FILE	*c_out = NULL;
EMO_FILE	*s_out = NULL;
EMO_FILE	*spool_out = NULL;
int exiting = 0;
char my_clean_reverse_name[MAXHOSTNAMELEN];
static	int cdscp = 0;
static	int sdscp = 0;
static	int client_ssl = 0;
static	int server_ssl = 0;
static	int server_stls = 0;
static	int transp = 0;
static	int extendperm = 0;
static	int auth = 0;
static	int user_limit = AUTH_USIZ;
static	int pass_limit = AUTH_PWSIZ;
static	char **filters = (char**) 0;
static	char **validusers = (char**) 0;
static	char **validests = (char**) 0;
static	char **domains = (char**) 0;
static	struct app_data adata;

static	struct	capabilities {
	int 		esmtp;
	int		starttls;
	long long	size;
	int		sasl_login;
	int		sasl_plain;
	} server_cap = { 0, 0, 0, 0, 0 };

static	fwparm	options[] = {
		{FWPARM_DSCP,	"-client-dscp", (char*) &cdscp 			},
		{FWPARM_DSCP,	"-server-dscp",	(char*) &sdscp 			},
		{FWPARM_BOOL,	"-ssl-client", 	(char*) &client_ssl		},
		{FWPARM_BOOL,	"-ssl-server", 	(char*) &server_ssl		},
		{FWPARM_BOOL,	"-stls-server",	(char*) &server_stls		},
		{FWPARM_BOOL,	"-authall", 	(char*) &auth			},
		{FWPARM_STRING,	"-authuser", 	proxy_stats.authuser 		},
		{FWPARM_BOOL,	"-transparent",	(char*) &transp 		},
		{FWPARM_BOOL,	"-extnd",	(char*) &extendperm 		},
		{FWPARM_LIST,	"-filter",	(char*) &filters		},
		{FWPARM_LIST,	"-dest",	(char*) &validests		},
		{FWPARM_LIST,	"-relay-domains",(char*) &domains		},
		{FWPARM_LIST,	"-user",	(char*) &validusers		},
		{FWPARM_STRING,	"-plug-to",	proxy_stats.dst 		},
		{FWPARM_PORT,	"-port",	(char*)&proxy_stats.port	},
		{FWPARM_OPLOG,	"-log",		(char*)&esmtp_descr		},
		{FWPARM_OPDENY,	"-deny",	(char*)&esmtp_descr		},
		{0,		0,		0 			}};

extern  int auth_perm(Cfg*,char*,char*,char*,char**);

static void smtp_exit(int val) ATTR_NORETURN;
static char * cleanitup(const char *s);
static char * smtp_cleanitup(const char *s);

int	state;

/*
 * Return to the initial state
 */

void
reset_state(int s)
{
	s &= OK_HELO | OK_EHLO | OK_TLS;
  /*
   * we must throw away anything in the output buffer 
   */
	if (spool_out)
		emo_flush(spool_out);
	if (s_out)
		emo_flush(s_out);
}


/*
 * Write a possibly multi-segement reply to client
 */
int
writereply(EMO_FILE *out,
	   int code,
	   int more,
	   ...)
{
  	va_list	ap;
  	char 	*msg;

  	va_start(ap, more);
  	emo_printf(out,"%3d%s", code, (more) ? "-" : " ");

  	while (!emo_error(out) && (msg = va_arg(ap, char *)) != NULL) 
		emo_puts(out, msg);
  	if (!emo_error(out)) 
		emo_puts(out, CRLF);
  	va_end(ap);
	emo_flush(out);
  	return (!emo_error(out));
}

/*
 * open a new spoolfile with appropriate lock and permissions 
 */

static EMO_FILE*
smtp_open_spoolfile(const char *s)
{
  	EMO_FILE	*f = NULL;

  	if (!s) {
		char	buffer[MAXPATHLEN];
		int	fd;
		snprintf(buffer, MAXPATHLEN, "%s/spoolXXXXXXXX", spooldir);
		if ((fd = mkstemp(buffer)) != -1)
			f = emo_fdopen(fd, 4096);
  	} else
		f = emo_open(s, 4096);

	if (!f) {
    		syslog(LLEV, "fwtksyserr: couldn't create spool file %s!", spoolfile);
    		smtp_exit(EX_CONFIG);
	}
  	if (lock_fd(f->fd) != 0) {
    		syslog(LLEV, "fwtksyserr: couldn't lock spool file %s", spoolfile);
    		smtp_exit(EX_TEMPFAIL);
  	}

  	return (f);
}

/*
 * close spoolfile, unlock, and open permissions 
 */

static void
smtp_close_spoolfile(EMO_FILE *spool)
{
  	if (spoolfile == NULL) {
    		syslog(LLEV, "fwtksyserr: attempt to close NULL spoolfile!");
    		smtp_exit(EX_CONFIG);
  	}
  	if (locktest_fd(spool->fd) == 0)
    		if (lockun_fd(spool->fd) != 0) {
      			syslog(LLEV, "fwtksyserr: couldn't unlock spool file %s", spoolfile);
      			smtp_exit(EX_OSERR);
    		}
	emo_close(spool);
  	chmod(spoolfile, 0750);	/*
				 * Mark file as 'complete' 
				 */
  	free(spoolfile);
  	spoolfile = NULL;
}

/*
 * unlock spoolfile and remove it 
 */
static void
smtp_nuke_spoolfile(EMO_FILE *spool)
{
  	if (spoolfile == NULL) {
    		syslog(LLEV, "fwtksyserr: attempt to remove NULL spoolfile!");
    		smtp_exit(EX_SOFTWARE);
  	}
  	if (unlink(spoolfile) != 0) {
    		syslog(LLEV, "fwtksyserr: couldn't remove spool file %s! (%s)", spoolfile, strerror(errno));
    		free(spoolfile);
    		spoolfile = NULL;
    		smtp_exit(EX_CONFIG);
  	}
  	if (locktest_fd(spool->fd) == 0)
    		if (lockun_fd(spool->fd) != 0) {
      			syslog(LLEV, "fwtksyserr: couldn't unlock spool file %s using lockf! (%s)", spoolfile, strerror(errno));
      			free(spoolfile);
      			spoolfile = NULL;
      			smtp_exit(EX_OSERR);
    		}
  	emo_close(spool);
  	free(spoolfile);
  	spoolfile = NULL;
}

/*
 * Try to say something meaningful to our client and then exit.
 */

static void
smtp_exit(int val)
{
  	if (val != 0) {
    		/*
     	 	 * we're leaving the client hanging. attempt to tell them we're
     	 	 * going away 
     	 	 */
    		if (exiting++<1) 
       			writereply(c_out, 421, 0, "exiting sucked", NULL);
     
    		/*
    	 	 * if we have an open spool file that's unclosed, blast 
    	 	 * it out of existence 
    	 	 */
    		if (exiting++<2) 
       			if (spoolfile != NULL) 
	      			smtp_nuke_spoolfile(spool_out);
  	} else {
    		if (exiting++<1) {
       			if (spoolfile != NULL) {
	  			smtp_close_spoolfile(spool_out);
	  			exiting++;
       			}
    		}
  	}
  	if (exiting++<3)
		emo_flush(c_out); 

    	accumlog(LLEV, 0);		/* flush? */
     
  	syslog(LLEV,"exit host=%.128s/%.128s bytes=%llu",proxy_stats.rladdr,proxy_stats.riaddr,proxy_stats.outbytes);
  	exit(val);
}

/*
 * clean up things (mostly hostnames) that will go into the syslogs 
 */

static char *
cleanitup(const char *s)
{
  	static unsigned char *bufs[20];
  	static int first = 1, next_buffer;
  	unsigned char *dst, *buffer_addr;
  	const unsigned char *src;
  	int saw_weird_char, saw_bad_char, saw_high_bit = 0;

  	if (first) {
    		int i;

    		for (i = 0; i < 20; i += 1) 
      			bufs[i] = xmalloc(1024);
    		first = 0;
  	}
  	src = (const unsigned char *) s;
  	dst = bufs[(next_buffer++) % 20];
  	buffer_addr = dst;
  	saw_weird_char = 0;
  	saw_bad_char = 0;
  	saw_high_bit = 0;

  	while (*src != '\0') {
    		unsigned char xch, ch;

    		xch = *src++;
    		ch = xch & 0x7f;

    		if (ch != xch) {
      			saw_high_bit = 1;
      			*dst++ = '^';
      			*dst++ = '=';
    		}
    		if (isalnum(ch) || strchr(" -/:=.@_[]", ch) != NULL) 
      			*dst++ = ch;
    		else if (strchr("\\`$|;()*", ch) != NULL) {
      			saw_bad_char = 1;

      			*dst++ = '^';
      			*dst++ = 'x';
      			*dst++ = "0123456789abcdef"[(xch >> 4) & 0xf];
      			*dst++ = "0123456789abcdef"[(xch) & 0xf];

    		} else {
      			saw_weird_char = 1;

      			switch (ch) {
      			case '\n':
				*dst++ = '^';
				*dst++ = 'n';
				break;
      			case '\r':
				*dst++ = '^';
				*dst++ = 'r';
				break;
      			case '\t':
				*dst++ = '^';
				*dst++ = 't';
				break;
      			case '\b':
				*dst++ = '^';
				*dst++ = 'b';
				break;
      			case '^':
				*dst++ = '^';
				*dst++ = '^';
				break;
      			default:
				*dst++ = '^';
				*dst++ = 'x';
				*dst++ = "0123456789abcdef"[(xch >> 4) & 0xf];
				*dst++ = "0123456789abcdef"[(xch) & 0xf];
      			}

    		}

    		if (dst - buffer_addr > 1024 - 10) {
      			syslog(LLEV, "INFO(cleanitup) - buffer overflow - chopping");
      			break;
    		}
  	}

  	*dst = '\0';


  	if (saw_bad_char) 
    		syslog(LLEV, "securityalert: (cleanitup) - saw VERY unusual character (cleaned string is \"%s\")", buffer_addr);
  	if (saw_weird_char) 
    		syslog(LLEV, "INFO(cleanitup) - saw unusual character (cleaned string is \"%s\")", buffer_addr);
  	if (saw_high_bit) 
    		syslog(LLEV, "INFO(cleanitup) - saw character with high bit set (cleaned string is \"%s\")", buffer_addr);
  	return ((char*) buffer_addr);
}

/*
 * less paranoid version of cleanitup that tries to handle mail addresses
 * without mangling them.
 */
static char *
smtp_cleanitup(const char *s)
{
  static unsigned char *bufs[20];
  static int first = 1, next_buffer;
  unsigned char *dst, *buffer_addr;
  const unsigned char *src;
  int firstone, arg_attempt, saw_weird_char, saw_bad_char, saw_high_bit = 0;

  	if (first) {
    		int i;

    		for (i = 0; i < 20; i += 1) 
      			bufs[i] = xmalloc(1024);
    		first = 0;
  	}
  	src = (unsigned char* ) s;
  	dst = bufs[(next_buffer++) % 20];
  	buffer_addr = dst;
  	saw_weird_char = 0;
  	saw_bad_char = 0;
  	saw_high_bit = 0;
  	arg_attempt = 0;

  	firstone = 1;
  	while (*src != '\0') {
    		unsigned char xch, ch;

    		xch = *src++;
    		ch = xch & 0x7f;

    		if (ch != xch) {
      			saw_high_bit = 1;
      			*dst++ = '^';
      			*dst++ = '=';
    		}
    		/*
     		 * <sjg> RFC822 allows both ' and " in local-part.
     		 * " is infact _required_ if local-part contains spaces as is
    		 * common in x400 (yuk).
    		 */
    		if (isalnum(ch) || strchr(" -,:=.@_!<>()[]/+%'\"", ch)!=NULL) {
      			if (firstone && (ch == '-')) {
				arg_attempt = 1;
				*dst++ = '^';
				*dst++ = '-';
      			} else 
				*dst++ = ch;

    		} else if (strchr("\\`$|*;", ch) != NULL) {
      			saw_bad_char = 1;

      			*dst++ = '^';
      			*dst++ = 'x';
      			*dst++ = "0123456789abcdef"[(xch >> 4) & 0xf];
      			*dst++ = "0123456789abcdef"[(xch) & 0xf];

    		} else {
      			saw_weird_char = 1;

      			switch (ch) {
      			case '\n':
				*dst++ = '^';
				*dst++ = 'n';
				break;
      			case '\r':
				*dst++ = '^';
				*dst++ = 'r';
				break;
      			case '\t':
				*dst++ = '^';
				*dst++ = 't';
				break;
      			case '\b':
				*dst++ = '^';
				*dst++ = 'b';
				break;
      			case '^':
				*dst++ = '^';
				*dst++ = '^';
				break;
      			default:
				*dst++ = '^';
				*dst++ = 'x';
				*dst++ = "0123456789abcdef"[(xch >> 4) & 0xf];
				*dst++ = "0123456789abcdef"[(xch) & 0xf];
      			}
    		}

    		if (dst - buffer_addr > 1024 - 10) {
      			syslog(LLEV, "INFO(smtp_cleanitup) - buffer overflow - chopping");
      			break;
    		}
    		firstone = 0;
  	}

  	*dst = '\0';
  	if (arg_attempt) 
    		syslog(LLEV, "securityalert: (smtp_cleanitup) - '-' as first character in address (cleaned string is \"%s\")", buffer_addr);
  	if (saw_bad_char) 
    		syslog(LLEV, "securityalert: (smtp_cleanitup) - saw VERY unusual character (cleaned string is \"%s\")", buffer_addr);
  	if (saw_weird_char) 
    		syslog(LLEV, "INFO(smtp_cleanitup) - saw unusual character (cleaned string is \"%s\")", buffer_addr);
  	if (saw_high_bit) 
    		syslog(LLEV, "INFO(smtp_cleanitup) - saw character with high bit set cleaned string is \"%s\")", buffer_addr);
  	return ((char *) buffer_addr);
}


int
sane_state(int state)
{
  	if ((OK_MAIL & state) && !(OK_HELO & state)) {
    		syslog(LLEV, "fwtksyserr: bad state. can't be OK_MAIL and not OK_HELO");
    		return (0);
  	}
  	if ((OK_RCPT & state) && !(OK_MAIL & state)) {
    		syslog(LLEV, "fwtksyserr: bad state. can't be OK_RCPT and not OK_MAIL");
    		return (0);
  	}
  	return (1);
}

static int
smtp_get_line(char *buffer, int maxline)
{
	int	x;

	memset(buffer,0,maxline);
	if ((x = emi_gets(c_in, buffer, maxline-1)) < 0) {
		if (((x = strlen(buffer))== (maxline-1))) {
			syslog(LLEV,"securityalert: client line too long, truncated");
			return(x);
		}
		return(-1);
	} else {
		proxy_stats.outbytes = emi_amount(c_in,0);
		return(x);
	}
	return(-1);
}

static int
smtp_get_response(char *buffer, int maxline)
{
	int	x;

	memset(buffer,0,maxline);
	do {
		if (emi_gets(s_in, buffer, maxline - 1) < 0) {
			if (((x = strlen(buffer))== (maxline-1))) {
				syslog(LLEV,"securityalert: server line too long, truncated");
				return(x);
			}
			return(-1);
		} else
			proxy_stats.inbytes = emi_amount(s_in,0);
	} while (buffer[3] == '-');
	return (atoi(buffer) / 100);
}

static int
smtp_response_ok(int expected)
{
	char	buffer[MAXLINE];
	int 	rc = 0;

	if ((rc = smtp_get_response(buffer, sizeof(buffer))) <= 0) {
		writereply(c_out, 451, 0, "Remote connection closed or invalid data received", NULL);
		smtp_exit(EX_PROTOCOL);
	}
	if (rc != expected) {
	/*	Just relay the response	*/
		emo_puts(c_out, buffer);
		emo_puts(c_out, "\r\n");
		emo_flush(c_out);
		return(0);
	}
	return(1);
}
 
static int
esmtp_helo_ehlo()
{
	char	buffer[MAXLINE];

	proxy_hostname(s_out->fd, myname, NULL, MAXHOSTNAMELEN);
	if (smtp_get_response(buffer, MAXLINE) != 2)
		return(-1); 
	emo_printf(s_out, "EHLO %s\r\n", myname);
	emo_flush(s_out);
	while(emi_gets(s_in, buffer, MAXLINE - 1) > 0) {
		if (atoi(buffer) == 500) {
			emo_printf(s_out, "HELO %s\r\n", myname);
			emo_flush(s_out);
		/*
		 *	For HELO, only single-line 250 respose is valid.
		 */
			if ((emi_gets(s_in, buffer, MAXLINE - 1) <= 0) ||
			    (atoi(buffer) != 250) || (buffer[3] == '-'))
				return(-1);
			else
				break;
		}
		if ((atoi(buffer) != 250) || (strlen(buffer) < 4))
			return(-1);
		server_cap.esmtp = 1;
		if (!strncasecmp("SIZE ", buffer + 4, 5)) {
			if (!(server_cap.size = atoll(buffer + 9))) 
				return(-1);
			
			maxsize = MIN(maxsize, server_cap.size);
		} else if (!strcasecmp("STARTTLS", buffer + 4))
			server_cap.starttls = 1;
		else if (!strncasecmp("AUTH ", buffer + 4, 5)) {
			char    tokbuf[MAX_STR];
			char    *tokav[MAX_ARG];
			int     tokac;
			int     i;
			tokac = enargv(buffer+4,tokav,56,tokbuf,sizeof(tokbuf));

			for (i = 1; i < tokac; i++) {
				if (!strcasecmp(tokav[i],"LOGIN"))
					 server_cap.sasl_login = 1;
				else if (!strcasecmp(tokav[i],"PLAIN"))
					server_cap.sasl_plain = 1;
			}
		}
		if (buffer[3] != '-')
			break;
	}
	return((!emi_error(s_in) && !emo_error(s_out)) ? 0 : -1);
}
			

int
process_esmtp_options(int start, int argc, char **argv)
{
	int 	i;

	if (start >= argc)
		return(0);
	/*
	 *	No ESMTP features allowed if session was started with old HELO
	 */
 	if (!(state & OK_EHLO)) {
		writereply(c_out, 501, 0, m501msg, NULL);
		return(1);
	}

	for (i = start; i < argc; i++) {
	/*
	 *	Process ESMTP MAIL, RCPT and DATA options here
	 */
		if (!strncasecmp("SIZE=", argv[i], 5)) {
			int	mysize;
			if ((mysize = atoi(argv[i] + 5)) == 0) {
				writereply(c_out, 501, 0, m501msg, NULL);
				return(1);
			}
			if (mysize >= maxsize) {
				writereply(c_out, 552, 0, m552msg, NULL);
				return(1);
			}
		} else {
			writereply(c_out, 501, 0, m501msg, NULL);
			return(1);
		}
	}
	return(0);
}

static void
negotiate_server_ssl()
{
	int     r;
	char    buffer[MAX_STR];
	char    *server_vrfy = proxy_conf_string(proxy_confp, "server-verify");

	if (!server_vrfy)
		server_vrfy = "basic";

        emio_sslconnect(s_in, s_out, proxy_stats.dst, "none", NULL, NULL);

	if (!s_in->sslio) {
		writereply(c_out, 554, 0, "SSL connection failed", NULL);
		smtp_exit(EX_PROTOCOL);
	}
	DEBUG_ASSERT(sizeof(buffer) > 512);
	r = proxy_SSLcheck_cert(s_out->sslio,proxy_stats.dst,
				strcmp(server_vrfy,"strict") ?
				"basic" : "checkaddr", buffer);
	if (PROXYSSL_CHECK_VRFY(server_vrfy, r)) {
		writereply(c_out, 554, 0, "Server certificate valiadation failed", NULL);
		smtp_exit(EX_PROTOCOL);
	}
	state |= OK_TLS;
}

static void
smtp_open_server()
{
	char    buffer[MAX_STR];
	int     server_fd = -1;
	
	proxy_update_operation("CONNECTING");

	if (proxy_stats.port == SMTPS_PORT)
		server_ssl = 1; /* Are we just being too smart here? */

	if ((server_fd = conn_server(proxy_stats.dst,proxy_stats.port,0,buffer)) < 0) {
		syslog(LLEV,"cannot connect to server %s/%d: %s", proxy_stats.dst,proxy_stats.port, buffer);
		writereply(c_out, 451, 0, "Destination server not available", NULL);
		proxy_exit();
	}
	if (sdscp)
		proxy_set_dscp(server_fd,sdscp);
	s_in = emi_fdopen (server_fd, 4096);
	s_out = emo_fdopen (server_fd, 4096);
	if (s_in == NULL || s_out == NULL) {
		syslog (LLEV, "fwtksyserr: fdopen() failed: %s", strerror(errno));
		exit(1);
	}
	emi_timeout (s_in, PROXY_TIMEOUT);
	emo_timeout (s_out, PROXY_TIMEOUT);
	
	if (server_ssl)
		negotiate_server_ssl();

	if (esmtp_helo_ehlo() < 0) {
		writereply(c_out, 554, 0, "Destination server does not talk SMTP", NULL);
		proxy_exit();
	}
	if (!s_in->sslio && server_stls && server_cap.starttls) {
		emo_puts(s_out,"STARTTLS\r\n");
		if (smtp_get_response(buffer, MAXLINE) != 2) {
			syslog(LLEV,"Server declared STLS capability, but rejected encryption request");
			writereply(c_out, 554, 0, "Destination server rejected STARTTLS command", NULL);
		} else
			negotiate_server_ssl();
	}
	/*
	 *	Client SSL connection may be already estabilished at this point.
	 *	If it is not, we may spoof the certificate.
	 */
	if (client_ssl) {
		if (!c_in->sslio)
			emio_sslaccept(c_in, c_out,
				proxy_stats.rladdr, proxy_stats.riaddr, "none",
				s_in->sslio ?
				 proxy_SSLspoof_server_cert(s_in->sslio) : NULL,
				NULL);
		if (!c_in->sslio) {
			syslog(LLEV, "Cannot negotiate client SSL conenction");
			exit(1);
		}
	}
}

int
do_helo(char *command, int flags, int argc, char **argv)
{
	char 	buf[64];
	if (argc != 2) {
    		writereply(c_out, 500, 0, m500msg, NULL);
      		return(1);
	}	
	if (strlen(argv[1]) > SMTP_MAXFQNAME) {
      		/*
      		 * someone gave us a *big* name for themselves. draw them to
       		 * our attention, and fail. 
      		 */
      		syslog(LLEV, "securityalert: more than %u bytes on HELO from %.128s/%.128s).",
	     	       SMTP_MAXFQNAME, proxy_stats.rladdr, 
		       proxy_stats.riaddr);

      		return(1);
    	}
    	client_claimed_name = xstrdup(cleanitup(argv[1]));
    	if (strcmp(argv[1], client_claimed_name) != 0) {
      		syslog(LLEV, "securityalert: suspicious characters in HELO: hostname from host %.128s/%.128s, exiting",
     		proxy_stats.rladdr, proxy_stats.riaddr);
		smtp_exit(EX_PROTOCOL);
    	}
    	/*
    	 * log the connection 
    	 */
      	accumlog(LLEV, 0);	/* flush anything left */
      	accumlog(LLEV, "relay=%s/%s",
       			proxy_stats.rladdr,proxy_stats.riaddr);
	if (strcasecmp(proxy_stats.rladdr, client_claimed_name))
		accumlog(LLEV, " as \"%s\"", client_claimed_name);
	if (!strcasecmp(argv[0],"HELO")) 
    		writereply(c_out, 250, 0, my_clean_reverse_name, " ",
	       		   m250helook, " ", client_claimed_name, NULL);
	else if (!strcasecmp(argv[0],"EHLO")) {
    		writereply(c_out, 250, 1, my_clean_reverse_name, " ",
	       		   m250helook, " ", client_claimed_name, NULL);
		if (auth || 
		    (s_out && (server_cap.sasl_login || server_cap.sasl_plain)))
			writereply(c_out, 250, 1, "AUTH LOGIN PLAIN", NULL);
		if (maxsize) {
			snprintf(buf, sizeof(buf), "SIZE %llu", maxsize);
			writereply(c_out, 250, 1, buf, NULL);
		}
		writereply(c_out, 250, 0, "HELP", NULL);
		state |= OK_EHLO;
	} else {
		syslog(LLEV, 
		       "fwtksyserr: unknown command in HELO handler, %s", 
		       argv[1]);
		smtp_exit(EX_CONFIG);
	}
	state |= OK_HELO;
	return(0);
			
}

static int
server_auth(const char *username, const char *password)
{
	char	*encoded;
	char	buf[SMTP_MAX_CMD_LINE];

	if (server_cap.sasl_plain) {
		char	*p;
		bzero(buf,sizeof(buf));
		p = memccpy(buf+1, auth_username, '\0', 
			    strlen(auth_username) + 1);
		p = memccpy(p, auth_password, '\0', 
			    strlen(auth_password) + 1);
		base64_encode(buf, p - buf, &encoded);
		emo_printf(s_out,"AUTH PLAIN %s\r\n", encoded);
		emo_flush(s_out);
		if (!smtp_response_ok(2))
			return(1);
	} else if (server_cap.sasl_login) {
		char	*encoded;
		emo_printf(s_out,"AUTH LOGIN\r\n");
		emo_flush(s_out);
		if (!smtp_response_ok(3))
			return(1);
		base64_encode(auth_username, strlen(auth_username), 
			      &encoded);
		emo_puts(s_out, encoded);
		emo_puts(s_out, "\r\n");
		emo_flush(s_out);
		if (!smtp_response_ok(3))
			return(1);
		base64_encode(auth_password, strlen(auth_password), 
			      &encoded);
		emo_puts(s_out, encoded);
		emo_puts(s_out, "\r\n");
		emo_flush(s_out);
		if (!smtp_response_ok(2))
			return(1);
	} else {
		writereply(c_out, 535, 0, "No suitable authentication method found", NULL);
		return(1);
	}
	return(0);
}

static int
authsrv_auth(const char *username, const char *password)
{
	char	buf[512];

	if (strlen(username) > AUTH_USIZ) {
		writereply(c_out, 501, 0, "Username too long", NULL);
		return(1);
	}
	if (auth_open(proxy_confp) || auth_recv(buf, sizeof(buf)) ||
	    (strncmp(buf,"Authsrv ready",13))) {
		auth_close();
		writereply(c_out, 501, 0, "Unable to connect to authentication server", NULL);
		return(1);
	}
	snprintf(buf, sizeof(buf), "authorize %s 'smtpd %s/%s'",
		 username, proxy_stats.rladdr, proxy_stats.riaddr);
	if (auth_send(buf) || auth_recv(buf, sizeof(buf))) {
		auth_close();
		writereply(c_out, 501, 0, "Authentication server error", NULL);
		return(1);
	}
	if (!strncmp(buf, "ok", 2)) {
		auth_close();
		return(0);
	} else if (!strncmp(buf, "password", 8)) 
		snprintf(buf, sizeof(buf), "response %s", password);
	else {
		writereply(c_out, 501, 0, "Challenge/response protocol unsupported at the moment", NULL);
		auth_close();
		return(1);
	}
	if (auth_send(buf) || auth_recv(buf, sizeof(buf))) {
		auth_close();
		writereply(c_out, 501, 0, "Authentication server error", NULL);
		return(1);
	}
	auth_close();
	if (!strncmp(buf, "ok", 2)) 
		return(0);
	else {
		writereply(c_out, 501, 0, "Authentication failed", NULL);
		return(1);
	}
}


int
do_auth(char *command, int flags, int argc, char **argv)
{
	char	buf1[SMTP_MAX_CMD_LINE];
	char	buf2[SMTP_MAX_CMD_LINE];
	if (argc < 2) {
		writereply(c_out, 501, 0, m501msg, NULL);
		return(1);
	}
	if (!strcmp(argv[1],"LOGIN")) {
		if (argc != 2) {
			writereply(c_out, 501, 0, m501msg, NULL);
			return(1);
		}
		/* "Username:" in base64 */
		writereply(c_out, 334, 0, "VXNlcm5hbWU6", NULL); 
		smtp_get_line(buf1,SMTP_MAX_CMD_LINE);
		bzero(buf2, sizeof(buf2));
		if (base64_decode(buf1,buf2) < 0) {
			writereply(c_out, 501, 0, m501msg, NULL);
			return(1);
		}
		auth_username = xstrdup(buf2);
		/* "Password:" in base64 */
		writereply(c_out, 334, 0, "UGFzc3dvcmQ6", NULL); 
		smtp_get_line(buf1,SMTP_MAX_CMD_LINE);
		bzero(buf2, sizeof(buf2));
		if (base64_decode(buf1,buf2) < 0) {
			writereply(c_out, 501, 0, m501msg, NULL);
			return(1);
		}
		auth_password = xstrdup(buf2);
		auth_type = "LOGIN";
	} else if (!strcmp(argv[1],"PLAIN")) {
		int 	len;
		char 	*p;
		if (argc != 3) {
			writereply(c_out, 501, 0, m501msg, NULL);
			return(1);
		}
		bzero(buf1,sizeof(buf1));
		if ((len = base64_decode(argv[2],buf1)) < 0) {
			writereply(c_out, 501, 0, m501msg, NULL);
			return(1);
		}
		buf1[len] = '\0';
		p += strlen(buf1) + 1;
		if ((p - buf1) > len) {
			writereply(c_out, 501, 0, m501msg, NULL);
			return(1);
		}
		auth_username = xstrdup(p);
		p += strlen(p) + 1;
		if ((p - buf1) > len) {
			writereply(c_out, 501, 0, m501msg, NULL);
			return(1);
		}
		auth_password = xstrdup(buf1);
		auth_type = "PLAIN";
	} else {
		writereply(c_out, 501, 0, m501msg, NULL);
		return(1);
	}
	if (s_out) {
		if (server_auth(auth_username, auth_password))
			return(1);
	} else if (auth) {
		if (authsrv_auth(auth_username, auth_password))
			return(1);
	}
	writereply(c_out, 235, 0, "Authentication successful", NULL);
	state |= OK_AUTH;
	return(0);
}

int
do_starttls(char *command, int flags, int argc, char **argv)
{
	if (argc != 1) {
		writereply(c_out, 501, 0, m501msg, NULL);
		return(1);
	}
	emio_sslaccept(c_in, c_out, proxy_stats.rladdr, proxy_stats.riaddr, 
		       "none", NULL, NULL);
	if (!c_in->sslio) {
		syslog(LLEV, "Cannot negotiate client SSL conenction");
		exit(1);
	}	
	writereply(c_out, 220, 0, m220tls, NULL);
	return(0);
}

int
do_mail(char *command, int flags, int argc, char **argv)
{
	char	*p;
	int	i;

    	/*
    	 * at this point I shouldn't have anything bigger than a return 
    	 * address and a FROM: left. Well, there may be options, but we do
	 * not really care.
    	 */
    	if (strlen(command) > SMTP_MAX_MAILPATH + 7) {
      	/*
      	 * someone gave us a *big* name * for themselves. draw them to 
      	 * our attention, and fail. 
      	 */
      		syslog(LLEV, "securityalert: more than %d bytes on MAIL from address %.128s/%.128s",
     	       	       SMTP_MAX_MAILPATH, proxy_stats.rladdr, 
	       	       proxy_stats.riaddr);
      		return(1);
    	}
    	if (strncasecmp(argv[1], "FROM:", 5) != 0) {
      		writereply(c_out, 501, 0, m501msg, NULL);
      		return(1);
    	}
	if (strlen(argv[1]) == 5) {
	/*
	 * 	We have white space after MAIL FROM:.
	 * 	Actually it is rfc5321 violation, but still pretty widespread,
	 * 	so let's work around.
	 */
	 	p = argv[2];
		i = 3;
	} else {
		p = argv[1] + 5;
		i = 2;
	}
	
	if (process_esmtp_options(i, argc, argv))
		return(1);
			
    	adata.smtp_sender = xstrdup(smtp_cleanitup(p));
      
	if (strcmp(p, adata.smtp_sender) != 0) {
      		syslog(LLEV,"securityalert: suspicious characters in FROM: address from host  %.128s/%.128s, exiting",
       	      	       proxy_stats.rladdr, proxy_stats.riaddr);
		smtp_exit(EX_PROTOCOL);
    	}
	if (s_out) {
		emo_printf(s_out, "MAIL FROM:%s\r\n", adata.smtp_sender);
		emo_flush(s_out);
		if (!smtp_response_ok(2))
			return(1);
		if (emo_error(s_out)) {
			writereply(c_out, 554, 0, m554nofrom, NULL);
			return(1);
		}
	}  
	state |= OK_MAIL;

    	writereply(c_out, 250, 0, "sender is ", adata.smtp_sender,
       		   ", ", m250fromok, NULL);
      	accumlog(LLEV, " from=%s", adata.smtp_sender);
	return(0);
}


int
do_rcpt(char *command, int flags, int argc, char **argv)
{
	char	*p;
	int	i;
	char *victim;

    	/*
    	 * at this point I shouldn't have anything bigger than a return
    	 * address and a RCPT: left. 
    	 */
    	if (strlen(command) > SMTP_MAX_MAILPATH + 1) {
      	/*
      	 * someone gave us a *big* name for themselves. draw them to 
      	 * our attention, and fail. 
       	*/
      		syslog(LLEV, "securityalert: more than %d bytes on RCPT from address %.128s/%.128s",
		       SMTP_MAX_MAILPATH,proxy_stats.rladdr,proxy_stats.riaddr);
      		return(1);
    	}
 
    	if (strncasecmp(argv[1], "TO:", 3) != 0) {
      		writereply(c_out, 501, 0, m501msg, NULL);
      		return(1);
    	}
	if (strlen(argv[1]) == 5) {
	/*
	 * 	We have white space after RCPT TO:.
	 * 	Actually it is rfc5321 violation, but still pretty widespread,
	 * 	so let's work around.
	 */
	 	p = argv[2];
		i = 3;
	} else {
		p = argv[1] + 3;
		i = 2;
	}

	if (process_esmtp_options(i, argc, argv))
		return(1);
	
    	victim = xstrdup(smtp_cleanitup(p));
      
	if (strcmp(p, victim) != 0) {
      		syslog(LLEV,"securityalert: suspicious characters in RCPT TO: address from host  %.512s/%.20s, rejected",
       	      	       proxy_stats.rladdr, proxy_stats.riaddr);
		free(victim);
		writereply(c_out, 554, 0, m554norcpt, NULL);
		return(1);
    	}
    	if (maxrecip && (++currecip > maxrecip)) {
     		writereply(c_out, 550, 0, "too many recipients", NULL);
		free(victim);
     		syslog(LLEV,"too many recipients");
     		return(1);
    	}
	if (!(p = strchr(message_addr_strip_brackets(victim),'@')) || !(*++p)) {
		syslog(LLEV,"rejected non-fqdn recipient %s",  victim);
		free(victim);
		writereply(c_out, 554, 0, m554norcpt, NULL);
		return(1);
	}
	if (searchlisth(p, domains)) {
		syslog(LLEV,"deny host=%.512s/%.20s relay to %.128s", 
			    proxy_stats.rladdr, proxy_stats.riaddr, victim);
		free(victim);
		writereply(c_out, 554, 0, m554norcpt, NULL);
		return(1);
	}
	addlist(victim, &(adata.smtp_rcpts));
	if (s_out) {
		emo_printf(s_out, "RCPT TO:%s\r\n", victim);
		emo_flush(s_out);
		if (!smtp_response_ok(2)) {
			free(victim);
			return(1);
		}
		if (emo_error(s_out)) {
			writereply(c_out, 554, 0, m554norcpt, NULL);
			free(victim);
			return(1);
		}
	} 
	state |= OK_RCPT;
     	writereply(c_out, 250, 0, "recipient ", victim, ", ", m250rcptok, NULL);
	accumlog(LLEV, " to=%s",  victim);
	free(victim);
	return(0);

}

int
do_noop(char *command, int flags, int argc, char **argv)
{
	writereply(c_out, 250, 0, m250msg, NULL);
	return(0);
}

int
do_vrfy(char *command, int flags, int argc, char **argv)
{
	writereply(c_out, 252, 0, m252msg, NULL);
	return(0);
}


int
do_help(char *command, int flags, int argc, char **argv)
{
	writereply(c_out, 214, 1, "Commands", NULL);
	writereply(c_out, 214, 1, "HELO    MAIL    RCPT    DATA    RSET", NULL);
	writereply(c_out, 214, 0, "NOOP    QUIT    HELP    VRFY    EXPN", NULL);
	return(0);
}

int
do_wiz(char *command, int flags, int argc, char **argv)
{
	syslog(LLEV,"securityalert: %s command attempted from address "
		    "%.128s/%.128s, name %.256s", argv[0], 
	       proxy_stats.rladdr, proxy_stats.riaddr, client_claimed_name); 
	writereply(c_out, 250, 0, m250msg, NULL);
	return(0);
}

int
do_rset(char *command, int flags, int argc, char **argv)
{
	if (s_out) {
		emo_puts(s_out, "RSET\r\n");
		emo_flush(s_out);
		if (!smtp_response_ok(2))
			return(1);
	}
	writereply(c_out, 250, 0, m250msg, NULL);
	reset_state(state);
	return(0);
}

int
do_quit(char *command, int flags, int argc, char **argv)
{
	if (s_out) {
		emo_puts(s_out, "QUIT\r\n");
		emo_flush(s_out);
		if (!smtp_response_ok(2))
			return(1);
	}
	writereply(c_out, 250, 0, m221msg, NULL);
	smtp_exit(EX_OK);
}

static void
forge_received(EMO_FILE *out)
{
	char	head[512 + 1];
	time_t	tt;
	char	*foo;

	time(&tt);

	memset(head, 0, sizeof(head));
	strlcat(head, "Received: from ", 512);
	strlcat(head, proxy_stats.rladdr, 512);
	strlcat(head, "[", 512);
	strlcat(head, proxy_stats.riaddr, 512);
	if (strcasecmp(proxy_stats.rladdr, client_claimed_name) == 0) 
		strlcat(head, "]", 512);
	else {
		strlcat(head,"],claiming to be \"",512);
		strlcat(head, client_claimed_name, 512);
		strlcat(head, "\"", 512);
	}
	strlcat(head, " via smtpd by ", 512);
	strlcat(head, my_clean_reverse_name, 512);
	strlcat(head, "\r\n\tid ", 512);
	if ((foo = strrchr(spoolfile, '/')) != NULL) 
		strlcat(head, foo + 1, 512);
	else 
		strlcat(head, spoolfile, 512);
	strlcat(head, "; ", 512);
	strlcat(head, ctime(&tt), 512);
	strlcat(head, "\r", 512);

	emo_puts(out, head);
	emo_flush(out);
}

#if MAXLINE >= MAX_STR
#error "MAX_STR should be > MAXLINE"
#endif

static int
snarfdata(const long long limit, long long *size)
{
	char		buffer[MAX_STR + 1];
	long long	saved_amount;
	long long	svchdrsize;
	int	rc;
	char	*err_msg = NULL; 
	EMI_FILE 	*body_in;

  	/*
  	 * initial message size, ignore service header 
  	 */

	saved_amount = emi_amount(c_in,0);

	adata.headers = header_read(c_in, "smtp");
	if (emi_error(c_in)) {
		syslog(LLEV, "error receiving message header");
      		return (FAILURE);
	}
	svchdrsize = emi_amount(c_in,0) - saved_amount;

	if (svchdrsize + 512 >= limit) {
		writereply(c_out,552,0,m552msg, NULL);
		return(FAILURE);
	} 
	if (header_is_truncated(adata.headers)) {
		bodyfile = NULL;
	} else {
		snprintf(buffer, MAXLINE, "%s/msgbodyXXXXXX", spooldir);
		bodyfile = message_savebody(buffer, c_in, 
					    limit - svchdrsize - 512, &rc);

		switch (rc) {
		case 0:	break;
		case 1: writereply(c_out,552,0,m552msg, NULL);
			return(FAILURE);
		case 2: return(FAILURE);
		case 3: writereply(c_out,451,0,m451msg, NULL);
			return(ERROR);
		}

		*size = emi_amount(c_in, 0) - saved_amount;
	}
	if (!filters)
		goto	skip;

	/*	
	 *	Forge some url for icap.
	 *	The syntax is close to draft-earhart-url-smtp-00 as possible.
	 */

	snprintf(buffer, MAX_STR, "POST smtp://");
	if (auth_username) {
		strlcat(buffer, auth_username, MAX_STR);
		strlcat(buffer, ";auth=", MAX_STR);
		strlcat(buffer, auth_type, MAX_STR);
		strlcat(buffer, "@", MAX_STR);
	}
	if (s_out) {
		strlcat(buffer, proxy_stats.dst, MAX_STR);
		if (proxy_stats.port != 25) {
			char	portid[256];
			snprintf(portid, 255, ":%u", proxy_stats.port);
			strlcat(buffer, portid, MAX_STR);
		}
	} else 
		strlcat(buffer, myname, MAX_STR);

	switch(filter_inspect_object(bodyfile, NULL, buffer, REQMOD, filters,
				     &adata, &err_msg)) {
	case -1:
		syslog(LLEV,"fwtksyserr: content inspection failed: %s",err_msg);
		writereply(c_out,451,0,m451msg, NULL);
		return(ERROR);
	case 1:
		syslog(LLEV,"message rejected by filter");
		writereply(c_out,554,0,m554msg, NULL);
		return(FAILURE);
	case 0:
		break;
	default:
		syslog(LLEV,"fwtksyserr: internal logic error, filter_inspect_object return value is invalid");
		smtp_exit(EX_SOFTWARE);
	}
skip:
	if (!s_out) {
		char	**rcpt;

		if (bodyfile) {
			spoolfile = xstrdup(bodyfile);
			memcpy(strrchr(spoolfile,'/'), "/spoolXX", 8);
		} else
			spoolfile = NULL;

		spool_out = smtp_open_spoolfile(spoolfile);
		emo_puts(spool_out, "FROM ");
		emo_puts(spool_out, adata.smtp_sender);
		emo_puts(spool_out, CRLF);
		for (rcpt = adata.smtp_rcpts; *rcpt; rcpt++) {
			emo_puts(spool_out,"RCPT ");
			emo_puts(spool_out, *rcpt);
			emo_puts(spool_out, CRLF);
		}
		emo_puts(spool_out, "BODY");
		emo_puts(spool_out, CRLF);
		forge_received(spool_out);
		header_write(spool_out, adata.headers);
		if (bodyfile) {
			if (!(body_in = emi_open(bodyfile, 4096))) {
				syslog(LLEV,"fwtksyserr: cannot open spool file");
				writereply(c_out,451,0,m451msg, NULL);
				return(FAILURE);
			}
			emo_puts(spool_out, CRLF);
			http_copy(spool_out, 0, body_in, 0, -1, OUT);
			unlink(bodyfile);
		}
		smtp_close_spoolfile(spool_out);
		if (emo_error(spool_out)) {
			writereply(c_out,452,0,m452msg, NULL);
			return(ERROR);
		}
	} else {
		header_write(s_out, adata.headers);
		if (bodyfile) {
			if (!(body_in = emi_open(bodyfile, 4096))) {
				syslog(LLEV,"fwtksyserr: cannot open spool file");
				writereply(c_out,451,0,m451msg, NULL);
				return(FAILURE);
			}
			emo_puts(s_out, CRLF);
			http_copy(s_out, 0, body_in, 0, -1, OUT);
			emo_puts(s_out,".\r\n");
		}

		/* Truncated header already has "." at the end */

		emo_flush(s_out);
		if (emo_error(s_out)) {
			syslog(LLEV,"fwtksyserr: cannot write to server");
			writereply(c_out,455,0,"Remote server error", NULL);
			return(ERROR);
		}
		if (!smtp_response_ok(2))
			return(ERROR);
	}
	return(SUCCESS);
}

int
do_data(char *command, int flags, int argc, char **argv)
{
	long long	i;

	if (process_esmtp_options(1, argc, argv))
		return(1);


	if (s_out) {
		emo_puts(s_out, "DATA\r\n");
		emo_flush(s_out);
		if (!smtp_response_ok(3))
			return(1);
	}

	if (s_out && emo_error(s_out)) {
		syslog(LLEV, "fwtksyserr: couldn't write to output buffer, abandoning session");
		writereply(c_out, 421, 0, "421 message one", NULL);
		emo_flush(c_out);
		smtp_exit(EX_OSERR);
	}

	writereply(c_out, 354, 0, m354msg, NULL);
	state |= OK_DATA;

	proxy_update_status();
	switch (snarfdata(maxsize, &i)) {
	case SUCCESS:
	/*
	 * success 
	 */
		writereply(c_out,250,0,m250gotit, NULL);
		accumlog(LLEV, " bytes=%ld", i);
		accumlog(LLEV, 0);	/* flush */
		state &= !(OK_RCPT | OK_MAIL | OK_DATA);
		return(0);
	case FAILURE:
	/*
	 * error on client socket or fatal error in processing 
	 */
		smtp_nuke_spoolfile(spool_out);
		state &= !(OK_RCPT | OK_MAIL | OK_DATA);
		return(1);
	case ERROR:
	/*
	 * error on server socket or non-fatal error in processing,
	 * reply already sent to client 
	 */
	    	smtp_nuke_spoolfile(spool_out);
		state &= !(OK_RCPT | OK_MAIL | OK_DATA);
		return(1);
	default:
	/*
	 * muy trabajo 
	 */
		writereply(c_out,451,0,m451msg, NULL);
		syslog(LLEV,"fwtksyserr: internal logic error");
		smtp_nuke_spoolfile(spool_out);
		smtp_exit(EX_SOFTWARE);
	}
}


/*
 * The brains of this operation
 */

int
main(int argc, char **argv)
{
  	Cfg	*cfp;
	char	tokbuf[MAX_STR];
	char	*tokav[MAX_ARG];
	int	tokac;
    	char	line[SMTP_MAX_CMD_LINE + 1];

  	proxy_init(argc,argv);
  	proxy_SSLinit();
  	proxy_update_status();
	c_in = emi_fdopen(0, 4096);
	c_out = emo_fdopen(0, 4096);

	if (c_in == NULL || c_out == NULL) {
		syslog (LLEV, "fwtksyserr: fdopen() failed: %s", strerror(errno));
		exit(1);
	}
	emi_timeout (c_in, PROXY_TIMEOUT);
	emo_timeout (c_out, PROXY_TIMEOUT);


  	if ((cfp = proxy_conf_hosts(proxy_confp,proxy_stats.rladdr,
			   	    proxy_stats.riaddr)))
		proxy_parse_options(cfp,options);
  	else {
		if (!client_ssl)
			printf("550 Access denied\r\n");
		exit(1);
  	}

	if (client_ssl) {
		emio_sslaccept(c_in, c_out, proxy_stats.rladdr, proxy_stats.riaddr, 
			       "none", NULL, NULL);
		if (!c_in->sslio) {
			syslog(LLEV, "Cannot negotiate client SSL conenction");
			exit(1);
		}
	}

  	if (cdscp)
		proxy_set_dscp(0,cdscp);


	user_limit = proxy_conf_int(proxy_confp,"user-limit",1,sizeof(auth_username)-1,user_limit);
	pass_limit = proxy_conf_int(proxy_confp,"pass-limit",1,sizeof(auth_password)-1,pass_limit);	

/*
 *	998 is rfc2822 recommended limit for header line length
 */

  	maxsize = proxy_conf_long(proxy_confp,"maxbytes", 1, INT_MAX, LONG_MAX);
  	maxheader = proxy_conf_int(proxy_confp,"maxheaderline",128,1023,998);
  	maxrecip = proxy_conf_int(proxy_confp,"maxrecip",1,10000,2048);


  	if ((proxy_uid == 0) || (proxy_uid == -1)) {
    		syslog(LLEV, "fwtkcfgerr: sorry, I don't want to run as root! It's a bad idea!");
    		exit(EX_CONFIG);
  	}
  	if ((proxy_gid == 0) || (proxy_gid == -1)) {
    		syslog(LLEV, "fwtkcfgerr: sorry, I don't want to run as group 0. It's a bad idea!");
    		exit(EX_CONFIG);
  	}
      
	spooldir = proxy_conf_string(proxy_confp, "spooldir");
#ifdef STRICT_CHROOT
  	if (!*proxy_chroot) {
    		syslog(LLEV,"fwtkcfgerr: no directory specified");
    		abort();
  	}
#endif
  	if (spooldir == NULL) {
    		syslog(LLEV, "fwtkcfgerr: NULL spool directory! Aborting.");
    		abort();
  	}

	if (transp) 
		proxy_get_transparent_dst(proxy_stats.dst,&proxy_stats.port);

	if (*proxy_stats.dst)
		smtp_open_server();

  /*
   * Who's on the other end of this line?
   */


  /*
   * set who we are in case our caller didn't tell us to be someone
   * else 
   */

	proxy_hostname(0, my_clean_reverse_name, NULL, MAXHOSTNAMELEN);

  	writereply(c_out, 220, 0, my_clean_reverse_name, " ", m220msg, NULL);

#ifdef STRICT_CHROOT
  	proxy_chroot_setugid();
#else
  	proxy_setugid();
#endif

    	while (smtp_get_line(line, SMTP_MAX_CMD_LINE) >= 0) {
		const   struct  hash_entry      *he;
		int                             (*handler)(char*,int,int,char**);

		if (emi_error(c_in))
			break;
		if (!*line)
			continue;

		tokac = enargv(line,tokav,MAX_ARG,tokbuf,sizeof(tokbuf));
		if(!tokav[0])
			continue;
		if (!(he=find_he2(&esmtp_descr,tokav[0],strlen(tokav[0])))){
			writereply(c_out, 500, 0, m500msg, NULL);
			continue;
		} else {
			proxy_update_operation(line);
			if (he->flags & OP_DENY) {
				writereply(c_out, 502, 0, m502msg, NULL);
				syslog(LLEV,"deny host=%.512s/%.20s operation %.100s",proxy_stats.rladdr,proxy_stats.riaddr,line);
				continue;
			}
			if ((he->flags & STATE_MASK & state) != (he->flags & STATE_MASK)) {
				writereply(c_out, 503, 0, m503msg, NULL);
				continue;
			}
			if (!sane_state(state)) {
				syslog(LLEV, "fwtksyserr: CRITICAL - bad state, aborting");
				abort();
			}
			handler = (int(*)(char*,int,int,char**))he->handler;
			assert (ESMTP_HANDLER ((int(*)())handler));
			handler(line,he->flags,tokac,tokav);

			if (!sane_state(state)) {
				syslog(LLEV, "fwtksyserr: CRITICAL - bad state, aborting");
				abort();
			}
			emo_flush(c_out);
			continue;
		}
	}
	smtp_exit(EX_OK);
}
