/**
 * @file   gsb_common.c
 * @author Alex Ott <alexott@gmail.com>
 * 
 * @brief  Implementation of common routines
 * 
 * 
 */

#include "gsb_common.h"

/** 
 * Create and initialize 
 * 
 * @param name name of gsb_file
 * 
 * @return newly allocated structure, that keep initial data
 */
gsb_file* gsb_init_file(const char* name) {
	gsb_file *file=(gsb_file*)malloc(sizeof(gsb_file));
	if(!file)
		return NULL;

	file->name=strdup(name);
	file->minor_version=-1;
	file->major_version=-1;
	file->count=0;
	bzero(&file->modtime,sizeof(struct timespec));

	return file;
}

/** 
 * Free memory allocated for gsb_file, and all corresponding internal structures
 * 
 * @param file pointer to gsb_file
 */
void gsb_free_file(gsb_file* file) {
	if(!file)
		return ;
	free(file->name);

/* 	free structures that hold actual data */
	
	free(file);
}

/** 
 * Read given file from disk
 * 
 * @param fname file name
 * 
 * @return newly allocated structure, that keep all data
 */
gsb_file* gsb_read_file(const char* fname) {

	return NULL;
}


